(* ---------------------------------------------------------------
Title         Q&D Atlas
Author        PhG
Overview      find exact or approximate match in ATLAS.RSC
Usage         see help
Notes         very, very, very quick & dirty... :-(
Bugs
Wish List

--------------------------------------------------------------- *)

MODULE Atlas;

IMPORT FIO;
IMPORT Str;
IMPORT Lib;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, chkJoker,
isOption, GetOptIndex, GetLongCard, GetString, CharCount, same, cleantabs,
aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, str16, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
Elapsed,cmdSHOW,Animation,Work,pathtype;

FROM QD_Names IMPORT ProgCopyright;


FROM IO IMPORT WrStr, WrLn;

FROM QD_Sky IMPORT parseLongitude;

FROM SYSTEM IMPORT BYTE;

(* ------------------------------------------------------------ *)

TYPE
    lambdafmttype = (asis,forcehms,forcedeg);


CONST
    ProgEXEname   = "ATLAS";
    ProgTitle     = "Q&D Atlas (Win32)";
    ProgVersion   = "v0.9 (v1.0j DOS)";
    ProgBanner    = ProgTitle+" "+ProgVersion+" "+ProgCopyright;
CONST
    extEXE        = ".EXE"; (* upper case *)
    extRSC        = ".RSC";
CONST
    errNone          = 0;
    errHelp          = 1;
    errOption        = 2;
    errTooManyTowns  = 3;
    errAtlasNotFound = 4;
    errDept          = 5;
    errAtlasLayout   = 6;
    errFmt           = 7;
    errAborted       = 8;


VAR
    DEBUG         : BOOLEAN;
CONST
    nullchar    = CHR(0);
    semicolon   = ";";
    pound       = "#";
    openbracket = "[";
    closebracket= "]";
    quote       = "'";
    doublequote = '"';
    blank       = " ";
    tab         = CHR(9);
    coma        = ",";
    jokerStar     = "*";
    jokerQuestion = "?";

CONST
    alpha   = "abcdefghijklmnopqrstuvwxyz";
    special = ""+jokerStar+jokerQuestion;
    letters = alpha+special;
CONST
    msgAtLine      = ", ligne ";
    FRsection      = "france"; (* lower case required here *)
    FRsectionAL    = "france::al"; (* pirated from AstroLab *)
    sepID          = "::";
    sep            = " "; (* was "  " *)
    sepcolon       = " : ";
    sepline        = "------------------------------------------------------------";
    msgAllWanted   = "Donne approximative recherche";
    msgExactWanted = "Donne exacte recherche";
    msgDeptWanted  = "Dpartement recherch";
    msgNotFiltered = "Les entres identiques ne sont PAS filtres";
    msgFiltered    = "Les entres identiques sont filtres";
    pleasewait     = "Patience, recherche en cours... ";

(* ------------------------------------------------------------ *)

PROCEDURE eyeCandy(command:CARDINAL);
VAR
    i : CARDINAL;
BEGIN
    Animation(command); (* elapsed time would be silly here for search can restart *)
END eyeCandy;

VAR
  PATIENCE : BOOLEAN;

PROCEDURE StopCandy();
BEGIN
    eyeCandy(cmdStop);
    video(pleasewait,FALSE);
    PATIENCE:=FALSE;
END StopCandy;

PROCEDURE StartCandy();
BEGIN
    video(pleasewait,TRUE);
    eyeCandy(cmdInit);
    eyeCandy(cmdSHOW);
    PATIENCE:=TRUE;
END StartCandy;

PROCEDURE BusyCandy ();
BEGIN
  eyeCandy(cmdShow);
END BusyCandy;

(* ------------------------------------------------------------ *)

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
    nl     = 015C+012C; (* CHR(13)+CHR(10); *)
VAR
    S : str256;
BEGIN
  IF PATIENCE=TRUE THEN StopCandy();  END;
    CASE e OF
    | errHelp :
        (*
               00000000011111111112222222222333333333344444444445555555555666666666677777777778
               1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
        *)
        WrLn;
        WrStr("Titre   : "+ProgBanner+nl);
        WrLn;
        WrStr("Syntaxe : "+ProgEXEname+" <ville|dpartement|numro de dpartement>... [option]..."+nl);
        WrLn;
        WrStr("Utilit : si possible, trouver dans l'atlas "+ProgEXEname+extRSC+" une correspondance"+nl);
        WrStr("          exacte ou approche pour chaque <ville>, <dpartement> ou"+nl);
        WrStr("          <numro de dpartement>"+nl);
        WrLn;
        WrStr("Options : /t pour trouver toutes les correspondances, exactes ou non"+nl);
        WrStr("             (la recherche utilise alors la phontique et les codes Soundex)"+nl);
        WrStr("          /d pour ne plus filtrer les entres strictement identiques"+nl);
        WrStr("             (la comparaison est limite aux 500 premires entres)"+nl);
        WrStr("          /lh pour afficher la longitude en heures, minutes et secondes"+nl);
        WrStr("          /ld pour afficher la longitude en degrs et minutes"+nl);
        WrStr("          /i pour afficher l'identificateur (atlas source de l'entre)"+nl);
        WrLn;
        WrStr("Note 1  : Pour les villes de France, attention aux ventuelles variations :"+nl);
        WrStr("          L', La, Le, Les, sur, sous, Saint, Sainte, etc."+nl);
        WrStr("Note 2  : La recherche d'une ville peut exploiter les caractres gnriques :"+nl);
        WrStr("          ? correspond  un seul caractre, *  aucun ou plusieurs de suite."+nl);
        WrStr("Note 3  : Pour d'videntes raisons, certaines villes peuvent figurer"+nl);
        WrStr("          dans l'atlas sous deux formes : franaise et anglaise."+nl);
        WrStr("Note 4  : Les coordonnes gographiques fournies sont indicatives ! ;-)"+nl);
        WrLn;
        WrStr("Exemple : "+ProgEXEname+" Paris 42 "+doublequote+"Saint Etienne"+doublequote+" Londres London "+doublequote+"Paris Mairie VIII"+doublequote+nl);
    | errOption :
        Str.Concat(S,"L'option ",einfo);
        Str.Append(S," est illgale !");
    | errTooManyTowns :
        Str.Concat(S,"La ville ",einfo);
        Str.Append(S," est vraiment de trop !");
    | errAtlasNotFound :
        Str.Concat(S,"L'atlas (",einfo);
        Str.Append(S,") est introuvable !");
    | errDept :
        Str.Concat(S,"Le dpartement ",einfo);
        Str.Append(S," n'existe pas !");
    | errAtlasLayout :
            Str.Concat(S,"Le format de l'atlas (",einfo);
            Str.Append(S,") est incorrect !");
    | errFmt:
        S := "Les options /lh et /ld sont mutuellement exclusives !";
    | errAborted :
        S := "Recherche interrompue par l'utilisateur !";


    ELSE
        S := "This is illogical, Captain !";
    END;
    IF (e <> errNone) AND (e <> errHelp) THEN
        WrLn;
        WrStr(ProgEXEname+" : ");
        WrStr(S);
        WrLn;
    END;

    (* FPUalert(); *)

    (* Lib.SetReturnCode(SHORTCARD(e)); *)
    HALT (e);
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    firstdept = 1;
    lastdept  = 95;

PROCEDURE onlyDigits (S:ARRAY OF CHAR) : BOOLEAN;
CONST
    digits = "0123456789";
VAR
    i : CARDINAL;
BEGIN
    FOR i := 0 TO Str.Length(S)-1 DO
        IF Belongs(digits,S[i])=FALSE THEN RETURN FALSE; END;
    END;
    RETURN TRUE;
END onlyDigits;

PROCEDURE getCardinal (S : ARRAY OF CHAR;VAR n : CARDINAL) : BOOLEAN;
VAR
    v  : LONGCARD;
    ok : BOOLEAN;
BEGIN
    v := Str.StrToCard(S,10,ok);
    IF ok = FALSE THEN RETURN FALSE; END;
    IF v > MAX(CARDINAL) THEN RETURN FALSE; END;
    n := CARDINAL(v);
    RETURN TRUE;
END getCardinal;

PROCEDURE chkRange (S:ARRAY OF CHAR;min,max:CARDINAL) : BOOLEAN;
VAR
    n  : CARDINAL;
    ok : BOOLEAN;
BEGIN
    IF getCardinal(S,n)=FALSE THEN RETURN FALSE; END;
    IF (n < min) OR (n > max) THEN RETURN FALSE; END;
    RETURN TRUE;
END chkRange;

PROCEDURE CleanStr (VAR S : ARRAY OF CHAR);
BEGIN
    LtrimBlanks(S);
    Ltrim(S,quote);
    Ltrim(S,doublequote);
    RtrimBlanks(S);
    Rtrim(S,quote);
    Rtrim(S,doublequote);
END CleanStr;

PROCEDURE CleanSectionStr (VAR S : ARRAY OF CHAR );
BEGIN
    Ltrim(S,openbracket);
    Rtrim(S,closebracket);
    CleanStr(S);
END CleanSectionStr;

PROCEDURE Filter (legal:ARRAY OF CHAR;VAR S : ARRAY OF CHAR);
VAR
    R : str128;
    i : CARDINAL;
BEGIN
    Str.Copy(R,"");
    FOR i := 0 TO Str.Length(S)-1 DO
        IF Belongs(legal,S[i]) THEN
            Str.Append(R,S[i]);
        END;
    END;
    Str.Copy(S,R);
END Filter;

(* assume legal "town,latitude,longitude" without quotes *)
PROCEDURE ParseEntry (S:ARRAY OF CHAR;c:CHAR;
                      VAR town,latitude,longitude:ARRAY OF CHAR);
BEGIN
    Str.ItemS(town,S,coma,0);
    Str.ItemS(latitude,S,coma,1);
    Str.ItemS(longitude,S,coma,2);
END ParseEntry;

PROCEDURE samedata (S1,S2:ARRAY OF CHAR;tryJoker:BOOLEAN):BOOLEAN;
BEGIN
    CASE tryJoker OF
    | FALSE : RETURN same(S1,S2);
    | TRUE  : RETURN Str.Match(S2,S1); (* S1 is wanted pattern *)
    ELSE ; (* //ISO fixed silly requirement *) END;
END samedata;

(* ------------------------------------------------------------ *)

(* tester les diffrents rsultats avec "cayenne" et "saintlouis" *)

TYPE
    dumpedentry = RECORD
        id   : str128;
        hash : CARDINAL;
    END;
CONST
    firstdumped = 1;
    maxdumped   = 500; (* should be enough ! *)
VAR
    allowdups   : BOOLEAN;
    lastdumped  : CARDINAL;
    dumped      : ARRAY [firstdumped..maxdumped] OF dumpedentry;

PROCEDURE initVerbose (flag:BOOLEAN);
BEGIN
    allowdups  := flag;
    lastdumped := firstdumped-1; (* 1-1=... ;) *)
END initVerbose;

PROCEDURE isDuplicate (S:ARRAY OF CHAR):BOOLEAN;
CONST
    hashrange = MAX(CARDINAL);
VAR
    i    : CARDINAL;
    hash : CARDINAL;
    rc   : BOOLEAN;
BEGIN
    rc := FALSE; (* default to allow entry dump *)
    IF allowdups THEN RETURN rc; END; (* yes here only *)

    hash :=Lib.HashString(S,hashrange); (* try and speed up string lookup *)

    i := firstdumped;
    LOOP
        IF i > lastdumped THEN (* not found, so add it *)
            IF i > maxdumped THEN
                (* no room to add new string
                   let's make it the management way :
                   fake it's ok by doing nothing *)
            ELSE
                Str.Copy(dumped[i].id,S);
                dumped[i].hash := hash;
                lastdumped := i;
            END;
            EXIT;
        END;
        IF dumped[i].hash = hash THEN (* first check *)
            IF same(dumped[i].id,S) THEN (* found *)
                rc:=TRUE;
                EXIT;
            END;
        END;
        INC(i);
    END;
    RETURN rc;
END isDuplicate;

(* ------------------------------------------------------------ *)

(*

;
; coordonnes gographiques
;

[France]

; atlas de France (merci  Le Corre !)
; dpartement, puis prfecture, puis arrondissement(s) et ville(s)
; dpartement
;     prfecture
;     arrondissement
;         ville
; "nom,latitude,longitude" (ne pas oublier les "" !!!)
; ddN/Smm hHmmMssE/W with leading zeroes

"Ain"
 "Bourg,46N12,0H20M55E"
 "Belley,45N46,0H22M44E"
 "Gex,46N20,0H24M12E"
 "Nantua,46N09,0H22M24E"
"Aisne"
 "Laon,49N34,0H14M30E"
 "Saint-Quentin,49N51,0H13M08E"
 "Soissons,49N23,0H13M20E"
 "Vervins,49N50,0H15M40E"

......

"Val-de-Marne"
 "Crteil,48N47,0H09M52E"
 "Nogent-sur-Marne,48N50,0H10M00E"
  "Saint-Mand",48N50,0H09M40E"
"Val-d'Oise"
 "Pontoise,49N03,0H08M20E"
 "Argenteuil,48N57,0H09M00E"
 "Montmorency,48N59,0H09M16E"

;
; autres atlas
;

[Principaut d'Andorre]

"Andorre"
 "Andorre-la-Vieille,42N30,0H06M00E"

[Principaut de Monaco]

"Monaco"
 "Monte-Carlo,43N45,0H29M36E"

*)

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead; (* was 24 *)
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
VAR
    ioBuffer        : ARRAY [firstBufferByte..lastBufferByte] OF BYTE;

(* ------------------------------------------------------------ *)

TYPE
    matchmode = (exact,phonetic,soundex,soundexorg,numero);

PROCEDURE FRspecial (country:ARRAY OF CHAR;departement:CARDINAL):BOOLEAN;
BEGIN
    (* this check prevents problems with several FRANCE sections ! *)
    (* departement is then 0 so we abort here *)
    IF (departement < firstdept) OR (departement > lastdept) THEN RETURN FALSE; END;
    (* now back to previous normal code *)
    LowerCase(country);
    RETURN (same(country,FRsection) OR same(country,FRsectionAL));
END FRspecial;

PROCEDURE infoprefix ( mode:matchmode ; tryJoker:BOOLEAN);
VAR
    S:str128;
BEGIN
    CASE mode OF
    | exact :
      IF tryJoker THEN
        S:="Joker !    : ";
      ELSE
        S:="Identique  : ";
      END;
    | phonetic :
        S:="Phontique : ";
    | soundex :
        S:="Soundex    : ";
    | soundexorg :
        S:="SoundexORG : ";
    | numero :
        S:="Prfecture : ";
    ELSE ; (* //ISO fixed silly requirement *) END;
    WrStr(S);
END infoprefix;

PROCEDURE fmtval (v:LONGCARD;digits:CARDINAL;padchar:CHAR):str16;
VAR
    ok:BOOLEAN;
    S:str16;
    i:CARDINAL;
BEGIN
    Str.CardToStr(v,S,10,ok);
    FOR i:=1 TO digits DO
        IF Str.Length(S) < i THEN Str.Prepend(S,padchar);END;
    END;
    RETURN S;
END fmtval;

PROCEDURE fmtLambda (lambdafmt:lambdafmttype;longitude:ARRAY OF CHAR ):str128;
CONST
    sEst  = "e";
    sOuest= "o";
CONST
    degRmul = 60.0;  (* degrees to minutes of arc *)
    degmax  = 60 * 360;
    hmsRmul = 240.0; (* degrees to seconds of time : 360--24*60*60 1--240 *)
    hmsmax  = 24 * 60 * 60;
VAR
    R : str128;
    c : CHAR;
    lambdadeg:LONGREAL;
    lambdasec:LONGINT;
    ok:BOOLEAN;
    Rmul:LONGREAL;
    maxi,v,hh,mm,ss,dd:LONGCARD;
BEGIN
    CASE lambdafmt OF
    | asis :
        Str.Copy(R,longitude);
    ELSE
        ok:=parseLongitude (longitude, lambdadeg,lambdasec);
        IF lambdadeg < 0.0 THEN
            c:=sEst;
        ELSE
            c:=sOuest;
        END;
        lambdadeg:=ABS(lambdadeg);
        CASE lambdafmt OF
        | forcehms: Rmul:=hmsRmul; maxi:=hmsmax;
        | forcedeg: Rmul:=degRmul; maxi:=degmax;
        ELSE ; (* //ISO fixed silly requirement *) END;
        v := VAL (LONGCARD,(lambdadeg * Rmul + 0.5));
        IF v >= maxi THEN v := 0; END;
        CASE lambdafmt OF
        | forcehms:
           (* degrees / 15 = hours, then hours * 60 * 60 = seconds *)
           hh := v DIV (60 * 60);
           mm := (v - hh * 60 * 60) DIV 60;
           ss := (v - hh * 60 * 60) MOD 60;
           Str.Concat(R, "~h~m~", c);
           Str.Subst(R,"~",fmtval(hh,2,"0"));
           Str.Subst(R,"~",fmtval(mm,2,"0"));
           Str.Subst(R,"~",fmtval(ss,2,"0"));
        | forcedeg:
           dd := v DIV 60;
           mm := v MOD 60;
           Str.Concat(R, "~~", c);
           Str.Subst(R,"~",fmtval(dd,3,"0"));
           Str.Subst(R,"~",fmtval(mm,2,"0"));
        ELSE ; (* //ISO fixed silly requirement *) END;
    END;
    RETURN R;
END fmtLambda;

PROCEDURE dmpDept (deptnum:CARDINAL;country,deptname,entry:ARRAY OF CHAR;
                  mode:matchmode;showid:BOOLEAN;
                  lambdafmt:lambdafmttype);
VAR
    ok          : BOOLEAN;
    S           : str128;
    departement : str128;
    town        : str128;
    latitude    : str128;
    longitude   : str128;
    p           : CARDINAL;
BEGIN
    ParseEntry(entry,coma,town,latitude,longitude);
    infoprefix(mode,FALSE);
    WrStr(latitude);
    WrStr(sep);

    WrStr( fmtLambda(lambdafmt,longitude) );

    WrStr(sep);
    WrStr(town);
    WrStr(sep);
    IF showid THEN
        WrStr(country);
    ELSE
        Str.Copy(S,country);
        p:=Str.Pos(S,sepID);
        IF p # MAX(CARDINAL) THEN S[p]:=nullchar ;END;
        WrStr(S);
    END;
    WrStr(sep);
    IF FRspecial(country,deptnum) THEN
        Str.CardToStr(LONGCARD(deptnum),S,10,ok);
        IF deptnum < 10 THEN Str.Prepend(S,"0"); END;
        Str.Concat(departement,deptname," (");
        Str.Append(departement,S);
        Str.Append(departement,")");
        WrStr(departement);
    END;
    WrLn;
END dmpDept;

PROCEDURE dmpTown (deptnum:CARDINAL;
                   country,deptname,town,latitude,longitude:ARRAY OF CHAR;
                   mode:matchmode;tryJoker,showid:BOOLEAN;
                   lambdafmt:lambdafmttype);
VAR
    ok          : BOOLEAN;
    S           : str128;
    departement : str128;
    p           : CARDINAL;
BEGIN
    infoprefix(mode,tryJoker);
    WrStr(latitude);
    WrStr(sep);

    WrStr( fmtLambda(lambdafmt,longitude) );

    WrStr(sep);
    WrStr(town);
    WrStr(sep);
    IF showid THEN
        WrStr(country);
    ELSE
        Str.Copy(S,country);
        p:=Str.Pos(S,sepID);
        IF p # MAX(CARDINAL) THEN S[p]:=nullchar ;END;
        WrStr(S);
    END;
    WrStr(sep);
    IF FRspecial(country,deptnum) THEN
        Str.CardToStr(LONGCARD(deptnum),S,10,ok);
        IF deptnum < 10 THEN Str.Prepend(S,"0"); END;
        Str.Concat(departement,deptname," (");
        Str.Append(departement,S);
        Str.Append(departement,")");
        WrStr(departement);
    END;
    WrLn;
END dmpTown;

(* ------------------------------------------------------------ *)

PROCEDURE ComputeCode (town:ARRAY OF CHAR;mode:matchmode;
                       VAR R : ARRAY OF CHAR);
BEGIN
    LowerCase(town);
    FixAE(town);                 (* just in case *)
    Filter(letters,town);
    CASE mode OF
    | exact :
        Str.Copy(R,town);
    | phonetic :
        CodePhonetic(town,R);
    | soundex :
        CodeSoundex(town,R);
    | soundexorg :
        CodeSoundexOrg(town,R);
    ELSE ; (* //ISO fixed silly requirement *) END;
END ComputeCode;

(* ------------------------------------------------------------ *)

PROCEDURE FindMatch (file,town:ARRAY OF CHAR;mode:matchmode;
                    showid:BOOLEAN;lambdafmt:lambdafmttype);
VAR
    hnd            : FIO.File;
    S              : str256;
    wantedcode     : str128;
    currcode       : str128;
    state          : (waiting,deptMatch,alert);
    ok             : BOOLEAN;
    currdept       : CARDINAL;
    countryname    : str128;
    deptname       : str128;
    townname       : str128;
    latitude       : str128;
    longitude      : str128;
    currline       : CARDINAL;
    i              : CARDINAL;
    c              : CHAR;
    tryJoker       : BOOLEAN;
BEGIN

  StartCandy();

    CleanStr(town);
    ComputeCode(town,mode,wantedcode);
    IF Str.CharPos(town,jokerStar) # MAX(CARDINAL) THEN
        tryJoker:=TRUE;
    ELSIF Str.CharPos(town,jokerQuestion) # MAX(CARDINAL) THEN
        tryJoker:=TRUE;
    ELSE
        tryJoker:=FALSE;
    END;

    state := waiting;
    currline := 0;
    Str.Copy(countryname,"");

    hnd := FIO.OpenRead(file);
    FIO.AssignBuffer(hnd,ioBuffer);
    FIO.EOF:=FALSE; (* safety ! *)
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hnd,S);
        IF (S[0]=nullchar) AND FIO.EOF THEN EXIT; END; (* weird hack ! *)
        IF ChkEscape() THEN
            StopCandy();
            abort(errAborted,"");
        END;

  BusyCandy();


        INC(currline);
        CleanStr(S); (* remove leading and trailing blanks, tabs, quotes *)
        CASE S[0] OF
        | nullchar,semicolon,pound :
            (* ignore empty string or comment *)
        | openbracket :
            CleanSectionStr(S);
            Str.Copy(countryname,S);
            currdept := firstdept-1;
        ELSE
            CASE CharCount(S,coma) OF
            | 2 : (* "town,latitude,longitude" *)
                CASE state OF
                | deptMatch :
  StopCandy();
                    dmpDept(currdept,countryname,deptname,S,mode,showid,lambdafmt);
  StartCandy();
                    state:=waiting;
                ELSE
                    ParseEntry(S,coma,townname,latitude,longitude);
                    ComputeCode(townname,mode,currcode);
                    IF samedata(wantedcode,currcode,tryJoker) THEN
  StopCandy();
                        IF isDuplicate(S)=FALSE THEN (* town entries only *)
                            dmpTown(currdept,countryname,deptname,townname,latitude,longitude,mode,tryJoker,showid,lambdafmt);
                        END;
                        (* IF ChkEscape() THEN abort(errAborted,""); END; (* candy turned off here *) *)
  StartCandy();
                    END;
                END;
            ELSE
                CASE state OF
                | deptMatch :
                    state:=alert;
                    EXIT; (* departement then departement again ! *)
                ELSE
                    INC(currdept);
                    Str.Copy(deptname,S);
                    ComputeCode(deptname,mode,currcode);
                    IF same(wantedcode,currcode) THEN state:=deptMatch;END;
                END;
            END;
        END;
    END;
    FIO.Close(hnd);

  StopCandy();

    (* here, S contains offending line *)
    IF state = alert THEN
        Str.CardToStr(LONGCARD(currline),S,10,ok);
        Str.Prepend(S,msgAtLine);
        Str.Prepend(S,file);
        abort(errAtlasLayout,S);
    END;
END FindMatch;

(* ------------------------------------------------------------ *)

PROCEDURE FindDept (file,deptnumstr:ARRAY OF CHAR;
                   showid:BOOLEAN;lambdafmt:lambdafmttype);
VAR
    hnd            : FIO.File;
    S              : str256;
    state          : (waiting,FranceFound,DeptFound,PrefectureFound);
    ok             : BOOLEAN;
    wanteddept     : CARDINAL;
    currdept       : CARDINAL;
    countryname    : str128;
    deptname       : str128;
    currline       : CARDINAL;
    i              : CARDINAL;
    c              : CHAR;
BEGIN
  StartCandy();
    ok := getCardinal(deptnumstr,wanteddept); (* already checked, so always TRUE *)

    state := waiting;
    currline := 0;
    Str.Copy(countryname,"");

    hnd := FIO.OpenRead(file);
    FIO.AssignBuffer(hnd,ioBuffer);
    FIO.EOF:=FALSE; (* safety ! *)
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hnd,S);
        IF (S[0]=nullchar) AND FIO.EOF THEN EXIT; END; (* weird hack ! *)
  BusyCandy();

        INC(currline);
        CleanStr(S); (* remove leading and trailing blanks, tabs, quotes *)
        CASE S[0] OF
        | nullchar,semicolon,pound :
            (* ignore empty string or comment *)
        | openbracket :
            CASE state OF
            | waiting :
                CleanSectionStr(S);
                Str.Copy(countryname,S);
                LowerCase(S);
                Filter(letters,S);
                IF same(S,FRsection) THEN
                    state    := FranceFound;
                    currdept := firstdept-1;
                END;
            ELSE
                EXIT; (* argh, nothing between [France] and next section ! *)
            END;
        ELSE
            CASE state OF
            | FranceFound :
                 IF CharCount(S,coma) # 2 THEN (* not "town,latitude,longitude" *)
                     INC(currdept);
                     IF currdept > lastdept THEN EXIT; END; (* data for > 95 dept !!! *)
                     IF currdept = wanteddept THEN
                         Str.Copy(deptname,S);
                         state := DeptFound;
                     END;
                 END;
            | DeptFound :
                 IF CharCount(S,coma) =2 THEN
                     state := PrefectureFound;
                 END;
                 EXIT; (* either ok or not expected "town,latitude,longitude" *)
            ELSE ; (* //ISO fixed silly requirement *) END;
        END;
    END;
    FIO.Close(hnd);

  StopCandy();

    (* here, S either contains offending line or town,latitude,longitude *)
    IF state # PrefectureFound THEN
        Str.CardToStr(LONGCARD(currline),S,10,ok);
        Str.Prepend(S,msgAtLine);
        Str.Prepend(S,file);
        abort(errAtlasLayout,S);
    END;
    dmpDept(wanteddept,countryname,deptname,S,numero,showid,lambdafmt);
END FindDept;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    firsttown = 1; (* yes, [1.. required ! *)
    maxtown   = 32;
VAR
    town      : ARRAY[firsttown..maxtown] OF str128;
    lasttown  : CARDINAL;
VAR
    parmcount : CARDINAL;
    i         : CARDINAL;
    opt       : CARDINAL;
    S         : pathtype;
    R         : pathtype;
    atlasfile : pathtype;
    tryall    : BOOLEAN;
    showid    : BOOLEAN;
    lambdafmt : lambdafmttype;
    firsttime : BOOLEAN; (* should it be in external module ? *)
BEGIN
    (* Lib.DisableBreakCheck(); *)
    FIO.IOcheck := FALSE;
    DEBUG := FALSE;
  PATIENCE:=FALSE;
    tryall := FALSE;
    showid := FALSE;
    lambdafmt:= asis;
    initVerbose(FALSE);

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,""); END;

    lasttown := firsttown-1;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R,"?"+delim+"H"+delim+"HELP"+delim+
                                 "DEBUG"+delim+
                                 "T"+delim+"ALL"+delim+
                                 "D"+delim+"DUPS"+delim+
                                 "LH"+delim+"HMS"+delim+
                                 "LD"+delim+"DM"+delim+"DEG"+delim+
                                 "I"+delim+"ID");
            CASE opt OF
            | 1,2,3 :   abort(errHelp,"");
            | 4 :       DEBUG := TRUE;
            | 5,6 :     tryall := TRUE;
            | 7,8 :     initVerbose(TRUE);
            | 9,10:     CASE lambdafmt OF
                        | asis,forcehms : lambdafmt:=forcehms;
                        ELSE
                            abort(errFmt,"");
                        END;
            | 11,12,13: CASE lambdafmt OF
                        | asis,forcedeg : lambdafmt:=forcedeg;
                        ELSE
                            abort(errFmt,"");
                        END;
            | 14,15:    showid := TRUE;
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            INC(lasttown);
            IF lasttown > maxtown THEN abort(errTooManyTowns,S); END;
            IF onlyDigits(S) THEN
                IF chkRange(S,firstdept,lastdept)=FALSE THEN abort(errDept,S); END;
            END;
            Str.Copy(town[lasttown],S);
        END;
    END;
    IF lasttown<firsttown THEN abort(errHelp,""); END;

    Lib.ParamStr(atlasfile,0);
    UpperCase(atlasfile); (* useless... but for added safety ! *)
    Str.Subst(atlasfile,extEXE,extRSC);
    IF FIO.Exists(atlasfile)=FALSE THEN abort(errAtlasNotFound,atlasfile); END;

    (* FPUfix(); *)

    WrLn;
    WrStr(ProgBanner);
    WrLn;

    FOR i := firsttown TO lasttown DO
        WrLn;
        IF onlyDigits(town[i]) THEN
            WrStr(msgDeptWanted);
        ELSE
            IF tryall THEN
                WrStr(msgAllWanted);
            ELSE
                WrStr(msgExactWanted);
            END;
        END;

        WrStr(sepcolon);
        WrStr(town[i]);
        WrLn;

        IF allowdups THEN
            WrStr(msgNotFiltered);
        ELSE
            WrStr(msgFiltered);
        END;
        WrLn;
        WrLn;

        IF onlyDigits(town[i]) THEN
            FindDept(atlasfile,town[i],showid,lambdafmt);
        ELSE
            FindMatch(atlasfile,town[i],exact,showid,lambdafmt);
            IF tryall THEN
                FindMatch(atlasfile,town[i],phonetic,showid,lambdafmt);
                FindMatch(atlasfile,town[i],soundex,showid,lambdafmt);
                (* FindMatch(atlasfile,town[i],soundexorg); *)
            END;
        END;
        IF i < lasttown THEN
            WrLn;
            WrStr(sepline);
            WrLn;
        END;
    END;

    abort(errNone,"");
END Atlas.