(* ---------------------------------------------------------------
Title         Q&D Calc
Author        PhG
Overview      some use(d|ful) subroutines for Q&D astro tools
Notes		  must be alone to avoid dynamic pool overflow
Bugs
Wish List

--------------------------------------------------------------- *)

IMPLEMENTATION MODULE QD_Calc;

FROM QD_Sky IMPORT int,frac,mod,JDtoJD0H;

FROM MATHLIB IMPORT Sin,Cos,Tan,ATan2,ASin,Sqrt,ATan;

(* ------------------------------------------------------------ *)

(* stripped from SkyCalc and SolarSys *)

CONST
    longzero = LONGREAL(0.0);

VAR
    PI         : LONGREAL;
    MaxDegrees : LONGREAL;
    MaxRadians : LONGREAL;
    DegRad     : LONGREAL;

PROCEDURE DegreesToRadians (VAR Angle : LONGREAL);
BEGIN
    Angle := mod (Angle,MaxDegrees);
    Angle := Angle * DegRad;
END DegreesToRadians;

PROCEDURE SecondsToRadians (VAR Angle : LONGREAL);
BEGIN
    Angle := mod (Angle / 3600.0,MaxDegrees);
    Angle := Angle * DegRad;
END SecondsToRadians;

PROCEDURE RadiansToDegrees (VAR Angle : LONGREAL);
BEGIN
    Angle := mod (Angle,MaxRadians);
    Angle := Angle / DegRad;
END RadiansToDegrees;

(* warning ! we MAY get 360 instead of 0 ! *)

PROCEDURE Mod360 (VAR Angle : LONGREAL); (* for angles whose range is 0-360 *)
BEGIN
    Angle := mod (Angle,MaxDegrees); (* should be ALREADY done but... *)
    IF Angle < longzero THEN
        Angle := Angle + MaxDegrees;
        Angle := mod (Angle,MaxDegrees); (* probably useless... *)
(*
    ELSIF Angle >= MaxDegrees THEN
        Angle := Angle - MaxDegrees;
        (* Angle := mod (Angle,MaxDegrees); *)
*)
    END;
END Mod360;

PROCEDURE ModRadians (VAR Angle : LONGREAL); (* Mod360 in radians *)
BEGIN
    Angle := mod (Angle,MaxRadians); (* should be ALREADY done but... *)
    IF Angle < longzero THEN
        Angle := Angle + MaxRadians;
        Angle := mod (Angle,MaxRadians); (* probably useless... *)
    END;
END ModRadians;

(* === GLOBAL VARIABLES ======================================= *)

(* On sait, on sait... mais c'est (un peu) plus rapide comme a ! *)

VAR
    X,Y    : LONGREAL; (* coordonnes rectangulaires du Soleil *)
    l      : LONGREAL; (* longitude hliocentrique vraie *)
    b      : LONGREAL; (* latitude hliocentrique vraie *)
    r      : LONGREAL; (* rayon vecteur *)
    lambda : LONGREAL; (* longitude cliptique -- gocentrique *)
    beta   : LONGREAL; (* latitude cliptique *)
    alpha  : LONGREAL; (* ascension droite -- quatoriale *)
    delta  : LONGREAL; (* declinaison *)
VAR
    T,T2,T3       : LONGREAL;

PROCEDURE CalcT (JD : LONGREAL);
BEGIN
    T  := (JD - 2415020.0) / 36525.0;
    T2 := T * T;
    T3 := T * T2;
END CalcT;

(* === GLOBAL SETUP =========================================== *)

VAR
    L0,M0    : LONGREAL;
    L1,M1,U1 : LONGREAL;
    L2,M2,U2 : LONGREAL;
    L3,M3,U3 : LONGREAL;
    L4,M4,U4 : LONGREAL;
    L5,M5,U5 : LONGREAL;
    L6,M6,U6 : LONGREAL;
    L7,M7,U7 : LONGREAL;
    L8,M8,U8 : LONGREAL;
    L9,M9    : LONGREAL;
    OmegaLune : LONGREAL; (* longitude du noeud ascendant *)
    Epsilon   : LONGREAL; (* obliquit de l'cliptique *)
    NutationLambda  : LONGREAL;
    NutationEpsilon : LONGREAL;
    NutationSoleil  : LONGREAL;

(*
  -- lments moyens des orbites plantaires
     (rapports  l'quinoxe moyen de la date)
  -- L = longitude moyenne
     M = anomalie moyenne
     U = argument de latitude
  -- procdure  excuter AVANT tout autre calcul !!!
*)

PROCEDURE CalcSetup ();
VAR
    OmegaNutation : LONGREAL; (* temporaire et # d'OmegaLune *)
BEGIN
    L0 := 279.6964027 +  36000.7695173 * T;     (* Soleil *)
    M0 := 358.4758635 +  35999.0494965 * T;
    L1 := 178.178814  + 149474.071386  * T;     (* Mercure *)
    M1 := 102.279426  + 149472.515334  * T;
    U1 := 131.032888  + 149472.885872  * T;
    L2 := 342.766738  +  58519.212542  * T;     (* Vnus *)
    M2 := 212.6018923 +  58517.8063877 * T;
    U2 := 266.987445  +  58518.311835  * T;
    L3 := 293.747201  +  19141.699879  * T;     (* Mars *)
    M3 := 319.5292728 +  19139.8588872 * T;
    U3 := 244.960887  +  19140.928953  * T;
    L4 := 237.352259  +   3034.906621  * T;     (* Jupiter *)
    M4 := 225.4445943 +   3034.9066206 * T;     (* 444539 ? *)
    U4 := 138.419219  +   3034.906621  * T;
    L5 := 265.869357  +   1222.116843  * T;     (* Saturne *)
    M5 := 175.758477  +   1222.116843  * T;
    U5 := 153.521637  +   1222.116843  * T;
    L6 := 243.362437  +    429.898403  * T;     (* Uranus *)
    M6 :=  74.313637  +    429.898403  * T;
    U6 := 169.872293  +    429.388747  * T;
    L7 :=  85.024943  +    219.863377  * T;     (* Neptune *)
    M7 :=  41.269103  +    219.863377  * T;
    U7 := 314.346275  +    218.761885  * T;
    L8 :=  92.312712  +    146.674728  * T;     (* Pluton *)
    M8 := 229.488633  +    145.278567  * T;
    U8 := 343.369233  +    145.278567  * T;

    L9 := 270.435378  + 481267.880862  * T;     (* Lune *)
    M9 := 296.095334  + 477198.867586  * T;
    OmegaLune := 259.181303 - 1934.137823 * T;  (* pour phmrides Lune *)

    DegreesToRadians (L0);
    DegreesToRadians (M0);
    DegreesToRadians (L1);
    DegreesToRadians (M1);
    DegreesToRadians (U1);
    DegreesToRadians (L2);
    DegreesToRadians (M2);
    DegreesToRadians (U2);
    DegreesToRadians (L3);
    DegreesToRadians (M3);
    DegreesToRadians (U3);
    DegreesToRadians (L4);
    DegreesToRadians (M4);
    DegreesToRadians (U4);
    DegreesToRadians (L5);
    DegreesToRadians (M5);
    DegreesToRadians (U5);
    DegreesToRadians (L6);
    DegreesToRadians (M6);
    DegreesToRadians (U6);
    DegreesToRadians (L7);
    DegreesToRadians (M7);
    DegreesToRadians (U7);
    DegreesToRadians (L8);
    DegreesToRadians (M8);
    DegreesToRadians (U8);
    DegreesToRadians (L9);
    DegreesToRadians (M9);
    DegreesToRadians (OmegaLune);

    Epsilon   := 23.4522944 - 0.0130125 * T - 0.000001638 * T2;
    DegreesToRadians (Epsilon);

(*
    --- MUST be computed right after CalcSetup and BEFORE CalcSoleil
    --- deltaPSI is too small to consider here for apparent TSL
    --- WAS a separate procedure
*)

    OmegaNutation := 259.18327 - 1934.142 * T;
    DegreesToRadians (OmegaNutation);

    NutationLambda  := - 0.004722 * Sin(OmegaNutation);
    NutationEpsilon := 0.00256 * Cos(OmegaNutation);
    NutationSoleil  := - 0.00569 - 0.00479 * Sin(OmegaNutation);

    DegreesToRadians (NutationLambda);
    DegreesToRadians (NutationEpsilon);
    DegreesToRadians (NutationSoleil);

    Epsilon := Epsilon + NutationEpsilon; (* !!! *)

END CalcSetup;

(* === COORDINATES ============================================ *)

PROCEDURE CalcAlphaDelta (); (* we use global variables here *)
VAR
    numerator,divisor : LONGREAL;
BEGIN
    numerator := Sin(lambda) * Cos(Epsilon) - Tan(beta) * Sin(Epsilon);
    divisor   := Cos(lambda);
    alpha     := ATan2 (divisor,numerator); (* REVERSE LOGIC here !!! *)
    delta := Sin(beta) * Cos(Epsilon) + Cos(beta) * Sin(Epsilon) * Sin(lambda);
    delta := ASin(delta);
END CalcAlphaDelta;

(* === COMMON TO PLANETS ====================================== *)

VAR
    Mode : calcmode; (* partial=astrology, full=astronomy *)

PROCEDURE Store (VAR P : skypos);
BEGIN
    RadiansToDegrees(lambda);
    Mod360 (lambda);
    P.lambda := lambda;
    CASE Mode OF
    | partial :
         P.beta := 0.0;
         P.alpha:= 0.0;
         P.delta:= 0.0;
         P.l    := 0.0;
         P.b    := 0.0;
         P.r    := 0.0;
         P.azimut  := 0.0;
         P.altitude:= 0.0;
         P.domitude:= 0.0;
         P.planetaryhour := 0.0;
         P.truehouse:= 0;
         P.RETrank  := 0;
         P.RETzone  := 0;
    | full :
         RadiansToDegrees(beta);
         RadiansToDegrees(alpha);
         RadiansToDegrees(delta);
         Mod360 (alpha);
         P.beta  := beta;
         P.alpha := alpha;
         P.delta := delta;

         RadiansToDegrees(l);
         RadiansToDegrees(b);
         Mod360(l);
         P.l := l;
         P.b := b;
         P.r := r;
         P.azimut  := 0.0;
         P.altitude:= 0.0;
         P.domitude:= 0.0;
         P.planetaryhour := 0.0;
         P.truehouse:= 0;
         P.RETrank := 0;
         P.RETzone := 0;
    ELSE ; (* //ISO fixed silly requirement *) END;
END Store;

(* === COMMON FOR PLANETS ===================================== *)

PROCEDURE CalcPlanete (VAR P : skypos);
VAR
    x,y,z    : LONGREAL;
    bigdelta : LONGREAL;
BEGIN
    x := X + r * Cos(b) * Cos(l);
    y := Y + r * Cos(b) * Sin(l);
    lambda := ATan2 (x,y); (* arc tangente de Y sur X !!! *)
    lambda := lambda + NutationLambda; (* !!! *)
    IF Mode = full THEN
        z := r * Sin(b);
        bigdelta := Sqrt (x * x + y * y + z * z);
        beta := ASin (z / bigdelta);
        CalcAlphaDelta ();
    END;
    Store (P);
END CalcPlanete;

(* === PLANETS ================================================ *)

(*
  pour chaque plante, calcul des lments l, b, r
  (rapports  l'quinoxe moyen de la date)
*)

(* L1 M1 M2 U1 *)
PROCEDURE CalcMercure (VAR P : skypos);
BEGIN
    l := 8.0 * Sin(M1) * T;
    l := l + 84378.0 * Sin(M1);
    l := l + 10733.0 * Sin(2.0 * M1);
    l := l +  1892.0 * Sin(3.0 * M1);
    l := l +   381.0 * Sin(4.0 * M1);
    l := l +    83.0 * Sin(5.0 * M1);
    l := l +    19.0 * Sin(6.0 * M1);
    l := l -   646.0 * Sin(2.0 * U1);
    l := l -   306.0 * Sin(M1 - 2.0 * U1);
    l := l -   274.0 * Sin(M1 + 2.0 * U1);
    l := l -    92.0 * Sin(2.0 * M1 + 2.0 * U1);
    l := l -    28.0 * Sin(3.0 * M1 + 2.0 * U1);
    l := l +    25.0 * Sin(2.0 * M1 - 2.0 * U1);
    l := l -     9.0 * Sin(4.0 * M1 + 2.0 * U1);
    l := l +     7.0 * Cos(2.0 * M1 - 5.0 * M2);
    SecondsToRadians (l);
    l := L1 + l;
    b := 24134.0 * Sin(U1);
    b := b -   10.0 * Sin(3.0 * U1);
    b := b + 5180.0 * Sin(M1 - U1);
    b := b + 4910.0 * Sin(M1 + U1);
    b := b + 1124.0 * Sin(2.0 * M1 + U1);
    b := b +  271.0 * Sin(3.0 * M1 + U1);
    b := b +  132.0 * Sin(2.0 * M1 - U1);
    b := b +   67.0 * Sin(4.0 * M1 + U1);
    b := b +   18.0 * Sin(3.0 * M1 - U1);
    b := b +   17.0 * Sin(5.0 * M1 + U1);
    b := b -    9.0 * Sin(M1 - 3.0 * U1);
    SecondsToRadians (b);
    r := 0.39528 - 0.07834 * Cos(M1);
    r := r - 0.00795 * Cos(2.0 * M1);
    r := r - 0.00121 * Cos(3.0 * M1);
    r := r - 0.00022 * Cos(4.0 * M1);
    CalcPlanete (P);
END CalcMercure;

(* L2 M2 U2 M0 *)
PROCEDURE CalcVenus (VAR P : skypos);
BEGIN
    l := -20.0 * Sin(M2) * T;
    l := l + 2814.0 * Sin(M2);
    l := l +   12.0 * Sin(2.0 * M2);
    l := l -  181.0 * Sin(2.0 * U2);
    l := l -   10.0 * Cos(2.0 * M0 - 2.0 * M2);
    l := l +    7.0 * Cos(3.0 * M0 - 3.0 * M2);
    SecondsToRadians (l);
    l := L2 + l;
    b := 12215.0 * Sin(U2);
    b := b + 83.0 * Sin(M2 + U2);
    b := b + 83.0 * Sin(M2 - U2);
    SecondsToRadians (b);
    r := 0.72335 - 0.00493 * Cos(M2);
    CalcPlanete (P);
END CalcVenus;

(* L3 M0 M2 M3 M4 U3 *)
PROCEDURE CalcMars (VAR P : skypos);
BEGIN
    l := 37.0 * Sin(M3) * T;
    l := l +     4.0 * Sin(2.0 * M3) * T;
    l := l + 38451.0 * Sin(M3);
    l := l +  2238.0 * Sin(2.0 * M3);
    l := l +   181.0 * Sin(3.0 * M3);
    l := l +    17.0 * Sin(4.0 * M3);
    l := l -    52.0 * Sin(2.0 * U3);
    l := l -    22.0 * Cos(M3 - 2.0 * M4);
    l := l -    19.0 * Sin(M3 - M4);
    l := l +    17.0 * Cos(M3 - M4);
    l := l -    16.0 * Cos(2.0 * M3 - 2.0 * M4);
    l := l +    13.0 * Cos(M0 - 2.0 * M3);
    l := l -    10.0 * Sin(M3 - 2.0 * U3);
    l := l -    10.0 * Sin(M3 + 2.0 * U3);
    l := l +     7.0 * Cos(M0 - M3);
    l := l -     7.0 * Cos(2.0 * M0 - 3.0 * M3);
    l := l -     5.0 * Sin(M2 - 3.0 * M3);
    l := l -     5.0 * Sin(M0 - M3);
    l := l -     5.0 * Sin(M0 - 2.0 * M3);
    l := l -     4.0 * Cos(2.0 * M0 - 4.0 * M3);
    l := l +     4.0 * Cos(M4);
    l := l +     3.0 * Cos(M2 - 3.0 * M3);
    l := l +     3.0 * Sin(2.0 * M3 - 2.0 * M4);
    SecondsToRadians (l);
    l := L3 + l;
    b := 6603.0 * Sin(U3);
    b := b + 622.0 * Sin(M3 - U3);
    b := b + 615.0 * Sin(M3 + U3);
    b := b +  64.0 * Sin(2.0 * M3 + U3);
    SecondsToRadians (b);
    r := 1.53031 - 0.1417 * Cos(M3);
    r := r - 0.0066 * Cos(2.0 * M3);
    r := r - 0.00047 * Cos(3.0 * M3);
    CalcPlanete (P);
END CalcMars;

(* L4 M4 M5 M6 *)
PROCEDURE CalcJupiter (VAR P : skypos);
BEGIN
    l := 2511.0 + 5023.0 * T;
    l := l + 19934.0 * Sin(M4);
    l := l +   601.0 * Sin(2.0 * M4);
    l := l +  1093.0 * Cos(2.0 * M4 - 5.0 * M5);
    l := l -   479.0 * Sin(2.0 * M4 - 5.0 * M5);
    l := l -   185.0 * Sin(2.0 * M4 - 2.0 * M5);
    l := l +   137.0 * Sin(3.0 * M4 - 5.0 * M5);
    l := l -   131.0 * Sin(M4 - 2.0 * M5);
    l := l +    79.0 * Cos(M4 - M5);
    l := l -    76.0 * Cos(2.0 * M4 - 2.0 * M5);
    l := l -    74.0 * Cos(M4) * T;
    l := l +    68.0 * Sin(M4) * T;
    l := l +    66.0 * Cos(2.0 * M4 - 3.0 * M5);
    l := l +    63.0 * Cos(3.0 * M4 - 5.0 * M5);
    l := l +    53.0 * Cos(M4 - 5.0 * M5);
    l := l +    49.0 * Sin(2.0 * M4 - 3.0 * M5);
    l := l -    43.0 * Sin(2.0 * M4 - 5.0 * M5) * T;
    l := l -    37.0 * Cos(M4);
    l := l +    25.0 * Sin(2.0 * L4);
    l := l +    25.0 * Sin(3.0 * M4);
    l := l -    23.0 * Sin(M4 - 5.0 * M5);
    l := l -    19.0 * Cos(2.0 * M4 - 5.0 * M5) * T;
    l := l +    17.0 * Cos(2.0 * M4 - 4.0 * M5);
    l := l +    17.0 * Cos(3.0 * M4 - 3.0 * M5);
    l := l -    14.0 * Sin(M4 - M5);
    l := l -    13.0 * Sin(3.0 * M4 - 4.0 * M5);
    l := l -     9.0 * Cos(2.0 * L4);
    l := l +     9.0 * Cos(M5);
    l := l -     9.0 * Sin(M5);
    l := l -     9.0 * Sin(3.0 * M4 - 2.0 * M5);
    l := l +     9.0 * Sin(4.0 * M4 - 5.0 * M5);
    l := l +     9.0 * Sin(2.0 * M4 - 6.0 * M5 + 3.0 * M6);
    l := l -     8.0 * Cos(4.0 * M4 - 10.0 * M5);
    l := l +     7.0 * Cos(3.0 * M4 - 4.0 * M5);
    l := l -     7.0 * Cos(M4 - 3.0 * M5);
    l := l -     7.0 * Sin(4.0 * M4 - 10.0 * M5);
    l := l -     7.0 * Sin(M4 - 3.0 * M5);
    l := l +     6.0 * Cos(4.0 * M4 - 5.0 * M5);
    l := l -     6.0 * Sin(3.0 * M4 - 3.0 * M5);
    l := l +     5.0 * Cos(2.0 * M5);
    l := l -     4.0 * Sin(4.0 * M4 - 4.0 * M5);
    l := l -     4.0 * Cos(3.0 * M5);
    l := l +     4.0 * Cos(2.0 * M4 - M5);
    l := l -     4.0 * Cos(3.0 * M4 - 2.0 * M5);
    l := l -     4.0 * Cos(2.0 * M4) * T;
    l := l +     3.0 * Sin(2.0 * M4) * T;
    l := l +     3.0 * Cos(5.0 * M5);
    l := l +     3.0 * Cos(5.0 * M4 - 10.0 * M5);
    l := l +     3.0 * Sin(2.0 * M5);
    l := l -     2.0 * Sin(2.0 * L4 - M4);
    l := l +     2.0 * Sin(2.0 * L4 + M4);
    l := l -     2.0 * Sin(3.0 * M4 - 5.0 * M5) * T;
    l := l -     2.0 * Sin(M4 - 5.0 * M5) * T;
    SecondsToRadians (l);
    l := L4 + l;
    b := -4692.0 * Cos(M4);
    b := b + 259.0 * Sin(M4);
    b := b + 227.0;
    b := b - 227.0 * Cos(2.0 * M4);
    b := b +  30.0 * Sin(M4) * T;
    b := b +  21.0 * Cos(M4) * T;
    b := b +  16.0 * Sin(3.0 * M4 - 5.0 * M5);
    b := b -  13.0 * Sin(M4 - 5.0 * M5);
    b := b -  12.0 * Cos(3.0 * M4);
    b := b +  12.0 * Sin(2.0 * M4);
    b := b +   7.0 * Cos(3.0 * M4 - 5.0 * M5);
    b := b -   5.0 * Cos(M4 - 5.0 * M5);
    SecondsToRadians (b);
    r := 5.20883 - 0.25122 * Cos(M4);
    r := r - 0.00604 * Cos(2.0 * M4);
    r := r + 0.0026  * Cos(2.0 * M4 - 2.0 * M5);
    r := r - 0.0017  * Cos(3.0 * M4 - 5.0 * M5);
    r := r - 0.0016  * Sin(2.0 * M4 - 2.0 * M5);
    r := r - 0.00091 * Sin(M4) * T;
    r := r - 0.00084 * Cos(M4) * T;
    r := r + 0.00069 * Sin(2.0 * M4 - 3.0 * M5);
    r := r - 0.00067 * Sin(M4 - 5.0 * M5);
    r := r + 0.00066 * Sin(3.0 * M4 - 5.0 * M5);
    r := r + 0.00063 * Sin(M4 - M5);
    r := r - 0.00051 * Cos(2.0 * M4 - 3.0 * M5);
    r := r - 0.00046 * Sin(M4);
    r := r - 0.00029 * Cos(M4 - 5.0 * M5);
    r := r + 0.00027 * Cos(M4 - 2.0 * M5);
    r := r - 0.00022 * Cos(3.0 * M4);
    r := r - 0.00021 * Sin(2.0 * M4 - 5.0 * M5);
    CalcPlanete (P);
END CalcJupiter;

(* M4 M5 M6 L5 *)
PROCEDURE CalcSaturne (VAR P : skypos);
BEGIN
    l := 2507.0 + 5014.0 * T;
    l := l + 23043.0 * Sin(M5);
    l := l -  2689.0 * Cos(2.0 * M4 - 5.0 * M5);
    l := l +  1177.0 * Sin(2.0 * M4 - 5.0 * M5);
    l := l -   826.0 * Cos(2.0 * M4 - 4.0 * M5);
    l := l +   802.0 * Sin(2.0 * M5);
    l := l +   425.0 * Sin(M4 - 2.0 * M5);
    l := l -   229.0 * Cos(M5) * T;
    l := l -   142.0 * Sin(M5) * T;
    l := l -   153.0 * Cos(2.0 * M4 - 6.0 * M5);
    l := l -   114.0 * Cos(M5);
    l := l +   101.0 * Sin(2.0 * M4 - 5.0 * M5) * T;
    l := l -    70.0 * Cos(2.0 * L5);
    l := l +    67.0 * Sin(2.0 * L5);
    l := l +    66.0 * Sin(2.0 * M4 - 6.0 * M5);
    l := l +    60.0 * Cos(2.0 * M4 - 5.0 * M5) * T;
    l := l +    41.0 * Sin(M4 - 3.0 * M5);
    l := l +    39.0 * Sin(3.0 * M5);
    l := l +    31.0 * Sin(M4 - M5);
    l := l +    31.0 * Sin(2.0 * M4 - 2.0 * M5);
    l := l -    29.0 * Cos(2.0 * M4 - 3.0 * M5);
    l := l -    28.0 * Sin(2.0 * M4 - 6.0 * M5 + 3.0 * M6);
    l := l +    28.0 * Cos(M4 - 3.0 * M5);
    l := l +    22.0 * Sin(2.0 * M4 - 4.0 * M5) * T;
    l := l -    22.0 * Sin(M5 - 3.0 * M6);
    l := l +    20.0 * Sin(2.0 * M4 - 3.0 * M5);
    l := l +    20.0 * Cos(4.0 * M4 - 10.0 * M5);
    l := l +    19.0 * Cos(2.0 * M5 - 3.0 * M6);
    l := l +    19.0 * Sin(4.0 * M4 - 10.0 * M5);
    l := l -    17.0 * Cos(2.0 * M5) * T;
    l := l -    16.0 * Cos(M5 - 3.0 * M6);
    l := l -    12.0 * Sin(2.0 * M4 - 4.0 * M5);
    l := l +    12.0 * Cos(M4);
    l := l -    12.0 * Sin(2.0 * M5 - 2.0 * M6);
    l := l -    11.0 * Sin(2.0 * M5) * T;
    l := l -    11.0 * Cos(2.0 * M4 - 7.0 * M5);
    l := l +    10.0 * Sin(2.0 * M5 - 3.0 * M6);
    l := l +    10.0 * Cos(2.0 * M4 - 2.0 * M5);
    l := l +     9.0 * Sin(4.0 * M4 - 9.0 * M5);
    l := l -     8.0 * Sin(M5 - 2.0 * M6);
    l := l -     8.0 * Cos(2.0 * L5 + M5);
    l := l +     8.0 * Cos(2.0 * L5 - M5);
    l := l +     8.0 * Cos(M5 - M6);
    l := l -     8.0 * Sin(2.0 * L5 - M5);
    l := l +     7.0 * Sin(2.0 * L5 + M5);
    l := l -     7.0 * Cos(M4 - 2.0 * M5);
    l := l -     7.0 * Cos(2.0 * M5);
    l := l -     6.0 * Sin(4.0 * M4 - 10.0 * M5) * T;
    l := l +     6.0 * Cos(4.0 * M4 - 10.0 * M5) * T;
    l := l +     6.0 * Sin(2.0 * M4 - 6.0 * M5) * T;
    l := l -     5.0 * Sin(3.0 * M4 - 7.0 * M5);
    l := l -     5.0 * Cos(3.0 * M4 - 3.0 * M5);
    l := l -     5.0 * Cos(2.0 * M5 - 2.0 * M6);
    l := l +     5.0 * Sin(3.0 * M4 - 4.0 * M5);
    l := l +     5.0 * Sin(2.0 * M4 - 7.0 * M5);
    l := l +     4.0 * Sin(3.0 * M4 - 3.0 * M5);
    l := l +     4.0 * Sin(3.0 * M4 - 5.0 * M5);
    l := l +     4.0 * Cos(M4 - 3.0 * M5) * T;
    l := l +     3.0 * Cos(2.0 * M4 - 4.0 * M5) * T;
    l := l +     3.0 * Cos(2.0 * M4 - 6.0 * M5 + 3.0 * M6);
    l := l -     3.0 * Sin(2.0 * L5) * T;
    l := l +     3.0 * Cos(2.0 * M4 - 6.0 * M5) * T;
    l := l -     3.0 * Cos(2.0 * L5) * T;
    l := l +     3.0 * Cos(3.0 * M4 - 7.0 * M5);
    l := l +     3.0 * Cos(4.0 * M4 - 9.0 * M5);
    l := l +     3.0 * Sin(3.0 * M4 - 6.0 * M5);
    l := l +     3.0 * Sin(2.0 * M4 - M5);
    l := l +     3.0 * Sin(M4 - 4.0 * M5);
    l := l +     2.0 * Cos(3.0 * M5 - 3.0 * M6);
    l := l +     2.0 * Sin(M4 - 2.0 * M5) * T;
    l := l +     2.0 * Sin(4.0 * M5);
    l := l -     2.0 * Cos(3.0 * M4 - 4.0 * M5);
    l := l -     2.0 * Cos(2.0 * M4 - M5);
    l := l -     2.0 * Sin(2.0 * M4 - 7.0 * M5 + 3.0 * M6);
    l := l +     2.0 * Cos(M4 - 4.0 * M5);
    l := l +     2.0 * Cos(4.0 * M4 - 11.0 * M5);
    l := l -     2.0 * Sin(M5 - M6);
    SecondsToRadians (l);
    l := L5 + l;
    b := 185.0 + 8297.0 * Sin(M5);
    b := b - 3346.0 * Cos(M5);
    b := b +  462.0 * Sin(2.0 * M5);
    b := b -  189.0 * Cos(2.0 * M5);
    b := b +   79.0 * Cos(M5) * T;
    b := b -   71.0 * Cos(2.0 * M4 - 4.0 * M5);
    b := b +   46.0 * Sin(2.0 * M4 - 6.0 * M5);
    b := b -   45.0 * Cos(2.0 * M4 - 6.0 * M5);
    b := b +   29.0 * Sin(3.0 * M5);
    b := b -   20.0 * Cos(2.0 * M4 - 3.0 * M5);
    b := b +   18.0 * Sin(M5) * T;
    b := b -   14.0 * Cos(2.0 * M4 - 5.0 * M5);
    b := b -   11.0 * Cos(3.0 * M5);
    b := b -   10.0 * T;
    b := b +    9.0 * Sin(M4 - 3.0 * M5);
    b := b +    8.0 * Sin(M4 - M5);
    b := b -    6.0 * Sin(2.0 * M4 - 3.0 * M5);
    b := b +    5.0 * Sin(2.0 * M4 - 7.0 * M5);
    b := b -    5.0 * Cos(2.0 * M4 - 7.0 * M5);
    b := b +    4.0 * Sin(2.0 * M4 - 5.0 * M5);
    b := b -    4.0 * Sin(2.0 * M5) * T;
    b := b -    3.0 * Cos(M4 - M5);
    b := b +    3.0 * Cos(M4 - 3.0 * M5);
    b := b +    3.0 * Sin(2.0 * M4 - 4.0 * M5) * T;
    b := b +    3.0 * Sin(M4 - 2.0 * M5);
    b := b +    2.0 * Sin(4.0 * M5);
    b := b -    2.0 * Cos(2.0 * M4 - 2.0 * M5);
    SecondsToRadians (b);
    r := 9.55774 - 0.00028 * T;
    r := r - 0.53252 * Cos(M5);
    r := r - 0.01878 * Sin(2.0 * M4 - 4.0 * M5);
    r := r - 0.01482 * Cos(2.0 * M5);
    r := r + 0.00817 * Sin(M4 - M5);
    r := r - 0.00539 * Cos(M4 - 2.0 * M5);
    r := r - 0.00524 * Sin(M5) * T;
    r := r + 0.00349 * Sin(2.0 * M4 - 5.0 * M5);
    r := r + 0.00347 * Sin(2.0 * M4 - 6.0 * M5);
    r := r + 0.00328 * Cos(M5) * T;
    r := r - 0.00225 * Sin(M5);
    r := r + 0.00149 * Cos(2.0 * M4 - 6.0 * M5);
    r := r - 0.00126 * Cos(2.0 * M4 - 2.0 * M5);
    r := r + 0.00104 * Cos(M4 - M5);
    r := r + 0.00101 * Cos(2.0 * M4 - 5.0 * M5);
    r := r + 0.00098 * Cos(M4 - 3.0 * M5);
    r := r - 0.00073 * Cos(2.0 * M4 - 3.0 * M5);
    r := r - 0.00062 * Cos(3.0 * M5);
    r := r + 0.00042 * Sin(2.0 * M5 - 3.0 * M6);
    r := r + 0.00041 * Sin(2.0 * M4 - 2.0 * M5);
    r := r - 0.0004  * Sin(M4 - 3.0 * M5);
    r := r + 0.0004  * Cos(2.0 * M4 - 4.0 * M5);
    r := r - 0.00023 * Sin(M4);
    r := r + 0.0002  * Sin(2.0 * M4 - 7.0 * M5);
    CalcPlanete (P);
END CalcSaturne;

(* L6 M4 M5 M6 M7 U6 T2 *)
PROCEDURE CalcUranus (VAR P : skypos);
BEGIN
    l := 32.0 * T2;
    l := l + 19397.0 * Sin(M6);
    l := l +   570.0 * Sin(2.0 * M6);
    l := l -   536.0 * Cos(M6) * T;
    l := l +   143.0 * Sin(M5 - 2.0 * M6);
    l := l +   110.0 * Sin(M6) * T;
    l := l +   102.0 * Sin(M5 - 3.0 * M6);
    l := l +    76.0 * Cos(M5 - 3.0 * M6);
    l := l -    49.0 * Sin(M4 - M6);
    l := l -    30.0 * Cos(2.0 * M6) * T;
    l := l +    29.0 * Sin(2.0 * M4 - 6.0 * M5 + 3.0 * M6);
    l := l +    29.0 * Cos(2.0 * M6 - 2.0 * M7);
    l := l -    28.0 * Cos(M6 - M7);
    l := l +    23.0 * Sin(3.0 * M6);
    l := l -    21.0 * Cos(M4 - M6);
    l := l +    20.0 * Sin(M6 - M7);
    l := l +    20.0 * Cos(M5 - 2.0 * M6);
    l := l -    19.0 * Cos(M5 - M6);
    l := l +    17.0 * Sin(2.0 * M6 - 3.0 * M3);
    l := l +    14.0 * Sin(3.0 * M6 - 3.0 * M7);
    l := l +    13.0 * Sin(M5 - M6);
    l := l -    12.0 * Cos(M6) * T2;
    l := l -    12.0 * Cos(M6);
    l := l +    10.0 * Sin(2.0 * M6 - 2.0 * M7);
    l := l -     9.0 * Sin(2.0 * U6);
    l := l -     9.0 * Sin(M6) * T2;
    l := l +     9.0 * Cos(2.0 * M6 - 3.0 * M7);
    l := l +     8.0 * Cos(M5 - 2.0 * M6) * T;
    l := l +     7.0 * Cos(M5 - 3.0 * M6) * T;
    l := l -     7.0 * Sin(M5 - 3.0 * M6) * T;
    l := l +     7.0 * Sin(2.0 * M6) * T;
    l := l +     6.0 * Sin(2.0 * M4 - 6.0 * M5 + 2.0 * M6);
    l := l +     6.0 * Cos(2.0 * M4 - 6.0 * M5 + 2.0 * M6);
    l := l +     5.0 * Sin(M5 - 4.0 * M6);
    l := l -     4.0 * Sin(3.0 * M6 - 4.0 * M7);
    l := l +     4.0 * Cos(3.0 * M6 - 3.0 * M7);
    l := l -     3.0 * Cos(M7);
    l := l -     2.0 * Sin(M7);
    SecondsToRadians (l);
    l := L6 + l;
    b := 2775.0 * Sin(U6);
    b := b + 131.0 * Sin(M6 - U6);
    b := b + 130.0 * Sin(M6 + U6);
    SecondsToRadians (b);
    r := 19.21216 - 0.90154 * Cos(M6);
    r := r - 0.02488 * Sin(M6) * T;
    r := r - 0.02121 * Cos(2.0 * M6);
    r := r - 0.00585 * Cos(M5 - 2.0 * M6);
    r := r - 0.00508 * Cos(M6) * T;
    r := r - 0.00451 * Cos(M4 - M6);
    r := r + 0.00336 * Sin(M5 - M6);
    r := r + 0.00198 * Sin(M4 - M6);
    r := r + 0.00118 * Cos(M5 - 3.0 * M6);
    r := r + 0.00107 * Sin(M5 - 2.0 * M6);
    r := r - 0.00103 * Sin(2.0 * M6) * T;
    r := r - 0.00081 * Cos(3.0 * M6 - 3.0 * M7);
    CalcPlanete (P);
END CalcUranus;

(* L6 L7 M4 M5 M6 M7 U7 T2 *)
PROCEDURE CalcNeptune (VAR P : skypos);
BEGIN
    l := 3523.0 * Sin(M7);
    l := l - 50.0 * Sin(2.0 * U7);
    l := l - 43.0 * Cos(M7) * T;
    l := l + 29.0 * Sin(M4 - M7);
    l := l + 19.0 * Sin(2.0 * M7);
    l := l - 18.0 * Cos(M4 - M7);
    l := l + 13.0 * Cos(M5 - M7);
    l := l + 13.0 * Sin(M5 - M7);
    l := l -  9.0 * Sin(2.0 * M6 - 3.0 * M7);
    l := l +  9.0 * Cos(2.0 * M6 - 2.0 * M7);
    l := l -  5.0 * Cos(2.0 * M6 - 3.0 * M7);
    l := l -  4.0 * Sin(M7) * T;
    l := l +  4.0 * Cos(M6 - 2.0 * M7);
    l := l +  4.0 * Sin(M7) * T2;
    SecondsToRadians (l);
    l := L7 + l;
    b := 6404.0 * Sin(U7);
    b := b + 55.0 * Sin(M7 + U7);
    b := b + 55.0 * Sin(M7 - U7);
    b := b - 33.0 * Sin(U7) * T;
    SecondsToRadians (b);
    r := 30.07175 - 0.25701 * Cos(M7);
    r := r - 0.00787 * Cos(2.0 * L6 - M6 - 2.0 * L7);
    r := r + 0.00409 * Cos(M4 - M7);
    r := r - 0.00314 * Sin(M7) * T;
    r := r + 0.0025  * Sin(M4 - M7);
    r := r - 0.00194 * Sin(M5 - M7);
    r := r + 0.00185 * Cos(M5 - M7);
    CalcPlanete (P);
END CalcNeptune;

(* L8 M8 U8 T2 *)
PROCEDURE CalcPluton (VAR P : skypos);
BEGIN
    l := 101577.0 * Sin(M8);
    l := l + 15517.0 * Sin(2.0 * M8);
    l := l -  3593.0 * Sin(2.0 * U8);
    l := l +  3414.0 * Sin(3.0 * M8);
    l := l -  2201.0 * Sin(M8 - 2.0 * U8);
    l := l -  1871.0 * Sin(M8 + 2.0 * U8);
    l := l +   839.0 * Sin(4.0 * M8);
    l := l -   757.0 * Sin(2.0 * M8 + 2.0 * U8);
    l := l -   285.0 * Sin(3.0 * M8 + 2.0 * U8);
    l := l +   227.0 * Sin(M8) * T2;
    l := l +   218.0 * Sin(2.0 * M8 - 2.0 * U8);
    l := l +   200.0 * Sin(M8) * T;
    SecondsToRadians (l);
    l := L8 + l;
    b := 57726.0 * Sin(U8);
    b := b + 15257.0 * Sin(M8 - U8);
    b := b + 14102.0 * Sin(M8 + U8);
    b := b +  3870.0 * Sin(2.0 * M8 + U8);
    b := b +  1138.0 * Sin(3.0 * M8 + U8);
    b := b +   472.0 * Sin(2.0 * M8 - U8);
    b := b +   353.0 * Sin(4.0 * M8 + U8);
    b := b -   144.0 * Sin(M8 - 3.0 * U8);
    b := b -   119.0 * Sin(3.0 * U8);
    b := b -   111.0 * Sin(M8 + 3.0 * U8);
    SecondsToRadians (b);
    r := 40.74638 - 9.58235 * Cos(M8);
    r := r - 1.16703 * Cos(2.0 * M8);
    r := r - 0.22649 * Cos(3.0 * M8);
    r := r - 0.04996 * Cos(4.0 * M8);
    CalcPlanete (P);
END CalcPluton;

(* === MOON =================================================== *)

PROCEDURE CalcLune (VAR P : skypos);
VAR
  F : LONGREAL; (* distance angulaire de la Lune moyenne au noeud ascendant *)
  D : LONGREAL; (* longation moyenne *)
  M : LONGREAL; (* alias for M9 in original *)
BEGIN
    F :=  11.254075 + 483202.018685 * T;
    D := 350.738975 + 445267.111345 * T;
    DegreesToRadians (F);
    DegreesToRadians (D);
    M := M9; (* suis flemmard, l... *)
    lambda := 22640.0 * Sin(M);
    lambda := lambda - 4586.0 * Sin(M - 2.0 * D);
    lambda := lambda + 2370.0 * Sin(2.0 * D);
    lambda := lambda +  769.0 * Sin(2.0 * M);
    lambda := lambda -  668.0 * Sin(M0);
    lambda := lambda -  412.0 * Sin(2.0 * F);
    lambda := lambda -  212.0 * Sin(2.0 * M - 2.0 * D);
    lambda := lambda -  206.0 * Sin(M - 2.0 * D + M0);
    lambda := lambda +  192.0 * Sin(M + 2.0 * D);
    lambda := lambda +  165.0 * Sin(2.0 * D - M0);
    lambda := lambda +  148.0 * Sin(M - M0);
    lambda := lambda -  125.0 * Sin(D);
    lambda := lambda -  110.0 * Sin(M + M0);
    lambda := lambda -   55.0 * Sin(2.0 * F - 2.0 * D);
    lambda := lambda -   45.0 * Sin(M + 2.0 * F);
    lambda := lambda +   40.0 * Sin(M - 2.0 * F);
    lambda := lambda -   38.0 * Sin(M - 4.0 * D);
    lambda := lambda +   36.0 * Sin(3.0 * M);
    lambda := lambda -   31.0 * Sin(2.0 * M - 4.0 * D);
    lambda := lambda +   28.0 * Sin(M - 2.0 * D - M0);
    lambda := lambda -   24.0 * Sin(2.0 * D + M0);
    lambda := lambda +   19.0 * Sin(M - D);
    lambda := lambda +   18.0 * Sin(D + M0);
    lambda := lambda +   15.0 * Sin(M + 2.0 * D - M0);
    lambda := lambda +   14.0 * Sin(2.0 * M + 2.0 * D);
    lambda := lambda +   14.0 * Sin(4.0 * D);
    lambda := lambda -   13.0 * Sin(3.0 * M - 2.0 * D);
    lambda := lambda -   11.0 * Sin(M + 16.0 * L0 - 18.0 * L2);
    lambda := lambda +   10.0 * Sin(2.0 * M - M0);
    lambda := lambda +    9.0 * Sin(M - 2.0 * F - 2.0 * D);
    lambda := lambda +    9.0 * Cos(M + 16.0 * L0 - 18.0 * L2);
    lambda := lambda -    9.0 * Sin(2.0 * M - 2.0 * D + M0);
    lambda := lambda -    8.0 * Sin(M + D);
    lambda := lambda +    8.0 * Sin(2.0 * D - 2.0 * M0);
    lambda := lambda -    8.0 * Sin(2.0 * M + M0);
    lambda := lambda -    7.0 * Sin(2.0 * M0);
    lambda := lambda -    7.0 * Sin(M - 2.0 * D + 2.0 * M0);
    lambda := lambda +    7.0 * Sin(OmegaLune);
    lambda := lambda -    6.0 * Sin(M - 2.0 * F + 2.0 * D);
    lambda := lambda -    6.0 * Sin(2.0 * F + 2.0 * D);
    lambda := lambda -    4.0 * Sin(M - 4.0 * D + M0);
    lambda := lambda +    4.0 * Cos(M + 16.0 * L0 - 18.0 * L2) * T;
    lambda := lambda -    4.0 * Sin(2.0 * M + 2.0 * F);
    lambda := lambda +    4.0 * Sin(M + 16.0 * L0 - 18.0 * L2) * T;
    lambda := lambda +    3.0 * Sin(M - 3.0 * D);
    lambda := lambda -    3.0 * Sin(M + 2.0 * D + M0);
    lambda := lambda -    3.0 * Sin(2.0 * M - 4.0 * D + M0);
    lambda := lambda +    3.0 * Sin(M - 2.0 * M0);
    lambda := lambda +    3.0 * Sin(M - 2.0 * D - 2.0 * M0);
    lambda := lambda -    2.0 * Sin(2.0 * M - 2.0 * D - M0);
    lambda := lambda -    2.0 * Sin(2.0 * F - 2.0 * D + M0);
    lambda := lambda +    2.0 * Sin(M + 4.0 * D);
    lambda := lambda +    2.0 * Sin(4.0 * M);
    lambda := lambda +    2.0 * Sin(4.0 * D - M0);
    lambda := lambda +    2.0 * Sin(2.0 * M - D);
    SecondsToRadians (lambda);
    lambda := L9 + lambda;
    lambda := lambda + NutationLambda; (* !!! *)
    IF Mode = full THEN
        beta := 18461.0 * Sin(F);
        beta := beta + 1010.0 * Sin(M + F);
        beta := beta + 1000.0 * Sin(M - F);
        beta := beta -  624.0 * Sin(F - 2.0 * D);
        beta := beta -  199.0 * Sin(M - F - 2.0 * D);
        beta := beta -  167.0 * Sin(M + F - 2.0 * D);
        beta := beta +  117.0 * Sin(F + 2.0 * D);
        beta := beta +   62.0 * Sin(2.0 * M + F);
        beta := beta +   33.0 * Sin(M - F + 2.0 * D);
        beta := beta +   32.0 * Sin(2.0 * M - F);
        beta := beta -   30.0 * Sin(F - 2.0 * D + M0);
        beta := beta -   16.0 * Sin(2.0 * M + F - 2.0 * D);
        beta := beta +   15.0 * Sin(M + F + 2.0 * D);
        beta := beta +   12.0 * Sin(F - 2.0 * D - M0);
        beta := beta -    9.0 * Sin(M - F - 2.0 * D + M0);
        beta := beta -    8.0 * Sin(F + OmegaLune);
        beta := beta +    8.0 * Sin(F + 2.0 * D - M0);
        beta := beta -    7.0 * Sin(M + F - 2.0 * D + M0);
        beta := beta +    7.0 * Sin(M + F - M0);
        beta := beta -    7.0 * Sin(M + F - 4.0 * D);
        beta := beta -    6.0 * Sin(F + M0);
        beta := beta -    6.0 * Sin(3.0 * F);
        beta := beta +    6.0 * Sin(M - F - M0);
        beta := beta -    5.0 * Sin(F + D);
        beta := beta -    5.0 * Sin(M + F + M0);
        beta := beta -    5.0 * Sin(M - F + M0);
        beta := beta +    5.0 * Sin(F - M0);
        beta := beta +    5.0 * Sin(F - D);
        beta := beta +    4.0 * Sin(3.0 * M + F);
        beta := beta -    4.0 * Sin(F - 4.0 * D);
        beta := beta -    3.0 * Sin(M - F - 4.0 * D);
        beta := beta +    3.0 * Sin(M - 3.0 * F);
        beta := beta -    2.0 * Sin(2.0 * M - F - 4.0 * D);
        beta := beta -    2.0 * Sin(3.0 * F - 2.0 * D);
        beta := beta +    2.0 * Sin(2.0 * M - F + 2.0 * D);
        beta := beta +    2.0 * Sin(M - F + 2.0 * D - M0);
        beta := beta +    2.0 * Sin(2.0 * M - F - 2.0 * D);
        beta := beta +    2.0 * Sin(3.0 * M - F);
        SecondsToRadians (beta);
        CalcAlphaDelta ();
    END;
    Store (P);
END CalcLune;

(* === SUN ==================================================== *)

(* v1.1 : attention ! faut toujours au moins faire le soleil *)
(* TOUJOURS A FAIRE MEME POUR UNE SEULE PLANETE !!! *)


(* L0 M0 M2 M3 M4 L9 T2 *)
PROCEDURE CalcSoleil (VAR P : skypos);
VAR
    e,e2,e3 : LONGREAL; (* eccentricit *)
BEGIN
    lambda := (1.919444 - 0.004722 * T) * Sin(M0);
    lambda := lambda + 0.02     * Sin(2.0 * M0);
    lambda := lambda - 0.001944 * Cos(M0 - M4);
    lambda := lambda + 0.001666 * Sin(L9 - L0);
    lambda := lambda + 0.001388 * Sin(4.0 * M0 - 8.0 * M3 + M4);
    lambda := lambda - 0.001388 * Cos(2.0 * M0 - 2.0 * M2);
    lambda := lambda - 0.001111 * Sin(M0 - M2);
    lambda := lambda + 0.001111 * Cos(4.0 * M0 - 8.0 * M3 + M4);
    lambda := lambda + 0.000833 * Sin(2.0 * M0 - 2.0 * M2);
    lambda := lambda - 0.000833 * Sin(M4);
    lambda := lambda - 0.000833 * Sin(2.0 * M0 - 2.0 * M4);
    DegreesToRadians (lambda);
    lambda := L0 + lambda;
    e  := 0.01675104 - 0.0000418 * T - 0.000000126 * T2;
    e2 := e * e;
    e3 := e * e2;
    r := 1.0 + e2 / 2.0;
    r := r - (e - e3 * 3.0 / 8.0) * Cos(M0);
    r := r - e2 / 2.0 * Cos(2.0 * M0);
    r := r - e3 * 3.0 / 8.0 * Cos(3.0 * M0);
    r := 1.0000002 * r;

(*  lambda := lambda + NutationSoleil; (* !!! *) moins prcis on dirait ! *)

    X := r * Cos(lambda); (* beta = 0 donc * Cos(beta) = * 1 ici *)
    Y := r * Sin(lambda);

    lambda := lambda + NutationSoleil; (* !!! *)

    IF Mode = full THEN
        beta   := 0.0; (* latitude du Soleil toujours infrieure  1"2 *)
        CalcAlphaDelta ();
    END;
    Store (P);
END CalcSoleil;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE zeroThem (  ); (* zero global vars *)
BEGIN
    beta     := 0.0;
    alpha    := 0.0;
    delta    := 0.0;
    l        := 0.0;
    b        := 0.0;
    r        := 0.0;
END zeroThem;

(*
   in all, NutationLambda is useless to correct things...
   all these nodes are MEAN nodes here !
   try and fix this, one of these days... LATER !
   nothing about that in ASTRONOMICAL papers !
*)

PROCEDURE CalcLilith (VAR P : skypos);
BEGIN
    zeroThem();
    lambda := L9 - M9 + PI;
    Store (P);
END CalcLilith;

PROCEDURE CalcDragonHead (VAR P : skypos);
BEGIN
    zeroThem();
    lambda := OmegaLune;
    Store (P);
END CalcDragonHead;

PROCEDURE CalcDragonTail (VAR P : skypos);
BEGIN
    zeroThem();
    lambda := OmegaLune + PI;
    Store (P);
END CalcDragonTail;

PROCEDURE CalcBlackSun (VAR P : skypos);
BEGIN
    zeroThem();
    lambda := L0 - M0;
    Store (P);
END CalcBlackSun;

(* === LOCAL SIDEREAL TIME ==================================== *)

CONST
    limitForPlacidus = LONGREAL(66.0); (* 90-epsilon *)
    limitForAll      = LONGREAL(89.0); (* avoid 90 ! *)

VAR
    placidusDomOK : BOOLEAN;
    allDomOK      : BOOLEAN;

PROCEDURE initDomFlags (phiDeg : LONGREAL);
BEGIN
    IF ABS(phiDeg) < limitForPlacidus THEN
        placidusDomOK := TRUE;
    ELSE
        placidusDomOK := FALSE;
    END;
    IF ABS(phiDeg) < limitForAll THEN
        allDomOK := TRUE;
    ELSE
        allDomOK := FALSE;
    END;
END initDomFlags;

VAR
    theLatitude : LONGREAL; (* for TSL AS MC *)
    TSL : LONGREAL; (* local sidereal time in radians used for AS and MC, A & h *)

PROCEDURE InitCalcLocal (jd,phiDeg,lambdaDeg:LONGREAL) : LONGREAL; (* return TSL *)
VAR
    JD0H   : LONGREAL;
    theLongitude : LONGREAL;
    tslDeg : LONGREAL;
    T0H, T0H2, T0H3 : LONGREAL;
BEGIN
    initDomFlags(phiDeg);
    theLatitude := phiDeg;
    theLongitude := lambdaDeg;
    DegreesToRadians (theLatitude);
    DegreesToRadians (theLongitude);
    JD0H := JDtoJD0H (jd);
    T0H  := (JD0H - 2415020.0) / 36525.0;
    T0H2 := T0H * T0H;
    T0H3 := T0H * T0H2;
    TSL := 0.276919398 + 100.0021359 * T0H + 0.000001075 * T0H2;
    TSL := MaxRadians * frac (TSL); (* TSL in radians here then *)
    TSL := TSL + 1.002737908 * MaxRadians * (jd - JD0H);
    TSL := TSL - theLongitude;
    tslDeg := TSL;
    RadiansToDegrees(tslDeg);
    Mod360(tslDeg);
    RETURN tslDeg; (* remember TSL is an ALPHA ! *)
END InitCalcLocal;

(* === LOCAL AS, MC, DS and FC ================================ *)

PROCEDURE CalcLambdaBetaSoleil ();
VAR
    numerator,divisor : LONGREAL;
BEGIN
    numerator := Sin (alpha);
    divisor   := Cos (alpha) * Cos (Epsilon);
    lambda    := ATan2 (divisor,numerator);
    beta      := 0.0;
END CalcLambdaBetaSoleil;

PROCEDURE CalcAlphaDeltaSoleil (); (* we use global variables here *)
VAR
    numerator,divisor : LONGREAL;
BEGIN
    numerator := Sin(lambda) * Cos(Epsilon);
    divisor   := Cos(lambda);
    alpha     := ATan2 (divisor,numerator); (* REVERSE LOGIC here !!! *)
    delta     := Sin(Epsilon) * Sin(lambda);
    delta     := ASin(delta);
END CalcAlphaDeltaSoleil;


PROCEDURE CalcAlphaDeltaSoleilSpecial (VAR P:skypos);
VAR
    numerator,divisor : LONGREAL;
BEGIN
    numerator := Sin(P.lambda) * Cos(Epsilon);
    divisor   := Cos(P.lambda);
    P.alpha   := ATan2 (divisor,numerator); (* REVERSE LOGIC here !!! *)
    P.delta   := Sin(Epsilon) * Sin(P.lambda);
    P.delta   := ASin(P.delta);
    RadiansToDegrees(P.alpha);
    RadiansToDegrees(P.delta);
    Mod360 (P.alpha);
END CalcAlphaDeltaSoleilSpecial;


(* ------------------------------------------------------------ *)

VAR
    ascendant,milieuduciel : LONGREAL; (* kept for later domification computation *)

PROCEDURE CalcAS (VAR P : skypos);
VAR
    numerator,divisor : LONGREAL;
BEGIN
    (* IF placidusDomOK = FALSE THEN RETURN; END; *)
    zeroThem();
    numerator := - Cos (TSL);
    divisor   := Sin(Epsilon) * Tan(theLatitude) + Cos(Epsilon) * Sin(TSL);
    lambda    := ATan2 (divisor,numerator); (* REVERSE LOGIC here !!! *)
    lambda    := lambda + PI;
    ascendant := lambda; (* for later use *)
    CalcAlphaDeltaSoleil ();
    Store (P);
END CalcAS;

PROCEDURE CalcMC (VAR P : skypos);
BEGIN
    (* IF placidusDomOK = FALSE THEN RETURN; END; *)
    zeroThem();
    alpha := TSL;
    CalcLambdaBetaSoleil ();
    milieuduciel := lambda; (* for later use *)
    CalcAlphaDeltaSoleil ();
    Store (P);
END CalcMC;

PROCEDURE CalcDS (VAR P : skypos);
VAR
    numerator,divisor : LONGREAL;
BEGIN
    (* IF placidusDomOK = FALSE THEN RETURN; END; *)
    zeroThem();
    numerator := - Cos (TSL);
    divisor   := Sin(Epsilon) * Tan(theLatitude) + Cos(Epsilon) * Sin(TSL);
    lambda    := ATan2 (divisor,numerator); (* REVERSE LOGIC here !!! *)
    lambda    := lambda + PI + PI; (* as + PI *)
    CalcAlphaDeltaSoleil ();
    Store (P);
END CalcDS;

PROCEDURE CalcFC (VAR P : skypos);
BEGIN
    (* IF placidusDomOK = FALSE THEN RETURN; END; *)
    zeroThem();
    alpha := TSL;
    CalcLambdaBetaSoleil ();
    lambda := lambda + PI; (* mc + PI *)
    CalcAlphaDeltaSoleil ();
    Store (P);
END CalcFC;

(* === CALC =================================================== *)

PROCEDURE Calc(planete : objet; VAR P : skypos; Cmode : calcmode);
BEGIN
    Mode := Cmode;

    CASE planete OF
    | soleil  :
                IF Mode=full THEN zeroThem; END; (* avoid silly results for lbr in astronomical mode *)
                CalcSoleil (P);
    | lune    :
                IF Mode=full THEN zeroThem; END; (* avoid silly results for lbr in astronomical mode *)
                CalcLune (P);
    | mercure : CalcMercure (P);
    | venus   : CalcVenus (P);
    | mars    : CalcMars (P);
    | jupiter : CalcJupiter (P);
    | saturne : CalcSaturne (P);
    | uranus  : CalcUranus (P);
    | neptune : CalcNeptune (P);
    | pluton  : CalcPluton (P);
    (* only if TSL has been computed ! *)
    | as      : CalcAS(P);
    | mc      : CalcMC(P);
    | ds      : CalcDS(P);
    | fc      : CalcFC(P);
    (* silly ! *)
    | lilith     : CalcLilith (P);
    | dragonhead : CalcDragonHead (P);
    | dragontail : CalcDragonTail (P);
    | blacksun   : CalcBlackSun (P);
    ELSE ; (* //ISO fixed silly requirement *) END;
END Calc;

PROCEDURE InitCalc (jd:LONGREAL;mode:calcmode);
VAR
    dummy : skypos;
BEGIN
    CalcT(jd);
    CalcSetup;
    (* now we CANNOT possibly forget it ! ;-) *)
    Calc(soleil,dummy,mode); (* soleil MUST ALWAYS be computed first ! *)
    initDomFlags(90.0); (* init both domifications flags to FALSE just in case *)
END InitCalc;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE StoreDom (cusp : LONGREAL) : LONGREAL;
BEGIN
     RadiansToDegrees(cusp);
     Mod360 (cusp);
     RETURN cusp;
END StoreDom;

PROCEDURE Symetrique (VAR P : housetype);
CONST
    halfcircle = LONGREAL(180.0);
VAR
    i : CARDINAL;v:LONGREAL;
BEGIN
    P[ORD(mDS)] := P[ORD(mAS)] + halfcircle;
    P[ORD(m8) ] := P[ORD(m2) ] + halfcircle;
    P[ORD(m9) ] := P[ORD(m3) ] + halfcircle;
    P[ORD(mFC)] := P[ORD(mMC)] + halfcircle;
    P[ORD(m5) ] := P[ORD(m11)] + halfcircle;
    P[ORD(m6) ] := P[ORD(m12)] + halfcircle;
    Mod360 (P[ORD(mDS)]);
    Mod360 (P[ORD(m8) ]);
    Mod360 (P[ORD(m9) ]);
    Mod360 (P[ORD(mFC)]);
    Mod360 (P[ORD(m5) ]);
    Mod360 (P[ORD(m6) ]);
END Symetrique;

(* ------------------------------------------------------------ *)

(*
   dtermination de la cuspide Placide avec itration
   c'est LA bonne mthode !
*)

VAR
    DifferenceDA : LONGREAL; (* diffrence ascensionnelle *)
    SemiArcSA    : LONGREAL; (* semi arc diurne *)

PROCEDURE CalcDAandSA;
BEGIN
    DifferenceDA := Tan (theLatitude) * Tan (delta);
    DifferenceDA := ASin (DifferenceDA);
    SemiArcSA := DifferenceDA + PI / 2.0;
    ModRadians (SemiArcSA);
END CalcDAandSA;

PROCEDURE DoPla (arc,pole,longfrom : LONGREAL) : LONGREAL;
VAR
    oldlambda : LONGREAL;
    DA : LONGREAL;
    ecart : LONGREAL;
BEGIN
    DegreesToRadians (pole);
    lambda := longfrom;         (* we start iteration from this angle *)
    REPEAT
        oldlambda := lambda;
        CalcAlphaDeltaSoleil ();
        CalcDAandSA ();
        alpha := TSL + pole + arc * DifferenceDA;
        CalcLambdaBetaSoleil ();
        ecart := lambda - oldlambda;
        ModRadians (ecart);     (* 0--360 *)
        IF ecart >= PI THEN
            ecart := ecart - PI;
        END;
    UNTIL ABS(ecart) < 0.0002; (* was .0005=about 2' and ABS() is important *)
    RETURN lambda;
END DoPla;

PROCEDURE DomPlacidus (VAR P : housetype);
VAR
    lo : LONGREAL;
BEGIN
    lo     := milieuduciel;
    P[ORD(mMC)] := StoreDom (lo);
    lo     := DoPla ( 1.0 / 3.0, 30.0, milieuduciel);
    P[ORD(m11)] := StoreDom (lo);
    lo     := DoPla ( 2.0 / 3.0, 60.0, lambda);
    P[ORD(m12)] := StoreDom (lo);
    lo     := ascendant;
    P[ORD(mAS)] := StoreDom (lo);
    lo     := DoPla ( 2.0 / 3.0, 120.0, ascendant);
    P[ORD(m2)]  := StoreDom (lo);
    lo     := DoPla ( 1.0 / 3.0, 150.0, lambda);
    P[ORD(m3)]  := StoreDom (lo);
END DomPlacidus;

(* ------------------------------------------------------------ *)

(* H. Gouchon's formulae seems wrong ! *)

(* this WORKS !
PROCEDURE doCampa (angleDeg:CARDINAL) : LONGREAL;
VAR
    KO:LONGREAL;
    divisor,numerator:LONGREAL;
    DN:LONGREAL;
BEGIN
    KO:=LONGREAL(angleDeg)+0.000001;
    DegreesToRadians(KO);

    DN:=ATan( Tan(KO)*Cos(theLatitude) );
    IF DN < 0.0 THEN DN:=DN+PI; END;
    IF Sin(KO) < 0.0 THEN DN:=DN+PI; END;

    numerator := Sin(TSL+DN);
    divisor := Cos(TSL+DN)*Cos(Epsilon)-Sin(DN)*Tan(theLatitude)*Sin(Epsilon);
    RETURN ATan2(divisor,numerator);
END doCampa;
*)

PROCEDURE doCampa (angleDeg:CARDINAL) : LONGREAL;
VAR
    angle : LONGREAL;
    numerator,divisor:LONGREAL;
    AO:LONGREAL;
BEGIN
    angle:=LONGREAL(angleDeg)+0.000001;
    DegreesToRadians(angle);
    numerator :=Sin(angle)*Cos(theLatitude); (* cos(phi) always >= 0 *)
    divisor   :=Cos(angle);
    angle:=ATan2(divisor,numerator);

    AO:=TSL+angle;
    numerator := Sin(AO);
    divisor := Cos(AO)*Cos(Epsilon)-Sin(angle)*Tan(theLatitude)*Sin(Epsilon);
    RETURN ATan2(divisor,numerator);
END doCampa;

PROCEDURE DomCampanus (VAR P : housetype);
VAR
    lo : LONGREAL;
BEGIN
    lo     := milieuduciel;
    P[ORD(mMC)] := StoreDom (lo);
    lo     := doCampa(30);
    P[ORD(m11)] := StoreDom (lo);
    lo     := doCampa(60);
    P[ORD(m12)] := StoreDom (lo);
    lo     := ascendant;
    P[ORD(mAS)] := StoreDom (lo);
    lo     := doCampa(120);
    P[ORD(m2)]  := StoreDom (lo);
    lo     := doCampa(150);
    P[ORD(m3)]  := StoreDom (lo);
END DomCampanus;

(* ------------------------------------------------------------ *)

(* H. Gouchon's formulae seems wrong ! *)

PROCEDURE doRegio (angleDeg:CARDINAL) : LONGREAL;
VAR
    angle : LONGREAL;
    numerator,divisor:LONGREAL;
    AO:LONGREAL;
BEGIN
    angle:=LONGREAL(angleDeg);
    DegreesToRadians(angle);
    AO:=TSL+angle;
    numerator := Sin(AO);
    divisor := Cos(AO)*Cos(Epsilon)-Sin(angle)*Tan(theLatitude)*Sin(Epsilon);
    RETURN ATan2(divisor,numerator);
END doRegio;

PROCEDURE DomRegiomontanus (VAR P : housetype);
VAR
    lo : LONGREAL;
BEGIN
    lo     := milieuduciel; (* id 0 *)
    P[ORD(mMC)] := StoreDom (lo);
    lo     := doRegio(30);
    P[ORD(m11)] := StoreDom (lo);
    lo     := doRegio(60);
    P[ORD(m12)] := StoreDom (lo);
    lo     := ascendant; (* id 90 *)
    P[ORD(mAS)] := StoreDom (lo);
    lo     := doRegio(120);
    P[ORD(m2)]  := StoreDom (lo);
    lo     := doRegio(150);
    P[ORD(m3)]  := StoreDom (lo);
END DomRegiomontanus;

(* ------------------------------------------------------------ *)

(* Ptolme : maisons gales  partir de l'AS *)

PROCEDURE doPtol (angle:INTEGER) : LONGREAL; (* yes, INTEGER because can be < 0 ! *)
VAR
    v : LONGREAL;
BEGIN
    v := LONGREAL(angle);
    DegreesToRadians(v);
    RETURN (ascendant+v);
END doPtol;

PROCEDURE DomPtolemeus (VAR P : housetype);
VAR
    lo : LONGREAL;
BEGIN
    lo     := doPtol(-105);
    P[ORD(mMC)] := StoreDom (lo);
    lo     := doPtol(-75);
    P[ORD(m11)] := StoreDom (lo);
    lo     := doPtol(-45);
    P[ORD(m12)] := StoreDom (lo);
    lo     := doPtol(-15);
    P[ORD(mAS)] := StoreDom (lo);
    lo     := doPtol(15);
    P[ORD(m2)]  := StoreDom (lo);
    lo     := doPtol(45);
    P[ORD(m3)]  := StoreDom (lo);
END DomPtolemeus;

(* ------------------------------------------------------------ *)

PROCEDURE dokoch (angleDeg:CARDINAL) : LONGREAL;
VAR
    angle : LONGREAL;
    numerator,divisor:LONGREAL;
    A1,portion,KN,quadrant,AO : LONGREAL;
BEGIN
    A1 := ASin( Sin(TSL) * Tan(theLatitude) * Tan(Epsilon) );
    IF angleDeg < 180 THEN
        KN :=  1.0; quadrant := 1.0;
    ELSE
        KN := -1.0; quadrant := 3.0;
    END;
    angle := LONGREAL(angleDeg);
    portion := (angle / 90.0) - quadrant;
    DegreesToRadians(angle);
    AO := TSL + angle + (A1 * portion);
    numerator := Sin(AO);
    divisor   := Cos(AO) * Cos(Epsilon) - KN * Tan(theLatitude) * Sin(Epsilon);
    RETURN ATan2(divisor,numerator);
END dokoch;

PROCEDURE DomKoch (VAR P : housetype);
VAR
    lo : LONGREAL;
BEGIN
    lo     := milieuduciel; (* id 0 *)
    P[ORD(mMC)] := StoreDom (lo);
    lo     := dokoch(30);
    P[ORD(m11)] := StoreDom (lo);
    lo     := dokoch(60);
    P[ORD(m12)] := StoreDom (lo);
    lo     := ascendant; (* id 90 *)
    P[ORD(mAS)] := StoreDom (lo);
    lo     := dokoch(120);
    P[ORD(m2)]  := StoreDom (lo);
    lo     := dokoch(150);
    P[ORD(m3)]  := StoreDom (lo);
END DomKoch;

(* ------------------------------------------------------------ *)

(* code shamelessly ripped from QD_Sky.MOD then adapted *)

(*

 no longer assume normalized [0..360[ angles : do it again just in case

 check planet in house cuspN, cuspN+1
 return %True if BI <= L < BS

 si bi < bs alors
                 si l < bi alors niet
                 si l < bs alors ok
                 niet
            sinon
                 si l < bs alors ok
                 si l < bi alors niet
                 ok
*)

PROCEDURE isBetweenThem (BorneInf,pos,BorneSup:LONGREAL) : BOOLEAN;
BEGIN
    Mod360(BorneInf);
    Mod360(BorneSup);
    Mod360(pos);

    IF BorneInf < BorneSup THEN
                                (* BI --- BS *)
        IF pos < BorneSup THEN
            IF pos < BorneInf THEN
            RETURN FALSE;       (* L < BI -- BS *)
            ELSE
                RETURN TRUE;        (* ok *)
            END;
        ELSE
            RETURN FALSE;           (* BI -- BS <= L *)
        END;
    ELSE
                                (* BI --- 0 --- BS *)
    IF pos < BorneInf THEN
        IF pos < BorneSup THEN
            RETURN TRUE;        (* ok *)
            ELSE
            RETURN FALSE;       (* BI --- 0 --- BS <= L *)
            END;
        ELSE
            RETURN TRUE;        (* BI <= L --- 0 --- BS *)
        END;
    END;
END isBetweenThem;

(*
   shortest distance between two degrees
*)

PROCEDURE minDegDistance (l1,l2:LONGREAL):LONGREAL;
VAR
    i : LONGREAL;
BEGIN
    i := ABS(l1-l2);
    IF i < 180.0 THEN
        RETURN i;
    ELSE
        RETURN ( 360.0 - i );
    END;
END minDegDistance;

PROCEDURE minDegDifference (l1,l2:LONGREAL):LONGREAL;
VAR
    i,k : LONGREAL;
BEGIN
    i := l2 - l1;
    IF ABS(i) < 180.0 THEN
        RETURN i;
    ELSE
        IF i < 0.0 THEN
            k:=-1.0;
        ELSE
            k:=1.0;
        END;
        RETURN k * ( ABS(i)-360.0 )
    END;
END minDegDifference;

PROCEDURE calcMidPoint (l1,l2:LONGREAL):LONGREAL;
VAR
    l:LONGREAL;
BEGIN
    l:=(l1+l2) / 2.0;
    IF minDegDistance(l1,l) < 90.0 THEN
        ;
    ELSE
        l:=l+180.0;
    END;
    Mod360(l);
    RETURN l;
END calcMidPoint;



(* ------------------------------------------------------------ *)

(* TSL, AS and MC must have been computed *)

PROCEDURE CalcDomification (required:domificationtype;
                            VAR cusp:housetype ) : domificationtype;
VAR
    i : CARDINAL;
BEGIN
    FOR i := firstCusp TO lastCusp DO
        cusp[i] := 0.0;
    END;
    IF allDomOK = FALSE THEN RETURN notapplicable; END;
    CASE required OF
    | notapplicable :
         RETURN notapplicable;
    | placidus :
         IF placidusDomOK = FALSE THEN RETURN notapplicable; END;
         DomPlacidus(cusp);
    | koch :
         IF placidusDomOK = FALSE THEN RETURN notapplicable; END;
         DomKoch(cusp);
    | campanus :
         DomCampanus(cusp);
    | regiomontanus :
         IF placidusDomOK = FALSE THEN RETURN notapplicable; END;
         DomRegiomontanus(cusp);
    | ptolemeus :
         IF placidusDomOK = FALSE THEN RETURN notapplicable; END;
         DomPtolemeus (cusp);
    | nonsensical :
         (* checking placidusDomOK is useless here *)
         DomPtolemeus(cusp);
    ELSE
         RETURN notapplicable;
    END;
    (* symetry already done if nonsensical *)
    IF required # nonsensical THEN Symetrique (cusp); END; (* 10 11 12 1 2 3 in mirror now *)
    RETURN required;
END CalcDomification;

(* ------------------------------------------------------------ *)

(*
    calculer ici la domitude en plus de la vraie maison
    formule et ide (!) piques  J.-C. Vitu
    warning ! rsultats parfois aberrants  proximit de 90-epsilon
*)

PROCEDURE CalcTrueHouse (VAR P : skypos) ;
VAR
    maison : LONGREAL;
    AH,DT,DD : LONGREAL;
    v : LONGREAL;
BEGIN
    P.truehouse := 0; (* already done by Store ! ;-) *)
    IF placidusDomOK = FALSE THEN RETURN; END;
    lambda := P.lambda;
    beta   := P.beta;
    DegreesToRadians (lambda);
    DegreesToRadians (beta);
    CalcAlphaDelta ();
    CalcDAandSA ();
    AH := TSL - alpha; (* angle horaire *)
    ModRadians (AH);
    DT := MaxRadians - SemiArcSA;
    DD := SemiArcSA / 3.0;
    IF (AH >= SemiArcSA) AND (AH <= DT) THEN (* the = were added 23/4/99 *)
        maison := (DT - AH) / ( (DT - SemiArcSA) / 6.0 ) + 1.0;

        (* 90+90*(180-AH)/(180-SAD) *)
        v:=(PI-AH)/(PI-SemiArcSA);
        v:=90.0+90.0*v;

    ELSIF AH < SemiArcSA THEN
        maison := 7.0 + (SemiArcSA - AH) / DD;

        (* 270-90*AH/SAD *)
        v:=AH/SemiArcSA;
        v:=270.0-90.0*v;

    ELSIF AH > DT THEN
        maison := 13.0 - (AH - DT) / DD;

        (* 270+90*(360-AH)/SAD *)
        v:=(MaxRadians - AH) / SemiArcSA;
        v:=270.0+90.0*v;

    ELSE
        maison := 0.0; (* should NEVER happen *)
        v := 0.0;
    END;
    P.truehouse := VAL (CARDINAL,maison);

    Mod360(v);
    P.domitude := v;

    P.planetaryhour := 15.0 * (4.0 * (360.0 - v) / 60.0); (* hours * 15 to get degrees *)
END CalcTrueHouse;

(* skypos MUST contain required data ! *)

PROCEDURE CalcAzimutAltitude (VAR P : skypos); (* alpha, delta, TSL, latitude *)
VAR
   numerator,divisor : LONGREAL;
   azimut,altitude:LONGREAL;
   H : LONGREAL;
BEGIN
   alpha := P.alpha;
   delta := P.delta;
   DegreesToRadians(alpha);
   DegreesToRadians(delta);
   H := TSL - alpha;
   numerator := Sin(H);
   divisor   := Cos(H) * Sin(theLatitude) - Tan(delta) * Cos(theLatitude);
   azimut    := ATan2 (divisor,numerator); (* REVERSE LOGIC here !!! *)
   altitude := Sin(theLatitude) * Sin(delta);
   altitude := altitude + Cos(theLatitude) * Cos(delta) * Cos(H);
   altitude := ASin(altitude);
   RadiansToDegrees(azimut);
   RadiansToDegrees(altitude);
   Mod360(azimut); (* only for azimut ! *)
   P.azimut := azimut;
   P.altitude:= altitude;
END CalcAzimutAltitude;

(* ------------------------------------------------------------ *)

PROCEDURE getLastCusp (id:domificationtype ):CARDINAL;
BEGIN
    CASE id OF
    | nonsensical :
        RETURN ORD(m8);
    ELSE
        RETURN ORD(m12);
    END;
END getLastCusp;

PROCEDURE chkDomificationRange (v:LONGCARD):BOOLEAN ;
VAR
    n : CARDINAL;
BEGIN
    n := CARDINAL(v);
    CASE n OF
    | ORD(placidus),ORD(koch),ORD(campanus),ORD(regiomontanus),ORD(ptolemeus),ORD(nonsensical):
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END chkDomificationRange;

PROCEDURE angleIsCusp (id:domificationtype):BOOLEAN ;
BEGIN
    CASE id OF
    | ptolemeus,nonsensical :
        RETURN FALSE;
    ELSE
        RETURN TRUE;
    END;
END angleIsCusp;

PROCEDURE clockwiseDom (id:domificationtype):BOOLEAN ;
BEGIN
    CASE id OF
    | nonsensical :
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END clockwiseDom;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

BEGIN
    PI         := LONGREAL(4.0) * ATan (1.0);
    MaxRadians := LONGREAL(8.0) * ATan (1.0); (* PI*2 *)
    MaxDegrees := LONGREAL(360.0);
    DegRad     := MaxRadians / MaxDegrees;
END QD_Calc.