(* ---------------------------------------------------------------
Title         Q&D Ini
Author        PhG
Overview      some use(d|ful) subroutines
Notes         warning ! only one inifile at a time !
Bugs
Wish List     speed up findentry by keeping some sort of indexes for sections

--------------------------------------------------------------- *)

IMPLEMENTATION MODULE QD_Ini;

IMPORT SYSTEM;
IMPORT Str;
IMPORT IO;
IMPORT FIO;
IMPORT Lib;

FROM Storage IMPORT ALLOCATE,DEALLOCATE;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, chkJoker,
isOption, GetOptIndex, GetLongCard, GetString, CharCount, same,
aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks;



FROM SYSTEM IMPORT BYTE,ADR;

(* ------------------------------------------------------------ *)

CONST
    nullchar    = CHR(0);
    semicolon   = ";";
    pound       = "#";
    openbracket = "[";
    closebracket= "]";
    quote       = "'";
    doublequote = '"';
    blank       = " ";
    tab         = CHR(9);
    comma       = ",";
    equate      = "=";
    dot         = ".";
    slash       = "/";

(* ------------------------------------------------------------ *)

PROCEDURE CleanStr (VAR S : ARRAY OF CHAR);
BEGIN
    LtrimBlanks(S);
    RtrimBlanks(S);
END CleanStr;

CONST
    MAXHASH       = 65535; (* max(cardinal) should be enough... *)

PROCEDURE processLine (VAR S : ARRAY OF CHAR;
                       VAR isSection:BOOLEAN;VAR len:CARDINAL;
                       VAR hashasis,hashupper:CARDINAL);
VAR
    R,T : str128;
    p : CARDINAL;
BEGIN
    CASE S[0] OF
    | openbracket : (* [XXX] *)
        isSection := TRUE;
        Ltrim(S,openbracket);
        Rtrim(S,closebracket);
        CleanStr(S);
        Str.Prepend(S,openbracket);
        Str.Copy(R,S);
        UpperCase(R);
        hashasis  := Lib.HashString(S,MAXHASH); (* "[section" *)
        hashupper := Lib.HashString(R,MAXHASH); (* "[SECTION" *)
    ELSE            (* XXXX=YYYY *)
        isSection := FALSE;
        p:=Str.CharPos(S,equate);
        IF p # MAX(CARDINAL) THEN (* should ALWAYS be true ! *)
            Str.Slice(R,S,0,p);
            Str.Delete(S,0,p+1);
            CleanStr(R); (* xxxx part of xxxx=yyyy *)
            CleanStr(S); (* yyyy part of xxxx=yyyy *)
            Str.Copy(T,R);
            UpperCase(T);
            hashasis  := Lib.HashString(R,MAXHASH); (* "xxxx" part of xxxx=YYYY *)
            hashupper := Lib.HashString(T,MAXHASH); (* "XXXX" part of xxxx=YYYY *)
            Str.Append(R,equate);
            Str.Append(R,S);
            Str.Copy(S,R);
        ELSE
            (* nada, let it be but we should abort or fake something... *)
            Str.Copy(T,S);
            UpperCase(T);
            hashasis  := Lib.HashString(S,MAXHASH);
            hashupper := Lib.HashString(T,MAXHASH);
        END;
    END;
    len := Str.Length(S);
END processLine;

(* ------------------------------------------------------------ *)

TYPE
    ptrIniEntry = POINTER TO iniEntry;
    iniEntry = RECORD
        next     : ptrIniEntry;
        isSection: BOOLEAN;
        hashasis : CARDINAL; (* original string  *)
        hashupper: CARDINAL; (* uppercase string *)
        slen     : SHORTCARD;
        string   : CHAR; (* after other data because variable length *)
    END;

VAR
    anchor : ptrIniEntry;

PROCEDURE FreeIni ();
VAR
    firstInList : ptrIniEntry;
    newInList   : ptrIniEntry;
    len         : CARDINAL;
    needed      : CARDINAL;
BEGIN
    firstInList := anchor;
    newInList   := firstInList;
    WHILE newInList <> NIL DO
         len         := CARDINAL(newInList^.slen);
         needed      := SIZE(iniEntry)-SIZE(CHAR)+len;
         firstInList := firstInList^.next;
         DEALLOCATE(newInList,needed);
         newInList := firstInList;
    END
END FreeIni;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead; (* 4 Kb buffer is enough *)
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
VAR
    ioBuffer        : ARRAY [firstBufferByte..lastBufferByte] OF BYTE;

PROCEDURE readIniFile (ini : ARRAY OF CHAR) : BOOLEAN;
VAR
    S        : str128;    (* should be enough *)
    hnd      : FIO.File;

    isSection: BOOLEAN;

    needed      : CARDINAL;
    firstInList : ptrIniEntry;
    newInList   : ptrIniEntry;
    len         : CARDINAL;
    hashasis    : CARDINAL;
    hashupper   : CARDINAL;
BEGIN
    FIO.EOF:=FALSE; (* safety ! *)
    hnd:=FIO.OpenRead(ini);
    FIO.AssignBuffer(hnd,ioBuffer);
    LOOP
        IF FIO.EOF THEN
            FIO.Close(hnd);
            RETURN FALSE; (* empty ini file ! *)
        END;
        FIO.RdStr(hnd,S);
        IF ((S[0]=nullchar) AND FIO.EOF) THEN
            FIO.Close(hnd);
            RETURN FALSE; (* weird hack ! *)
        END;
        CleanStr(S); (* remove leading and trailing blanks, tabs *)
        CASE S[0] OF
        | nullchar,semicolon,pound :
            (* ignore empty string or comment *)
        ELSE
            IF Str.Length(S) # 0 THEN EXIT; END; (* always here *)
        END;
    END;

    processLine(S,isSection,len,hashasis,hashupper);

    needed := SIZE(iniEntry)-SIZE(CHAR)+len; (* iniEntry.string is a CHAR *)

    ALLOCATE(firstInList,needed);
    firstInList^.next := NIL;

    firstInList^.isSection := isSection;
    firstInList^.hashasis  := hashasis;
    firstInList^.hashupper := hashupper;
    firstInList^.slen      := SHORTCARD(len);
    Lib.FastMove( ADR(S),ADR(firstInList^.string),len);

    newInList := firstInList;

    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hnd,S);
        IF ((S[0]=nullchar) AND FIO.EOF) THEN EXIT; END;
        CleanStr(S);
        CASE S[0] OF
        | nullchar,semicolon,pound :
            (* ignore empty string or comment *)
        ELSE
            IF Str.Length(S) # 0 THEN (* always here *)

                processLine(S,isSection,len,hashasis,hashupper);

                needed := SIZE(iniEntry)-SIZE(CHAR)+len; (* iniEntry.string is a CHAR *)

                ALLOCATE(newInList^.next,needed);
                newInList            := newInList^.next;

                newInList^.isSection := isSection;
                newInList^.hashasis  := hashasis;
                newInList^.hashupper := hashupper;
                newInList^.slen      := SHORTCARD(len);
                Lib.FastMove( ADR(S),ADR(newInList^.string),len);

                newInList^.next      := NIL;
            END;
        END;
    END;
    FIO.Close(hnd);
    anchor := firstInList;
    RETURN TRUE;
END readIniFile;

(* ------------------------------------------------------------ *)

PROCEDURE readIniFileEnh (ini : ARRAY OF CHAR) : BOOLEAN;
VAR
    S        : str128; (* should be enough *)
    hnd      : FIO.File;

    isSection: BOOLEAN;

    needed      : CARDINAL;
    firstInList : ptrIniEntry;
    newInList   : ptrIniEntry;
    len         : CARDINAL;
    hashasis    : CARDINAL;
    hashupper   : CARDINAL;

    i           : CARDINAL;
    rc          : BOOLEAN;
    c           : CHAR;
BEGIN

    FIO.EOF:=FALSE; (* safety ! *)
    hnd:=FIO.OpenRead(ini);
    FIO.AssignBuffer(hnd,ioBuffer);
    LOOP
        IF FIO.EOF THEN
            FIO.Close(hnd);
            RETURN FALSE; (* empty ini file ! *)
        END;
        FIO.RdStr(hnd,S);
        IF ((S[0]=nullchar) AND FIO.EOF) THEN
            FIO.Close(hnd);
            RETURN FALSE; (* weird hack ! *)
        END;


        CleanStr(S); (* remove leading and trailing blanks, tabs *)
        CASE S[0] OF
        | nullchar,semicolon,pound :
            (* ignore empty string or comment *)
        ELSE
            IF Str.Length(S) # 0 THEN EXIT; END; (* always here *)
        END;
    END;

    processLine(S,isSection,len,hashasis,hashupper);

    needed := SIZE(iniEntry)-SIZE(CHAR)+len; (* iniEntry.string is a CHAR *)

    ALLOCATE(firstInList,needed);
    firstInList^.next := NIL;

    firstInList^.isSection := isSection;
    firstInList^.hashasis  := hashasis;
    firstInList^.hashupper := hashupper;
    firstInList^.slen      := SHORTCARD(len);
    Lib.FastMove( ADR(S),ADR(firstInList^.string),len);

    newInList := firstInList;

    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hnd,S);
        IF ((S[0]=nullchar) AND FIO.EOF) THEN EXIT; END;


        CleanStr(S);
        CASE S[0] OF
        | nullchar,semicolon,pound :
            (* ignore empty string or comment *)
        ELSE
            IF Str.Length(S) # 0 THEN (* always here *)

                processLine(S,isSection,len,hashasis,hashupper);

                needed := SIZE(iniEntry)-SIZE(CHAR)+len; (* iniEntry.string is a CHAR *)

                ALLOCATE(newInList^.next,needed);
                newInList            := newInList^.next;

                newInList^.isSection := isSection;
                newInList^.hashasis  := hashasis;
                newInList^.hashupper := hashupper;
                newInList^.slen      := SHORTCARD(len);
                Lib.FastMove( ADR(S),ADR(newInList^.string),len);

                newInList^.next      := NIL;
            END;
        END;
    END;
    FIO.Close(hnd);
    anchor := firstInList;
    RETURN TRUE;
END readIniFileEnh;

(* ------------------------------------------------------------ *)

PROCEDURE listSections ();
VAR
    firstInList : ptrIniEntry;
    newInList   : ptrIniEntry;
    len         : CARDINAL;
    S           : str128;
    isSection   : BOOLEAN;
BEGIN
    firstInList := anchor;
    newInList   := firstInList;
    WHILE newInList <> NIL DO
         firstInList := firstInList^.next;
         len         := CARDINAL(newInList^.slen);
         Lib.FastMove( ADR(newInList^.string),ADR(S),len);
         S[len]:=nullchar; (* REQUIRED safety ! *)
         isSection := newInList^.isSection;

         IF isSection THEN
             IO.WrStr(S);
             IO.WrChar(closebracket);
             IO.WrLn;
         END;
         newInList := firstInList;
    END;
END listSections;

VAR
    globalNewInList : ptrIniEntry;

PROCEDURE getFirstSection (VAR R:ARRAY OF CHAR):BOOLEAN;
VAR
    firstInList : ptrIniEntry;
    len         : CARDINAL;
    S           : str128;
    isSection   : BOOLEAN;
BEGIN
    firstInList := anchor;
    globalNewInList   := firstInList;
    WHILE globalNewInList <> NIL DO
         firstInList := firstInList^.next;
         len         := CARDINAL(globalNewInList^.slen);
         Lib.FastMove( ADR(globalNewInList^.string),ADR(S),len);
         S[len]:=nullchar; (* REQUIRED safety ! *)
         isSection := globalNewInList^.isSection;

         globalNewInList := firstInList;
         IF isSection THEN
             Str.Concat(R,S,closebracket);
             RETURN TRUE;
         END;
    END;
    RETURN FALSE;
END getFirstSection;

PROCEDURE getNextSection (VAR R:ARRAY OF CHAR):BOOLEAN;
VAR
    firstInList : ptrIniEntry;
    len         : CARDINAL;
    S           : str128;
    isSection   : BOOLEAN;
BEGIN
    firstInList := globalNewInList;
    WHILE globalNewInList <> NIL DO
         firstInList := firstInList^.next;
         len         := CARDINAL(globalNewInList^.slen);
         Lib.FastMove( ADR(globalNewInList^.string),ADR(S),len);
         S[len]:=nullchar; (* REQUIRED safety ! *)
         isSection := globalNewInList^.isSection;
         globalNewInList := firstInList;

         IF isSection THEN
             Str.Concat(R,S,closebracket);
             RETURN  TRUE;
         END;
    END;
    RETURN FALSE;
END getNextSection;

PROCEDURE listAll ();
VAR
    firstInList : ptrIniEntry;
    newInList   : ptrIniEntry;
    len         : CARDINAL;
    S           : str128;
    isSection   : BOOLEAN;
BEGIN
    firstInList := anchor;
    newInList   := firstInList;
    WHILE newInList <> NIL DO
         firstInList := firstInList^.next;
         len         := CARDINAL(newInList^.slen);
         Lib.FastMove( ADR(newInList^.string),ADR(S),len);
         S[len]:=nullchar; (* REQUIRED safety ! *)
         isSection := newInList^.isSection;

         IF isSection THEN IO.WrLn; END;
         IO.WrStr(S);
         IF isSection THEN
             IO.WrChar(closebracket);
             IO.WrLn;
         END;
         IO.WrLn;
         newInList := firstInList;
    END;
END listAll;

(* ------------------------------------------------------------ *)

(* now letting value case unaffected *)

PROCEDURE findEntry (section,item:ARRAY OF CHAR;exact:BOOLEAN;
                     VAR value : ARRAY OF CHAR) : BOOLEAN;

VAR
    firstInList : ptrIniEntry;
    newInList   : ptrIniEntry;
    S,R         : str128;
    len,hashsection,hashitem,hash : CARDINAL;
    state       : (waiting,foundsection,founditem);
    p           : CARDINAL;
BEGIN
    CleanStr(item);
    Ltrim(section,openbracket);
    Rtrim(section,closebracket);
    CleanStr(section);
    Str.Prepend(section,openbracket);
    IF exact=FALSE THEN
        UpperCase(section);
        UpperCase(item);
    END;
    hashsection := Lib.HashString(section,MAXHASH);
    hashitem    := Lib.HashString(item   ,MAXHASH);

    state := waiting;

    firstInList := anchor;
    newInList   := firstInList;

    LOOP (* was WHILE newInList <> NIL DO *)
         IF newInList = NIL THEN EXIT; END;
         firstInList := firstInList^.next;
         IF newInList^.isSection THEN
             IF exact THEN
                 hash := newInList^.hashasis;
             ELSE
                 hash := newInList^.hashupper;
             END;
             IF hash = hashsection THEN
                 (* compare strings just in case... *)
                 len := CARDINAL(newInList^.slen);
                 Lib.FastMove( ADR(newInList^.string),ADR(S),len);
                 S[len]:=nullchar; (* REQUIRED safety ! *)
                 IF exact = FALSE THEN UpperCase(S); END;
                 IF same(section,S) THEN (* here=newInList *)
                     state := foundsection;
                     EXIT;
                 END;
             END;
         END;
         newInList := firstInList;
    END;
    IF state = waiting THEN RETURN FALSE; END; (* no match for section *)

    newInList := newInList^.next; (* next link, of course *)

    LOOP (* was WHILE newInList <> NIL DO *)
         IF newInList = NIL THEN EXIT; END;
         firstInList := firstInList^.next;
         IF newInList^.isSection THEN EXIT; END; (* section passed *)
         IF exact THEN
             hash := newInList^.hashasis;
         ELSE
             hash := newInList^.hashupper;
         END;
         IF hash = hashitem THEN
             (* compare strings just in case... *)
             len := CARDINAL(newInList^.slen);
             Lib.FastMove( ADR(newInList^.string),ADR(S),len);
             S[len]:=nullchar; (* REQUIRED safety ! *)
             (* IF exact = FALSE THEN UpperCase(S); END; *)
             p:=Str.CharPos(S,equate);
             IF p # MAX(CARDINAL) THEN
                 Str.Slice(R,S,0,p);
                 IF exact=FALSE THEN UpperCase(R);END;
                 IF same(item,R) THEN
                     Str.Delete(S,0,p+1); (* ok, we keep original case now *)
                     Str.Copy(value,S);
                     state := founditem;
                     EXIT;
                 END;
             END;
         END;
         newInList := firstInList;
    END;
    RETURN (state=founditem);
END findEntry;

(* ------------------------------------------------------------ *)

PROCEDURE findSection (section:ARRAY OF CHAR;exact:BOOLEAN):BOOLEAN;
VAR
    firstInList : ptrIniEntry;
    newInList   : ptrIniEntry;
    S           : str128;
    len,hashsection,hash: CARDINAL;
    state       : (waiting,foundsection);
    p           : CARDINAL;
BEGIN
    Ltrim(section,openbracket);
    Rtrim(section,closebracket);
    CleanStr(section);
    Str.Prepend(section,openbracket);
    IF exact=FALSE THEN UpperCase(section); END;
    hashsection := Lib.HashString(section,MAXHASH);

    state := waiting;

    firstInList := anchor;
    newInList   := firstInList;

    LOOP (* was WHILE newInList <> NIL DO *)
         IF newInList = NIL THEN EXIT; END;
         firstInList := firstInList^.next;
         IF newInList^.isSection THEN
             IF exact THEN
                 hash := newInList^.hashasis;
             ELSE
                 hash := newInList^.hashupper;
             END;
             IF hash = hashsection THEN
                 (* compare strings just in case... *)
                 len := CARDINAL(newInList^.slen);
                 Lib.FastMove( ADR(newInList^.string),ADR(S),len);
                 S[len]:=nullchar; (* REQUIRED safety ! *)
                 IF exact = FALSE THEN UpperCase(S); END;
                 IF same(section,S) THEN (* here=newInList *)
                     state := foundsection;
                     EXIT;
                 END;
             END;
         END;
         newInList := firstInList;
    END;
    RETURN (state=foundsection);
END findSection;

(* ------------------------------------------------------------ *)

PROCEDURE getItemStr (section,entry,default:ARRAY OF CHAR;
          VAR result:ARRAY OF CHAR);
VAR
    found : BOOLEAN;
    R     : str128;
BEGIN
    found:=findEntry (section,entry,FALSE,R);
    IF found=FALSE THEN Str.Copy(R,default); END;
    (* tss, we don't even check there are 2 quotes ! ;-) *)
    Ltrim(R,doublequote);
    Rtrim(R,doublequote);
    Str.Copy(result,R);
END getItemStr;

PROCEDURE getItemStrUp (section,entry,default:ARRAY OF CHAR;
          VAR result:ARRAY OF CHAR);
BEGIN
    getItemStr(section,entry,default,result);
    UpperCase(result);
END getItemStrUp;

(* ------------------------------------------------------------ *)

(* strtoreal does not like embedded spaces ! better slow safe than sorry ! *)

PROCEDURE numclean (VAR R : ARRAY OF CHAR);
BEGIN
    UpperCase(R);
    ReplaceChar(R,tab,"");
    ReplaceChar(R,blank,"");
END numclean;

(* ------------------------------------------------------------ *)

PROCEDURE getItemCard (section,entry,default:ARRAY OF CHAR;
          VAR result:CARDINAL);
VAR
    found : BOOLEAN;
    R     : str128;
    v     : LONGCARD;
    ok    : BOOLEAN;
BEGIN
    found:=findEntry (section,entry,FALSE,R);
    IF found=FALSE THEN Str.Copy(R,default); END;

    numclean(R);

    v:=Str.StrToCard(R,10,ok);
    IF (ok=FALSE) OR (v > MAX(CARDINAL)) THEN
        result:=MAX(CARDINAL);
    ELSE
        result:=CARDINAL(v);
    END;
END getItemCard;

(* ------------------------------------------------------------ *)

PROCEDURE getItemBool (section,entry,default:ARRAY OF CHAR;
          VAR result:BOOLEAN);
VAR
    found : BOOLEAN;
    R     : str128;
BEGIN
    found:=findEntry (section,entry,FALSE,R);
    IF found=FALSE THEN Str.Copy(R,default); END;

    UpperCase(R); (* really required here *)

    IF (same(R,"Y")) OR (same(R,"1")) OR (same(R,"O")) THEN
        result:=TRUE;
    ELSIF (same(R,"N")) OR (same(R,"0")) THEN
        result:=FALSE;
    ELSIF (same(R,"YES")) OR (same(R,"OUI")) THEN
        result:=TRUE;
    ELSIF (same(R,"NO")) OR (same(R,"NON")) THEN
        result:=FALSE;
    ELSE
        result:=FALSE;
    END;
END getItemBool;

(* ------------------------------------------------------------ *)

PROCEDURE getItemChar (section,entry,default:ARRAY OF CHAR;
          VAR result:CHAR);
VAR
    found : BOOLEAN;
    R     : str128;
BEGIN
    found:=findEntry (section,entry,FALSE,R);
    IF found=FALSE THEN Str.Copy(R,default); END;
    (* tsss, we don't even check there are two quotes ! ;-) *)
    (* tsss, what if we want a double quote alone ? ;-) *)
    Ltrim(R,doublequote);
    Rtrim(R,doublequote);
    Str.Copy(result,R[0]);
END getItemChar;

PROCEDURE getItemCharUp (section,entry,default:ARRAY OF CHAR;
          VAR result:CHAR);
BEGIN
    getItemChar(section,entry,default,result);
    UpperCase(result);
END getItemCharUp;

(* ------------------------------------------------------------ *)

(*
    read value and check it is in the ]0..1[ range
    entry decimal point is a comma or a dot, or entry is a fraction
    minimal error checking !!! -- we assume default is a legal value
    numbers MUST be >= 0 !
*)

PROCEDURE getItemReal0_1 (section,entry,default:ARRAY OF CHAR;
          VAR result:LONGREAL);
VAR
    found : BOOLEAN;
    R     : str128;
    v     : LONGREAL;
    ok    : BOOLEAN;
VAR
    p : CARDINAL;
    str : str128;
    numerator,divisor : LONGCARD;
BEGIN
    found:=findEntry (section,entry,FALSE,R);
    IF found=FALSE THEN Str.Copy(R,default); END;

    numclean(R);

    Str.Subst(R,comma,dot); (* just in CASE *)
    p := Str.CharPos(R,slash);
    IF p = MAX(CARDINAL) THEN
        v:=Str.StrToReal(R,ok);
    ELSE
        Str.Slice(str,R,0,p); (* numerator *)
        Str.Delete(R,0,p+1);  (* divisor *)
        numerator:=Str.StrToCard(str,10,ok);
        IF ok THEN
            divisor := Str.StrToCard(R,10,ok);
            IF ok THEN
                IF divisor=0 THEN
                    ok := FALSE;
                ELSE
                    v:=LONGREAL(numerator) / LONGREAL(divisor);
                END;
            END;
        END;
    END;
    IF ok THEN
        IF ( (v > 0.0) AND (v < 1.0) ) THEN
            (* good value ! *)
        ELSE
            ok:=FALSE;
        END;
    END;
    IF ok THEN
        result := v;
    ELSE
        (* if we're here, value is either illegal or not in range : let's use default *)
        IF found THEN (* try default *)
            Str.Copy(R,default);
            Str.Subst(R,comma,dot); (* just in CASE *)
            p := Str.CharPos(R,slash);
            IF p = MAX(CARDINAL) THEN
                v:=Str.StrToReal(R,ok);
            ELSE
                Str.Slice(str,R,0,p); (* numerator *)
                Str.Delete(R,0,p+1);  (* divisor *)
                numerator:=Str.StrToCard(str,10,ok);
                divisor := Str.StrToCard(R,10,ok);
                v:=LONGREAL(numerator) / LONGREAL(divisor);
            END;
            result := v; (* assume default is ok *)
        ELSE
            result:= 0.0; (* assume ugly result because default was bad *)
        END;
    END;
END getItemReal0_1;

(* ------------------------------------------------------------ *)

(*
    entry decimal point is a comma or a dot
    minimal error checking !!! -- we assume default is a legal value
*)

PROCEDURE getItemReal (section,entry,default:ARRAY OF CHAR;
          VAR result:LONGREAL);
VAR
    found : BOOLEAN;
    R     : str128;
    v     : LONGREAL;
    ok    : BOOLEAN;
VAR
    str : str128;
BEGIN
    found:=findEntry (section,entry,FALSE,R);
    IF found=FALSE THEN Str.Copy(R,default); END;

    numclean(R);

    Str.Subst(R,comma,dot); (* just in CASE *)
    v:=Str.StrToReal(R,ok);
    IF ok THEN
        result := v;
    ELSE
        (* if we're here, value is illegal : let's use default *)
        IF found THEN
            Str.Copy(R,default);
            Str.Subst(R,comma,dot); (* just in CASE *)
            v:=Str.StrToReal(R,ok);
            result := v; (* assume default is ok *)
        ELSE
            result:= 0.0; (* assume ugly result because default was bad *)
        END;
    END;
END getItemReal;

(* ------------------------------------------------------------ *)

(*
    read real value
    entry decimal point is a comma or a dot, or entry is a fraction
    minimal error checking !!! -- we assume default is a legal value
    number CAN be < 0 !
*)

PROCEDURE getItemFrac (section,entry,default:ARRAY OF CHAR;
          VAR result:LONGREAL);
VAR
    found : BOOLEAN;
    R     : str128;
    v     : LONGREAL;
    ok    : BOOLEAN;
VAR
    p : CARDINAL;
    str : str128;
    numerator,divisor : LONGINT; (* allow negative numbers *)
BEGIN
    found:=findEntry (section,entry,FALSE,R);
    IF found=FALSE THEN Str.Copy(R,default); END;

    numclean(R);

    Str.Subst(R,comma,dot); (* just in CASE *)
    p := Str.CharPos(R,slash);
    IF p = MAX(CARDINAL) THEN
        v:=Str.StrToReal(R,ok);
    ELSE
        Str.Slice(str,R,0,p); (* numerator *)
        Str.Delete(R,0,p+1);  (* divisor *)
        numerator:=Str.StrToInt(str,10,ok);
        IF ok THEN
            divisor := Str.StrToInt(R,10,ok);
            IF ok THEN
                IF divisor=0 THEN
                    ok := FALSE;
                ELSE
                    v:=LONGREAL(numerator) / LONGREAL(divisor);
                END;
            END;
        END;
    END;
    IF ok THEN
        result := v;
    ELSE
        (* if we're here, value is either illegal or not in range : let's use default *)
        IF found THEN (* try default *)
            Str.Copy(R,default);
            Str.Subst(R,comma,dot); (* just in CASE *)
            p := Str.CharPos(R,slash);
            IF p = MAX(CARDINAL) THEN
                v:=Str.StrToReal(R,ok);
            ELSE
                Str.Slice(str,R,0,p); (* numerator *)
                Str.Delete(R,0,p+1);  (* divisor *)
                numerator:=Str.StrToInt(str,10,ok);
                divisor := Str.StrToInt(R,10,ok);
                v:=LONGREAL(numerator) / LONGREAL(divisor);
            END;
            result := v; (* assume default is ok *)
        ELSE
            result:= 0.0; (* assume ugly result because default was bad *)
        END;
    END;
END getItemFrac;

(* ------------------------------------------------------------ *)

BEGIN
    FIO.IOcheck:=FALSE; (* avoid nasty problems ! *)

END QD_Ini.