(* ---------------------------------------------------------------
Title         Q&D Names
Author        PhG
Overview      some use(d|ful) subroutines
Notes
Bugs
Wish List

--------------------------------------------------------------- *)

IMPLEMENTATION MODULE QD_Names;

IMPORT FIO;
IMPORT Str;
IMPORT Lib;
IMPORT IO;

FROM QD_Calc IMPORT objet,cuspide,sens,domificationtype;

FROM QD_Box IMPORT str128, LtrimBlanks, RtrimBlanks,
UpperCase, same, str2048, ReplaceChar, str4096;

FROM SYSTEM IMPORT BYTE;

FROM QD_Make IMPORT FRENCH;

(* ------------------------------------------------------------ *)

PROCEDURE getObjectNameFR (id:objet; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | soleil :      S:="Soleil";
    | lune :        S:="Lune";
    | mercure :     S:="Mercure";
    | venus :       S:="Vnus";
    | mars :        S:="Mars";
    | jupiter :     S:="Jupiter";
    | saturne :     S:="Saturne";
    | uranus :      S:="Uranus";
    | neptune :     S:="Neptune";
    | pluton :      S:="Pluton";
    | as :          S:="Ascendant";
    | mc :          S:="Milieu du ciel";
    | ds :          S:="Descendant";
    | fc :          S:="Fond du ciel";
    | lilith :      S:="Lilith";
    | dragonhead :  S:="Noeud Nord";
    | dragontail :  S:="Noeud Sud";
    | blacksun :    S:="Soleil noir";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getObjectNameFR;

PROCEDURE getObjectName (id:objet; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | soleil :      S:="Soleil";
    | lune :        S:="Lune";
    | mercure :     S:="Mercure";
    | venus :       S:="Vnus";
    | mars :        S:="Mars";
    | jupiter :     S:="Jupiter";
    | saturne :     S:="Saturne";
    | uranus :      S:="Uranus";
    | neptune :     S:="Neptune";
    | pluton :      S:="Pluton";
    | as :          S:="Ascendant";
    | mc :          S:="Milieu du ciel";
    | ds :          S:="Descendant";
    | fc :          S:="Fond du ciel";
    | lilith :      S:="Lilith";
    | dragonhead :  S:="Noeud Nord";
    | dragontail :  S:="Noeud Sud";
    | blacksun :    S:="Soleil noir";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getObjectName;

PROCEDURE getObjectAbrev (id:objet; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | soleil :      S:="So";
    | lune :        S:="Lu";
    | mercure :     S:="Me";
    | venus :       S:="V";
    | mars :        S:="Ma";
    | jupiter :     S:="Ju";
    | saturne :     S:="Sa";
    | uranus :      S:="Ur";
    | neptune :     S:="Ne";
    | pluton :      S:="Pl";
    | as :          S:="AS";
    | mc :          S:="MC";
    | ds :          S:="DS";
    | fc :          S:="FC";
    | lilith :      S:="Li";
    | dragonhead :  S:="NN";
    | dragontail :  S:="NS";
    | blacksun :    S:="SN";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getObjectAbrev;

PROCEDURE getCuspideName (id:cuspide; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | mAS :  S:="Ascendant";
    | m2 :   S:="II";
    | m3 :   S:="III";
    | mFC :  S:="Fond du ciel";
    | m5 :   S:="V";
    | m6 :   S:="VI";
    | mDS :  S:="Descendant";
    | m8 :   S:="VIII";
    | m9 :   S:="IX";
    | mMC :  S:="Milieu du ciel";
    | m11 :  S:="XI";
    | m12 :  S:="XII";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getCuspideName;

(* house 1 is not always AS ! *)

PROCEDURE getCuspideNameAlt (id:cuspide; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | mAS :  S:="I";
    | m2 :   S:="II";
    | m3 :   S:="III";
    | mFC :  S:="IV";
    | m5 :   S:="V";
    | m6 :   S:="VI";
    | mDS :  S:="VII";
    | m8 :   S:="VIII";
    | m9 :   S:="IX";
    | mMC :  S:="X";
    | m11 :  S:="XI";
    | m12 :  S:="XII";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getCuspideNameAlt;

PROCEDURE getMotionStr (id:sens; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | direct :            S:="Direct";
    | retrograde :        S:="Rtrograde";
    | stationretrograde : S:="Rtrograde (aprs station)";
    | stationdirect :     S:="Direct (aprs station)";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getMotionStr;

PROCEDURE getDomificationStr (id:domificationtype; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | notapplicable : S:="Dtermination impossible en raison de la latitude !";
    | placidus :      S:="Placidus";
    | campanus :      S:="Campanus";
    | regiomontanus : S:="Regiomontanus";
    | ptolemeus :     S:="Ptolme";
    | koch      :     S:="Koch";
    | nonsensical  :      S:="Nonsensical";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getDomificationStr;

PROCEDURE getDomificationStrMap (id:domificationtype; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | notapplicable : S:="Aucune !";
    | placidus :      S:="Placidus";
    | campanus :      S:="Campanus";
    | regiomontanus : S:="Regiomontanus";
    | ptolemeus :     S:="Ptolme";
    | koch      :     S:="Koch";
    | nonsensical :       S:="Nonsensical";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getDomificationStrMap;

PROCEDURE getAspectNameFR (id:aspecttype; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | conjonction : S:="Conjonction";
    | opposition  : S:="Opposition";
    | carre       : S:="Carr";
    | trigone     : S:="Trigone";
    | sextile     : S:="Sextile";
    | semicarre   : S:="Semi-carr";
    | semisextile : S:="Semi-sextile";
    | quinconce   : S:="Quinconce";
    | sesquicarre : S:="Sesqui-carr";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getAspectNameFR;

PROCEDURE getAspectName (id:aspecttype; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | conjonction : S:="Conjonction";
    | opposition  : S:="Opposition";
    | carre       : S:="Carr";
    | trigone     : S:="Trigone";
    | sextile     : S:="Sextile";
    | semicarre   : S:="Semi-carr";
    | semisextile : S:="Semi-sextile";
    | quinconce   : S:="Quinconce";
    | sesquicarre : S:="Sesqui-carr";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getAspectName;

PROCEDURE getAspectAbrev (id:aspecttype; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | conjonction : S:="Co";
    | opposition  : S:="Op";
    | carre       : S:="Ca";
    | trigone     : S:="Tr";
    | sextile     : S:="Se";
    | semicarre   : S:="Sc";
    | semisextile : S:="Ss";
    | quinconce   : S:="Qu";
    | sesquicarre : S:="Sq";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getAspectAbrev;

PROCEDURE getAspectAbrevChar (id:aspecttype; VAR R : CHAR);
VAR
    S : CHAR;
BEGIN
    CASE id OF
    | conjonction : S:="C";
    | opposition  : S:="o";
    | carre       : S:="c";
    | trigone     : S:="t";
    | sextile     : S:="s";
    | semicarre   : S:="-";
    | semisextile : S:="+";
    | quinconce   : S:="-";
    | sesquicarre : S:="+";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getAspectAbrevChar;

PROCEDURE getSignNameFR (id:signtype; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | belier     : S:="Blier";
    | taureau    : S:="Taureau";
    | gemeaux    : S:="Gmeaux";
    | cancer     : S:="Cancer";
    | lion       : S:="Lion";
    | vierge     : S:="Vierge";
    | balance    : S:="Balance";
    | scorpion   : S:="Scorpion";
    | sagittaire : S:="Sagittaire";
    | capricorne : S:="Capricorne";
    | verseau    : S:="Verseau";
    | poissons   : S:="Poissons";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getSignNameFR;

PROCEDURE getSignName (id:signtype; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | belier     : S:="Blier";
    | taureau    : S:="Taureau";
    | gemeaux    : S:="Gmeaux";
    | cancer     : S:="Cancer";
    | lion       : S:="Lion";
    | vierge     : S:="Vierge";
    | balance    : S:="Balance";
    | scorpion   : S:="Scorpion";
    | sagittaire : S:="Sagittaire";
    | capricorne : S:="Capricorne";
    | verseau    : S:="Verseau";
    | poissons   : S:="Poissons";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getSignName;

PROCEDURE getSectorName (id:cuspide; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | mAS :  S:="Ascendant";
    | m2 :   S:="Secteur II";
    | m3 :   S:="Secteur III";
    | mFC :  S:="Fond du ciel";
    | m5 :   S:="Secteur V";
    | m6 :   S:="Secteur VI";
    | mDS :  S:="Descendant";
    | m8 :   S:="Secteur VIII";
    | m9 :   S:="Secteur IX";
    | mMC :  S:="Milieu du ciel";
    | m11 :  S:="Secteur XI";
    | m12 :  S:="Secteur XII";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getSectorName;

(* for those house system when AS is not house 1, such as Ptolme *)

PROCEDURE getSectorNameAlt (id:cuspide; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | mAS :  S:="Secteur I";
    | m2 :   S:="Secteur II";
    | m3 :   S:="Secteur III";
    | mFC :  S:="Secteur IV";
    | m5 :   S:="Secteur V";
    | m6 :   S:="Secteur VI";
    | mDS :  S:="Secteur VII";
    | m8 :   S:="Secteur VIII";
    | m9 :   S:="Secteur IX";
    | mMC :  S:="Secteur X";
    | m11 :  S:="Secteur XI";
    | m12 :  S:="Secteur XII";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getSectorNameAlt;

PROCEDURE getRomanStr (num:CARDINAL;VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE num OF
    | 1  :  S:="I";
    | 2  :  S:="II";
    | 3  :  S:="III";
    | 4  :  S:="IV";
    | 5  :  S:="V";
    | 6  :  S:="VI";
    | 7  :  S:="VII";
    | 8  :  S:="VIII";
    | 9  :  S:="IX";
    | 10 :  S:="X";
    | 11 :  S:="XI";
    | 12 :  S:="XII";
    ELSE
            S:="???";
    END;
    Str.Copy(R,S);
END getRomanStr;

PROCEDURE getDignityName (id:dignitytype;VAR R:ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | exaltation : S:="Exaltation";
    | domicile   : S:="Domicile";
    | exil       : S:="Exil";
    | chute      : S:="Chute";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getDignityName;

PROCEDURE getMotionAbrev (id:sens; VAR R : ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | direct :            S:="--D->";
    | retrograde :        S:="<-R--";
    | stationretrograde : S:="<-r-*";
    | stationdirect :     S:="*-d->";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getMotionAbrev;

PROCEDURE getSeasonName (id:seasontype;VAR R:ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE id OF
    | printempsT : S:="Printemps";
    | eteT       : S:="Et";
    | automneT   : S:="Automne";
    | hiverT     : S:="Hiver";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Copy(R,S);
END getSeasonName;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE isPlanete (id:objet) : BOOLEAN;
BEGIN
    CASE id OF
    | soleil,lune,
      mercure,venus,mars,
      jupiter,saturne,
      uranus,neptune,pluton :
        RETURN TRUE;
    | as,mc,ds,fc,
      lilith,dragonhead,dragontail,blacksun :
        RETURN FALSE;
    ELSE ; (* //ISO fixed silly requirement *) END;
END isPlanete;

PROCEDURE isFictitious (id:objet) : BOOLEAN;
BEGIN
    CASE id OF
    | soleil,lune,
      mercure,venus,mars,
      jupiter,saturne,
      uranus,neptune,pluton :
        RETURN FALSE;
    | as,mc,ds,fc,
      lilith,dragonhead,dragontail,blacksun :
        RETURN TRUE;
    ELSE ; (* //ISO fixed silly requirement *) END;
END isFictitious;

PROCEDURE isNonsense (id:objet) : BOOLEAN;
BEGIN
    CASE id OF
    | soleil,lune,
      mercure,venus,mars,
      jupiter,saturne,
      uranus,neptune,pluton,
      as,mc,ds,fc :
        RETURN FALSE;
    | lilith,dragonhead,dragontail,blacksun :
        RETURN TRUE;
    ELSE ; (* //ISO fixed silly requirement *) END;
END isNonsense;

PROCEDURE isAngle (id:objet) : BOOLEAN;
BEGIN
    CASE id OF
    | soleil,lune,
      mercure,venus,mars,
      jupiter,saturne,
      uranus,neptune,pluton,
      lilith,dragonhead,dragontail,blacksun :
        RETURN FALSE;
    | as,mc,ds,fc :
        RETURN TRUE;
    ELSE ; (* //ISO fixed silly requirement *) END;
END isAngle;

PROCEDURE isDSFC (id:objet) : BOOLEAN;
BEGIN
    CASE id OF
    | soleil,lune,
      mercure,venus,mars,
      jupiter,saturne,
      uranus,neptune,pluton,
      lilith,dragonhead,dragontail,blacksun,as,mc :
        RETURN FALSE;
    | ds,fc :
        RETURN TRUE;
    ELSE ; (* //ISO fixed silly requirement *) END;
END isDSFC;

PROCEDURE isASMC (id:objet) : BOOLEAN;
BEGIN
    CASE id OF
    | soleil,lune,
      mercure,venus,mars,
      jupiter,saturne,
      uranus,neptune,pluton,
      lilith,dragonhead,dragontail,blacksun,ds,fc :
        RETURN FALSE;
    | as,mc :
        RETURN TRUE;
    ELSE ; (* //ISO fixed silly requirement *) END;
END isASMC;

PROCEDURE isAS   (id:objet) : BOOLEAN;
BEGIN
    CASE id OF
    | soleil,lune,
      mercure,venus,mars,
      jupiter,saturne,
      uranus,neptune,pluton,
      lilith,dragonhead,dragontail,blacksun,mc,ds,fc :
        RETURN FALSE;
    | as :
        RETURN TRUE;
    ELSE ; (* //ISO fixed silly requirement *) END;
END isAS;

PROCEDURE isPlaneteLenteTR (id:objet) : BOOLEAN;
BEGIN
    CASE id OF
    | jupiter,saturne,
      uranus,neptune,pluton :
        RETURN TRUE;
    | soleil,lune,
      mercure,venus,mars,
      as,mc,ds,fc,
      lilith,dragonhead,dragontail,blacksun :
        RETURN FALSE;
    ELSE ; (* //ISO fixed silly requirement *) END;
END isPlaneteLenteTR;

PROCEDURE isPlaneteRapideTR (id:objet) : BOOLEAN;
BEGIN
    CASE id OF
    | soleil,lune,
      mercure,venus,mars :
        RETURN TRUE;
    | jupiter,saturne,
      uranus,neptune,pluton,
      as,mc,ds,fc,
      lilith,dragonhead,dragontail,blacksun :
        RETURN FALSE;
    ELSE ; (* //ISO fixed silly requirement *) END;
END isPlaneteRapideTR;

(* check for objects which are always opposite or whose aspects are useless *)

PROCEDURE isAxisOrUseless (id1,id2:objet) : BOOLEAN;
BEGIN
    CASE id1 OF
    | soleil,lune,
      mercure,venus,mars,
      jupiter,saturne,
      uranus,neptune,pluton,
      lilith,blacksun :
        RETURN FALSE;
    | as:
        (* was RETURN (id2=ds) *)
        CASE id2 OF
        | ds, mc, fc :
            RETURN TRUE;
        ELSE
            RETURN FALSE;
        END;
    | mc:
        CASE id2 OF
        | fc, as, ds :
            RETURN TRUE;
        ELSE
            RETURN FALSE;
        END;
    | ds:
        CASE  id2 OF
        | as, mc, fc :
            RETURN TRUE;
        ELSE
            RETURN FALSE;
        END;
    | fc:
        CASE id2 OF
        | mc, as, ds :
            RETURN TRUE;
        ELSE
            RETURN FALSE;
        END;
    | dragonhead:
        RETURN (id2=dragontail);
    | dragontail:
        RETURN (id2=dragonhead);
    ELSE ; (* //ISO fixed silly requirement *) END;
END isAxisOrUseless;

(* ------------------------------------------------------------ *)

PROCEDURE getEcart (id:aspecttype) : CARDINAL;
VAR
    n:CARDINAL;
BEGIN
    CASE id OF
    | conjonction : n:=0;
    | opposition  : n:=180;
    | carre       : n:=90;
    | trigone     : n:=120;
    | sextile     : n:=60;
    | semicarre   : n:=45;
    | semisextile : n:=30;
    | quinconce   : n:=150;
    | sesquicarre : n:=135;
    ELSE ; (* //ISO fixed silly requirement *) END;
    RETURN n;
END getEcart;

(* ------------------------------------------------------------ *)

PROCEDURE getGroupClass (group1,group2:CARDINAL):grouptype;
VAR
    g:grouptype;
BEGIN
    CASE group1 OF
    | 1 :
        CASE group2 OF
        | 1 : g:=group11;
        | 2 : g:=group12;
        | 3 : g:=group13;
        ELSE ; (* //ISO fixed silly requirement *) END;
    | 2 :
        CASE group2 OF
        | 1 : g:=group12;
        | 2 : g:=group22;
        | 3 : g:=group23;
        ELSE ; (* //ISO fixed silly requirement *) END;
    | 3 :
        CASE group2 OF
        | 1 : g:=group13;
        | 2 : g:=group23;
        | 3 : g:=group33;
        ELSE ; (* //ISO fixed silly requirement *) END;
    ELSE ; (* //ISO fixed silly requirement *) END;
    RETURN g;
END getGroupClass;

(* ------------------------------------------------------------ *)

PROCEDURE getDignity (id:objet;sign:signtype):dignitytype;
VAR
    d : dignitytype;
BEGIN
    d:=aucunedignite; (* default *)
    CASE id OF
    | soleil :
        CASE sign OF
        | lion          : d:= domicile;
        | verseau       : d:= exil;
        | belier        : d:= exaltation;
        | balance       : d:= chute;
        ELSE ; (* //ISO fixed silly requirement *) END;
    | lune :
        CASE sign OF
        | cancer        : d:= domicile;
        | capricorne    : d:= exil;
        | taureau       : d:= exaltation;
        | scorpion      : d:= chute;
        ELSE ; (* //ISO fixed silly requirement *) END;
    | mercure :
        CASE sign OF
        | gemeaux       : d:= domicile;
        | sagittaire    : d:= exil;
        | vierge        : d:= exaltation;
        | poissons      : d:= chute;
        ELSE ; (* //ISO fixed silly requirement *) END;
    | venus :
        CASE sign OF
        | taureau       : d:= domicile;
        | scorpion      : d:= exil;
        | poissons      : d:= exaltation;
        | vierge        : d:= chute;
        ELSE ; (* //ISO fixed silly requirement *) END;
    | mars :
        CASE sign OF
        | belier        : d:= domicile;
        | balance       : d:= exil;
        | capricorne    : d:= exaltation;
        | cancer        : d:= chute;
        ELSE ; (* //ISO fixed silly requirement *) END;
    | jupiter :
        CASE sign OF
        | sagittaire    : d:= domicile;
        | gemeaux       : d:= exil;
        | cancer        : d:= exaltation;
        | capricorne    : d:= chute;
        ELSE ; (* //ISO fixed silly requirement *) END;
    | saturne :
        CASE sign OF
        | capricorne    : d:= domicile;
        | cancer        : d:= exil;
        | balance       : d:= exaltation;
        | belier        : d:= chute;
        ELSE ; (* //ISO fixed silly requirement *) END;
    | uranus :
        CASE sign OF
        | verseau       : d:= domicile;
        | lion          : d:= exil;
        | scorpion      : d:= exaltation;
        | taureau       : d:= chute;
        ELSE ; (* //ISO fixed silly requirement *) END;
    | neptune :
        CASE sign OF
        | poissons      : d:= domicile;
        | vierge        : d:= exil;
        | lion          : d:= exaltation;
        | verseau       : d:= chute;
        ELSE ; (* //ISO fixed silly requirement *) END;
    | pluton :
        CASE sign OF
        | scorpion      : d:= domicile;
        | taureau       : d:= exil;
        | sagittaire    : d:= exaltation;
        | gemeaux       : d:= chute;
        ELSE ; (* //ISO fixed silly requirement *) END;
    ELSE ; (* //ISO fixed silly requirement *) END;
    RETURN d;
END getDignity;

(* for those two planets which have two domiciles exil *)

PROCEDURE getDignityAlt (id:objet;sign:signtype):dignitytype;
VAR
    d : dignitytype;
BEGIN
    d:=aucunedignite; (* default *)
    CASE id OF
    | mercure :
        CASE sign OF
        | vierge        : d:= domicile;
        | poissons      : d:= exil;
        ELSE ; (* //ISO fixed silly requirement *) END;
    | venus :
        CASE sign OF
        | balance       : d:= domicile;
        | belier        : d:= exil;
        ELSE ; (* //ISO fixed silly requirement *) END;
    ELSE ; (* //ISO fixed silly requirement *) END;
    RETURN d;
END getDignityAlt;

PROCEDURE getMaitrise (sign:signtype):objet; (* ArcDomicile *)
VAR
    r:objet;
BEGIN
    CASE sign OF
    | belier     : r:=mars;
    | taureau    : r:=venus;
    | gemeaux    : r:=mercure;
    | cancer     : r:=lune;
    | lion       : r:=soleil;
    | vierge     : r:=mercure;
    | balance    : r:=venus;
    | scorpion	 : r:=pluton;
    | sagittaire : r:=jupiter;
    | capricorne : r:=saturne;
    | verseau    : r:=uranus;
    | poissons   : r:=neptune;
    ELSE ; (* //ISO fixed silly requirement *) END;
    RETURN r;
END getMaitrise;

PROCEDURE getSaison (sign:signtype):seasontype;
VAR
    r:seasontype;
BEGIN
    CASE sign OF
    | belier     : r:=printempsT;
    | taureau    : r:=printempsT;
    | gemeaux    : r:=printempsT;
    | cancer     : r:=eteT;
    | lion       : r:=eteT;
    | vierge     : r:=eteT;
    | balance    : r:=automneT;
    | scorpion	 : r:=automneT;
    | sagittaire : r:=automneT;
    | capricorne : r:=hiverT;
    | verseau    : r:=hiverT;
    | poissons   : r:=hiverT;
    ELSE ; (* //ISO fixed silly requirement *) END;
    RETURN r;
END getSaison;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* parse file,hexoffset,hexhash of file,hexoffset *)

PROCEDURE validEntry (S:ARRAY OF CHAR;
                      VAR R:ARRAY OF CHAR;VAR offset:LONGCARD );
CONST
    equal   = "=";
VAR
    p:CARDINAL;
    ok:BOOLEAN;
BEGIN
    p:=Str.CharPos(S,equal); (* CANNOT be max(cardinal) ! *)
    Str.Slice(R,S,0,p); (* filename *)
    Str.Delete(S,0,p+1);
    offset:=Str.StrToCard(S,16,ok);
END validEntry;

TYPE
    indexentry = RECORD
        f8e3 : ARRAY [0..11] OF CHAR; (* 8+1+3 *)
        offset:LONGCARD;
        hash:CARDINAL; (* try and speed up string lookup *)
    END;

CONST
    firstIndex = 1;
    maxIndex   = 256; (* 168 needed for now... but for future expansion *)
    nullchar   = CHR(0);
    hashK      = MAX(CARDINAL);
VAR
    lastIndex  : CARDINAL;
    indexarray : ARRAY [firstIndex..maxIndex] OF indexentry;

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead; (* was 16 *)
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
VAR
    ioBuffer        : ARRAY [firstBufferByte..lastBufferByte] OF BYTE;

PROCEDURE readIndexFile (file:ARRAY OF CHAR) : BOOLEAN;
VAR
    hnd:FIO.File;
    S : str128;
    state : (ok,problem);
    rc:BOOLEAN;
    c:BYTE;
    i : CARDINAL;
BEGIN
    state := ok;
    lastIndex:=firstIndex; (* hence [1.. *)
    hnd := FIO.OpenRead(file);
    FIO.AssignBuffer(hnd,ioBuffer);
    FIO.EOF:=FALSE; (* safety ! *)
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hnd,S);
        IF (S[0]=nullchar) AND FIO.EOF THEN EXIT; END; (* weird hack ! *)
        LtrimBlanks(S);
        RtrimBlanks(S);
        CASE S[0] OF
        | nullchar :
            (* ignore empty string if any *)
        ELSE
            IF lastIndex > maxIndex THEN
                state:=problem;
                EXIT;
            END;
            validEntry(S,indexarray[lastIndex].f8e3,indexarray[lastIndex].offset); (* assume OK *)
            indexarray[lastIndex].hash:=Lib.HashString(indexarray[lastIndex].f8e3,hashK);
            INC(lastIndex);
        END;
    END;
    FIO.Close(hnd);
    CASE state OF
    | ok :
        IF lastIndex=firstIndex THEN RETURN FALSE; END;
        DEC(lastIndex); (* was at last+1 *)
        RETURN TRUE;
    | problem :
        RETURN FALSE;
    ELSE ; (* //ISO fixed silly requirement *) END;
END readIndexFile;

(* ------------------------------------------------------------ *)

(* find F8.E3 offset in file *)

PROCEDURE findIndexFileOffset (S,ext:ARRAY OF CHAR;VAR offset:LONGCARD):BOOLEAN;
VAR
    hashwanted,i:CARDINAL;
    filename:str128;
BEGIN
    Str.Concat(filename,S,ext);
    UpperCase(filename);
    hashwanted:=Lib.HashString(filename,hashK);
    i:=firstIndex;
    LOOP
        IF indexarray[i].hash=hashwanted THEN
            IF same(indexarray[i].f8e3,filename) THEN (* added safety in case hash collision *)
                offset:=indexarray[i].offset;
                RETURN TRUE;
            END;
        END;
        INC(i);
        IF i > lastIndex THEN
            offset:=0; (* just in case we forget to test result ! *)
            RETURN FALSE;
        END;
    END;
END findIndexFileOffset;

(* ------------------------------------------------------------ *)

CONST
    blank      = " ";
    underscore = "_";
    tabchar    = CHR(9);
    tabexpanded= "        ";
VAR
    hugestring : str4096; (* should be enough : longest paragraph is about 4050 chars *)

(* find :::XXXXXXXXXX in file *)

PROCEDURE findFileEntry (hnd:FIO.File;offset:LONGCARD;S:ARRAY OF CHAR) : BOOLEAN;
VAR
    section:str128;
    rc:BOOLEAN;
    i : CARDINAL;
    li:LONGCARD;
    c:BYTE;
BEGIN
    Str.Concat(section,sEntry,S);
    UpperCase(section);
    LtrimBlanks(section);
    RtrimBlanks(section);
    ReplaceChar(section,blank,underscore); (* do not use "" here for it would truncate string ! here we can replace AFTER trim *)

    FIO.Seek(hnd,offset);
    FIO.EOF:=FALSE;
    LOOP
        FIO.RdStr(hnd,hugestring);
        IF (hugestring[0]=nullchar) AND FIO.EOF THEN EXIT; END; (* weird hack ! *)
        IF same(hugestring,sEOF) THEN EXIT; END; (* *** *)
        (* ignore ###xxx but check for :::xxx *)
        IF hugestring[0]=sEntry[0] THEN (* speed up a little *)
            IF Str.Pos(hugestring,sEntry)=0 THEN
                IF same(hugestring,section) THEN RETURN TRUE; END;
            END;
        END;
    END;
    RETURN FALSE;
END findFileEntry;

(*
   read string after :::XXXXXXXXXX
   MUST be used immediately after findFileEntry because of CAinit
*)

PROCEDURE readFileEntry(hnd:FIO.File;VAR S:ARRAY OF CHAR) : BOOLEAN;
VAR
    i : CARDINAL;
    c : BYTE;
BEGIN
    FIO.EOF:=FALSE;
    FIO.RdStr(hnd,hugestring);
    IF (hugestring[0]=nullchar) AND FIO.EOF THEN RETURN FALSE; END; (* weird hack ! *)
    IF hugestring[0]=sEOF[0] THEN
        IF same(hugestring,sEOF) THEN RETURN FALSE; END; (* *** *)
    END;
    IF hugestring[0]=sEntry[0] THEN
        IF Str.Pos(hugestring,sEntry)=0 THEN RETURN FALSE; END; (* :::xxx *)
    END;

    (* found ! now replace first occurrence of TAB for paragraph *)
    Str.Subst(hugestring,tabchar,tabexpanded);
    Str.Copy(S,hugestring);
    RETURN TRUE;
END readFileEntry;

(* ------------------------------------------------------------ *)

(* find :::_c_h_d_planet1_&_planet2 in file *)

PROCEDURE findFileEntryAspect (hnd:FIO.File;offset:LONGCARD;
                               ABREV,P1,P2:ARRAY OF CHAR) : BOOLEAN;
VAR
    rc:BOOLEAN;
    i : CARDINAL;
    li:LONGCARD;
    c:BYTE;
    pos1,pos2:CARDINAL;
BEGIN
    (* clean things up *)
    UpperCase(ABREV);
    UpperCase(P1);
    LtrimBlanks(P1);
    RtrimBlanks(P1);
    ReplaceChar(P1,blank,underscore); (* do not use "" here for it would truncate string ! *)
    UpperCase(P2);
    LtrimBlanks(P2);
    RtrimBlanks(P2);
    ReplaceChar(P2,blank,underscore); (* do not use "" here for it would truncate string ! *)

    FIO.Seek(hnd,offset);
    FIO.EOF:=FALSE;
    LOOP
        FIO.RdStr(hnd,hugestring);
        IF (hugestring[0]=nullchar) AND FIO.EOF THEN EXIT; END; (* weird hack ! *)
        IF same(hugestring,sEOF) THEN EXIT; END; (* *** *)
        (* ignore ###xxx but check for :::xxx *)
        IF hugestring[0]=sEntry[0] THEN (* speed up a little *)
            IF Str.Pos(hugestring,sEntry)=0 THEN
                IF Str.Pos(hugestring,stEt) # MAX(CARDINAL) THEN
                    IF Str.Pos(hugestring,ABREV) # MAX(CARDINAL) THEN
                        pos1:=Str.Pos(hugestring,P1);
                        IF pos1 # MAX(CARDINAL) THEN
                            pos2:=Str.Pos(hugestring,P2);
                            IF pos2 # MAX(CARDINAL) THEN
                                IF pos1 # pos2 THEN
                                    RETURN TRUE; (* okay, we may leave here *)
                                ELSE
                                    (*
                                    well, seems that P1=P2 !
                                    remove its first occurrence then search it again
                                    *)
                                    Str.Subst(hugestring,P1,"");
                                    pos2:=Str.Pos(hugestring,P2);
                                    IF pos2 # MAX(CARDINAL) THEN
                                        RETURN TRUE; (* here too we may leave *)
                                    END;
                                END;
                                (* go on searching... go on searching... *)
                            END;
                        END;
                    END;
                END;
            END;
        END;
    END;
    RETURN FALSE;
END findFileEntryAspect;

(* ------------------------------------------------------------ *)

BEGIN
    FIO.IOcheck:=FALSE; (* avoid nasty problems *)

END QD_Names.

