(* ---------------------------------------------------------------
Title         Q&D T.U. (France)
Author        PhG
Overview      find correction required to get Universal Time for France
Usage         see help
Notes         very, very, very quick & dirty... :-(
Bugs
Wish List

--------------------------------------------------------------- *)

MODULE TU;

IMPORT FIO;
IMPORT Str;
IMPORT Lib;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey,  chkJoker,
isOption, GetOptIndex, GetLongCard, GetString, CharCount, same, cleantabs,
aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
Elapsed,cmdSHOW,Animation,Work,pathtype;

FROM QD_Sky IMPORT int, frac, mod, DateToJD, JDtoDate, JDtoJD0H,
JDtoWeekday, JDtoT, isDigit, MatchFormat, getLastCard, stripCard,
getCard, parseDate, parseTime, parseLongitude, DateTimeToJD,
CalcBornes, chkZoneFR, ToggleSign, rangeWarning;


FROM QD_Names IMPORT ProgCopyright;

FROM SYSTEM IMPORT BYTE;

FROM IO IMPORT WrStr, WrLn;

(* ------------------------------------------------------------ *)

CONST
CONST
    ProgEXEname   = "TU";
    ProgTitle     = "Q&D T.U. (Win32)";
    ProgVersion   = "v0.9 (v1.0e DOS)";
    ProgBanner    = ProgTitle+" "+ProgVersion+" "+ProgCopyright;
CONST
    extEXE        = ".EXE"; (* upper case *)
    extRSC        = ".RSC";
CONST
    errNone          = 0;
    errHelp          = 1;
    errOption        = 2;
    errParmOverflow  = 3;
    errMissingDate   = 4;
    errMissingTime   = 5;
    errDateFmt       = 6;
    errTimeFmt       = 7;
    errLambdaFmt     = 8;
    errRscNotFound   = 9;
    errCountryNotFound=10;
    errRscLayout     = 11;
    errImpossibleLegalHour = 12;


VAR
    DEBUG         : BOOLEAN;
CONST
    nullchar    = CHR(0);
    semicolon   = ";";
    pound       = "#";
    openbracket = "[";
    closebracket= "]";
    quote       = "'";
    doublequote = '"';
    blank       = " ";
    tab         = CHR(9);
    coma        = ",";
CONST
    alpha   = "abcdefghijklmnopqrstuvwxyz";
    special = "";
    letters = alpha+special;
CONST
    msgAtLine      = ", ligne ";
    FRsection      = "france"; (* lower case required here *)
    msgData        = "Donnes fournies : ";
    sep            = "  ";
    codeParis      = "P";
    codeLocal      = "L";
    msgWarn        = "Attention ! Par prcaution, recommencez en indiquant aussi la longitude !";
    pleasewait     = "Patience, recherche en cours... ";

(* ------------------------------------------------------------ *)

PROCEDURE eyeCandy(command:CARDINAL);
VAR
    i : CARDINAL;
BEGIN
    Animation(command); (* elapsed time would be silly here for search can restart *)
END eyeCandy;

VAR
  PATIENCE : BOOLEAN;

PROCEDURE StopCandy();
BEGIN
    eyeCandy(cmdStop);
    video(pleasewait,FALSE);
    PATIENCE:=FALSE;
END StopCandy;

PROCEDURE StartCandy();
BEGIN
    video(pleasewait,TRUE);
    eyeCandy(cmdInit);
    eyeCandy(cmdSHOW);
    PATIENCE:=TRUE;
END StartCandy;

PROCEDURE BusyCandy ();
BEGIN
  eyeCandy(cmdShow);
END BusyCandy;

(* ------------------------------------------------------------ *)

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
    nl     = 015C+012C; (* CHR(13)+CHR(10); *)
VAR
    S : str256;
BEGIN
  IF PATIENCE=TRUE THEN StopCandy();  END;
    CASE e OF
    | errHelp :
        (*
               00000000011111111112222222222333333333344444444445555555555666666666677777777778
               1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
        *)
        WrLn;
        WrStr("Titre   : "+ProgBanner+nl);
        WrLn;
        WrStr("Syntaxe : "+ProgEXEname+" <date> <heure lgale> [longitude]"+nl);
        WrLn;
        WrStr("Utilit : si possible, trouver la correction horaire requise pour obtenir"+nl);
        WrStr("          l'heure en Temps Universel (T.U.)  partir de l'heure lgale"+nl);
        WrLn;
        WrStr("Option  : /u pour ignorer la restriction  l'intervalle [1600..2200]"+nl);
        WrLn;
        WrStr("Exemple : "+ProgEXEname+" 20/9/16 7h"+nl);
    | errOption :
        Str.Concat(S,"L'option ",einfo);
        Str.Append(S," est illgale !");
    | errParmOverflow :
        Str.Concat(S,"Le paramtre ",einfo);
        Str.Append(S," est de trop !");
    | errMissingDate :
        S := "Les date et heure lgale sont indispensables !";
    | errMissingTime :
        S := "L'heure lgale est indispensable !";
    | errDateFmt :
        S := "La date est incorrecte !";
    | errTimeFmt :
        S := "L'heure est incorrecte !";
    | errLambdaFmt :
        S := "La longitude est incorrecte !";
    | errRscNotFound :
        Str.Concat(S,"Le tableau des corrections (",einfo);
        Str.Append(S,") est introuvable !");
    | errCountryNotFound :
        Str.Concat(S,"Le pays demand (",einfo);
        Str.Append(S,") ne figure pas dans le tableau !");
    | errRscLayout :
        Str.Concat(S,"Le format du tableau des corrections (",einfo);
        Str.Append(S,") semble incorrect :"+nl);
        Str.Append(S,"la section du pays demand est vide, ou Last BS read < JDL !");
    | errImpossibleLegalHour :
        S:= "Cette heure lgale est impossible (JDL < BI) !";


    ELSE
        S := "This is illogical, Captain !";
    END;
    IF (e <> errNone) AND (e <> errHelp) THEN
        WrLn;
        WrStr(ProgEXEname+" : ");
        WrStr(S);
        WrLn;
    END;

    (* FPUalert(); *)

    (* Lib.SetReturnCode(SHORTCARD(e)); *)
    HALT (e);
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE CleanStr (VAR S : ARRAY OF CHAR);
BEGIN
    LtrimBlanks(S);
    Ltrim(S,quote);
    Ltrim(S,doublequote);
    RtrimBlanks(S);
    Rtrim(S,quote);
    Rtrim(S,doublequote);
END CleanStr;

PROCEDURE CleanSectionStr (VAR S : ARRAY OF CHAR );
BEGIN
    Ltrim(S,openbracket);
    Rtrim(S,closebracket);
    CleanStr(S);
END CleanSectionStr;

PROCEDURE Filter (legal:ARRAY OF CHAR;VAR S : ARRAY OF CHAR);
VAR
    R : str128;
    i : CARDINAL;
BEGIN
    Str.Copy(R,"");
    FOR i := 0 TO Str.Length(S)-1 DO
        IF Belongs(legal,S[i]) THEN
            Str.Append(R,S[i]);
        END;
    END;
    Str.Copy(S,R);
END Filter;

(*
;
; corrections horaires pour le Temps Universel (TU)
;

[France]

; tableau des corrections horaires pour la France
; format de la correction : 1 1H 1H1 -1 -1H -1H30
;  du (T.U)      au (T.U.)  HL=T.U.plus
; "aaaa:jj-mm,hh/aaaa:jj-mm,hh=c"
; vrifier aprs 97 en fonction des ordres des crapules eurocrates
; pour l'heure, application du dernier dimanche de mars / octobre

"0000:00-00,00/1891:15-03,00=l"
"1891:15-03,00/1911:11-03,00=p"
"1911:11-03,00/1916:14-06,23=0"
"1916:14-06,23/1916:01-10,23=1"
"1916:01-10,23/1917:24-03,23=0"
......

"2001:25-03,01/2001:28-10,01=2"
"2001:28-10,01/2002:31-03,01=1"
"2002:31-03,01/2002:27-10,01=2"
"2002:27-10,01/9999:31-12,01=1"

;------------------------------------------------

[Principaut d'Andorre]

; idem France

"0000:00-00,00/1891:15-03,00=l"
"1891:15-03,00/1911:11-03,00=p"
"1911:11-03,00/1916:14-06,23=0"

*)

(* ------------------------------------------------------------ *)

(* synonyms could be in INI file... *)

PROCEDURE FindAlias (VAR S : ARRAY OF CHAR );
CONST
    firstAlias = 1;
    lastAlias  = 3;
TYPE
    aliases = ARRAY [1..3] OF str128;
CONST
    syn = aliases{"France,FRA,FR,F",
                  "Principaut de Monaco,Monaco,MON",
                  "Principaut d'Andorre,Andorre,AND"};
VAR
    i,j : CARDINAL;
    abrev : str128;
BEGIN
    FOR i:=firstAlias TO lastAlias DO
        FOR j := 0 TO CharCount(syn[i],coma) DO
            Str.ItemS(abrev,syn[i],coma,j);
            LowerCase(abrev);
            Filter(letters,abrev);
            IF same(S,abrev) THEN
                Str.ItemS(S,syn[i],coma,0);
                LowerCase(S);
                Filter(letters,S);
                RETURN;
            END;
        END;
    END;
END FindAlias;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead; (* was 24 *)
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
VAR
    ioBuffer        : ARRAY [firstBufferByte..lastBufferByte] OF BYTE;

(* ------------------------------------------------------------ *)

PROCEDURE FindDeltaTU (file,country:ARRAY OF CHAR;
                       lambdaHere:BOOLEAN;lambdaDeg:LONGREAL;
                       day,month,year,hh,mm:CARDINAL);
VAR
    hnd            : FIO.File;
    S              : str256;
    state          : (waiting,countryFound,borneFound,nextborneFound,problem);
    currline       : CARDINAL;
    countryname    : str128;
    countrysection : str128;
    legaljulian    : LONGREAL; (* julian day for legal date/time *)
    ok             : BOOLEAN;
    needZoneFR     : BOOLEAN; (* en zone occupe, -2H *)
    needmoment     : BOOLEAN;

    borneS         : str128;
    borneinf       : LONGREAL;
    bornesup       : LONGREAL;
    correctionTU   : str128; (* really oversized ! *)

    nextborneS     : str128;
    nextborneinf   : LONGREAL;
    nextbornesup   : LONGREAL;
    correctionTUalt: str128; (* really oversized ! *)

    deltaTU        : str128;
    deltaTUalt     : str128; (* alternate if AFTER hour change *)
    info           : str128;
    i              : CARDINAL;
    c              : CHAR;
BEGIN
  StartCandy();

    Str.Copy(countrysection,country);
    LowerCase(countrysection);
    Filter(letters,countrysection);
    FindAlias(countrysection); (* now useless : always France *)

    legaljulian := DateTimeToJD(day,month,year,hh,mm);

    state := waiting;
    currline := 0;

    hnd := FIO.OpenRead(file);
    FIO.AssignBuffer(hnd,ioBuffer);
    FIO.EOF:=FALSE; (* safety ! *)
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hnd,S);
  BusyCandy();
        IF (S[0]=nullchar) AND FIO.EOF THEN EXIT; END; (* weird hack ! *)


        INC(currline);
        CleanStr(S); (* remove leading and trailing blanks, tabs, quotes *)
        CASE S[0] OF
        | nullchar,semicolon,pound :
            (* ignore empty string or comment *)
        | openbracket :
            CASE state OF
            | waiting :
                CleanSectionStr(S);
                Str.Copy(countryname,S);
                LowerCase(S);
                Filter(letters,S);
                IF same(countrysection,S) THEN
                    state := countryFound;
                END;
            | borneFound : (* here we need nextborne but nada for we were already at the end of data : so fake it *)
                Str.Copy(nextborneS,borneS);
                nextborneinf:=bornesup;
                nextbornesup:=bornesup;
                Str.Copy(correctionTUalt,correctionTU);
                state:=nextborneFound; (* fake ! *)
                EXIT;
            ELSE (* either empty section or last BS read < JDL ! *)
                state := problem;
                EXIT;
            END;
        ELSE
            CASE state OF
            | waiting : (* ignore *)

            | countryFound :
                CalcBornes(S,lambdaDeg,borneinf,bornesup,correctionTU);
                IF  legaljulian <= bornesup THEN (* JD <= BS i.e. BS >= JD *)
                    Str.Copy(borneS,S);
                    state := borneFound;
                END;
            | borneFound : (* we need nextborne just in case *)
                CalcBornes(S,lambdaDeg,nextborneinf,nextbornesup,correctionTUalt);
                Str.Copy(nextborneS,S);
                state := nextborneFound;
                EXIT;
            ELSE (* oops ! *)
                state := problem;
                EXIT;
            END;
        END;
    END;
    FIO.Close(hnd);
  StopCandy();

    (* here, S contains latest offending line *)
    CASE state OF
    | waiting :
        abort(errCountryNotFound,country);
    | problem,countryFound,borneFound :
        Str.CardToStr(LONGCARD(currline),S,10,ok);
        Str.Prepend(S,msgAtLine);
        Str.Prepend(S,file);
        abort(errRscLayout,S);
    ELSE ; (* //ISO fixed silly requirement *) END;
    (* here nextborneFound ok : borneS and nextborneS for diag *)

    (*
            OLDBS        <   JD        <= BS
            OLDBS            JD   BI      BS   heure lgale impossible !
            OLDBS   BI       JD           BS
       BI   OLDBS            JD           BS
    *)

    (* impossible legal hour (JDL < BI) -- if JDL=BI, changement effectu *)
    IF legaljulian < borneinf THEN abort(errImpossibleLegalHour,"");END;

    (* maybe we should check that JD < (BS-correction) ? later... *)

    (*
               OLDBI   JD   OLDBS
                            BI      BS
    *)
    (*
        OLDBI   JD  <=   OLDBS   BS
                              BI         oldc is OK
                    BI                   oldc is OK
             BI                          ask before or after
    *)

    (* zone libre ou zone occupe ? *)
    needZoneFR  := chkZoneFR(day,month,year,hh,mm,countrysection);

    ToggleSign(correctionTU); (* always *)
    Str.Copy(deltaTU,correctionTU); (* with sign ok now *)

    Str.Copy(info,deltaTU);

    IF legaljulian < nextborneinf THEN
        needmoment:=FALSE;
        IF needZoneFR THEN
            Str.Copy(correctionTUalt,"-2H"); (* zone occupe *)
            Str.Append(info," en zone libre, -2H en zone occupe");
        ELSE
            Str.Copy(correctionTUalt,"");
            (* errinfo$="" *)
        END;
    ELSE
        ToggleSign (correctionTUalt);
        Str.Copy(deltaTUalt,correctionTUalt); (* with sign ok now *)
        needmoment:=TRUE;
        IF needZoneFR THEN
            Str.Append(info," avant le changement d'heure, ");
            Str.Append(info,deltaTUalt);
            Str.Append(info," aprs, et -2H en zone occupe");
        ELSE
            Str.Append(info," avant le changement d'heure, ");
            Str.Append(info,deltaTUalt);
            Str.Append(info," aprs");
        END;
    END;

    WrLn;
    WrStr(info);
    WrLn;
    IF lambdaHere=FALSE THEN
        IF (same(deltaTU,codeLocal)) OR (same(deltaTU,codeParis)) THEN
            WrLn;
            WrStr(msgWarn);
            WrLn;
        END;
    END;
    IF DEBUG THEN
        WrLn;
        WrStr(borneS);
        WrLn;
        WrStr(nextborneS);
        WrLn;
    END;
END FindDeltaTU;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    flagIgnoreRangeLimits:BOOLEAN;
    parmcount : CARDINAL;
    i         : CARDINAL;
    opt       : CARDINAL;
    S         : pathtype;
    R         : pathtype;
    rscfile   : pathtype;
    state     : (waiting,gotDate,gotTime,gotLongitude);
    date      : str128;
    time      : str128;
    longitude : str128;
    country   : str128;
    day,month,year : CARDINAL;
    hh,mm          : CARDINAL;
    lambdaHere     : BOOLEAN;
    lambdaDeg      : LONGREAL;
    lambdaSec      : LONGINT;

    firsttime      : BOOLEAN;
BEGIN
    (* Lib.DisableBreakCheck();     *)
    FIO.IOcheck := FALSE;
    DEBUG := FALSE;
  PATIENCE:=FALSE;



    flagIgnoreRangeLimits:=FALSE;

    Str.Copy(country,FRsection); (* always France *)
    lambdaHere:=FALSE;
    lambdaDeg := 0.0;
    lambdaSec := 0;


    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,""); END;

    state:=waiting;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R)=TRUE THEN
            opt := GetOptIndex(R,"?"+delim+"H"+delim+"HELP"+delim+
                                 "DEBUG"+delim+
                                 "U"+delim+"UNSAFE"
                                 );
            CASE opt OF
            | 1,2,3 :
                abort(errHelp,"");
            | 4 :
                DEBUG := TRUE;
            | 5,6:
                flagIgnoreRangeLimits:=TRUE;
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            CASE state OF
            | waiting :
                Str.Copy(date,S);
                state := gotDate;
            | gotDate :
                Str.Copy(time,S);
                state := gotTime;
            | gotTime :
                lambdaHere := TRUE;
                Str.Copy(longitude,S);
                state := gotLongitude;
            | gotLongitude :
                abort(errParmOverflow,S);
            ELSE ; (* //ISO fixed silly requirement *) END;
        END;
    END;
    IF state = waiting THEN abort(errMissingDate,""); END; (* could be errHelp *)
    IF state = gotDate THEN abort(errMissingTime,""); END;

    IF parseDate(flagIgnoreRangeLimits,date,day,month,year)=FALSE THEN abort(errDateFmt,""); END;
    IF parseTime(time,hh,mm)=FALSE THEN abort(errTimeFmt,""); END;
    IF lambdaHere THEN
        IF parseLongitude(longitude,lambdaDeg,lambdaSec)=FALSE THEN abort(errLambdaFmt,""); END;
    END;

    Lib.ParamStr(rscfile,0);
    UpperCase(rscfile); (* useless... but for added safety ! *)
    Str.Subst(rscfile,extEXE,extRSC);
    IF FIO.Exists(rscfile)=FALSE THEN abort(errRscNotFound,rscfile); END;

    rangeWarning(flagIgnoreRangeLimits,1,0);

    (* FPUfix(); *)

    WrLn;
    WrStr(ProgBanner);
    WrLn;

    WrLn;
    WrStr(msgData);
    WrStr(date);
    WrStr(sep);
    WrStr(time);
    IF lambdaHere THEN
        WrStr(sep);
        WrStr(longitude);
    END;
    WrLn;

    FindDeltaTU(rscfile,country,lambdaHere,lambdaDeg,day,month,year,hh,mm);

    abort(errNone,"");
END TU.