(*

a Q&D visual Wolfram rules tester

guess my original typo on name had something TO do WITH Wolfram & Hart ;-)
or with keyboard as well...

notes :
- we don't try and be clever for this small hack
  (we just wanted to quickly experiment CA rules,
  and at first it was merely designed TO check rule 30)
- safety checks are limited

to do :
- dump CA state as to check its randomness ?
- dump screen to PCX ?
- handle wider than 3 bits patterns ? 5 bits could do
- toggle infos on screen while running ?
- allow "?" bit state from cli ?

default rule is rule 30
   1. If the cell and its right-hand neighbor are both 0,
      then the new value of the cell is the value of its left-hand neighbor.
   2. Otherwise, the new value of the cell is the opposite
      of the value of its left-hand neighbor.

   3. Another description : x(n+1,i) = x(n,i-1) xor [x(n,i) or x(n,i+1)]

   4. Using patterns :
      current pattern           111 110 101 100 011 010 001 000
      new state for center cell  0   0   0   1   1   1   1   0

XDS does not handle VGA colors correctly : fed up trying to find a fix...
*)

MODULE wolfRule32;

IMPORT Graph;
IMPORT IO;
IMPORT Lib;
IMPORT Str;
IMPORT SYSTEM;
IMPORT FIO;

FROM IO IMPORT WrStr,WrLn;


(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* // QD_ASCII *)

CONST (* same as QD_TEXT *)
    cr            = 15C; (* was CHR(13) : silly XDS *requires* octal here for concatenation !!! *)
    lf            = 12C; (* was CHR(10) *)
    nl            = cr+lf;
    bs            = CHR(8);
CONST
    dash          = "-";
    slash         = "/";
    nullchar      = 0C; (* CHR(0) *)
    tabchar       = 11C; (* silly XDS passion for octal !!! *)
    space         = 40C;
    dot           = ".";
    deg           = "";      (* $f8 *)
    doublequote   = '"';
    quote         = "'";
    colon         = ":";
    percent       = "%";
    vbar          = "|";      (* $7c *)
    blank         = space;
    equal         = "=";
    dquote        = doublequote;
    charnull      = nullchar;
    singlequote   = quote;
    antislash     = "\";
    dollar        = "$";
    star          = "*";
    backslash     = antislash;
    coma          = ",";
    question      = "?";
    underscore    = "_";
    tabul         = tabchar;
    hbar          = vbar;  (* yes, nonsense but kept for historical reasons ! *)
    comma         = coma; (* typo ! *)
    semicolon     = ";";
    diese         = "#";
    pound         = diese;
    openbracket   = "[";
    closebracket  = "]";
    tilde         = "~";
    exclam        = "!";
    openbrace     = "{";
    closebrace    = "}";
    ampersand	  = "&";
CONST
    stardotstar   = star+dot+star;
    dotdot        = dot+dot;
CONST
    lettersUpp = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    lettersLow = "abcdefghijklmnopqrstuvwxyz";
    digits     = "0123456789";
    letters    = lettersUpp;   (* should be remed out *)

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* // QD_Box *)

TYPE
    str80 = ARRAY [0..79] OF CHAR;
    str2  = ARRAY [0..1] OF CHAR;
    str128= ARRAY [0..127] OF CHAR;
    str256= ARRAY [0..255] OF CHAR;
    str1024=ARRAY [0..1023] OF CHAR;
    str2048=ARRAY [0..2047] OF CHAR;
    str16  =ARRAY [0..15] OF CHAR;
    str4096=ARRAY [0..4095] OF CHAR;

CONST
    delim = ",";

CONST
    aR = FIO.readonly;
    aH = FIO.hidden;
    aS = FIO.system;
    aDD= FIO.directory; (* seems XDS already uses aD but error message is not informative ! *)
    aA = FIO.archive;
    everything       = FIO.FileAttr{aR,aH,aS,aDD,aA};
    allfiles         = FIO.FileAttr{aR,aH,aS,aA};

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE map (c : CHAR ;table : ARRAY OF CHAR) : CHAR;
BEGIN
    RETURN table [ ORD (c) ];
END map;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE Rtrim (VAR S : ARRAY OF CHAR; c:CHAR);
VAR
    i,len : CARDINAL;
BEGIN
    len := Str.Length (S);
    i   := len;
    LOOP
        IF i = 0 THEN EXIT; END;
        IF S[i-1] # c THEN EXIT; END;
        DEC (i);
    END;
    Str.Delete (S,i,len-i);
END Rtrim;

(* ------------------------------------------------------------ *)

PROCEDURE UpperCase (VAR S : ARRAY OF CHAR);
BEGIN
    Str.Caps(S);
END UpperCase;

PROCEDURE LowerCase (VAR S : ARRAY OF CHAR);
BEGIN
    Str.Lows(S);
END LowerCase;

(* ------------------------------------------------------------ *)

PROCEDURE ReplaceChar (VAR S : ARRAY OF CHAR;
                       oldchar, newchar : CHAR);
VAR
    p : CARDINAL;
BEGIN
    IF oldchar=newchar THEN RETURN; END; (* silly... *)
    LOOP
        p := Str.CharPos(S,oldchar);
        IF p = MAX(CARDINAL) THEN EXIT; END;
        IF newchar = "" THEN
            Str.Delete(S,p,1);
        ELSE
            S[p]:=newchar;
        END;
    END;
END ReplaceChar;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(*
    use dmpTTX this way :

    continue:=dmpTTX(Z,width,S,TRUE);
    WHILE continue DO
        WrStr(S);WrLn;
        continue:=dmpTTX(Z,width,S,FALSE);
    END;
*)

VAR
    basePos:CARDINAL;
    baseLen:CARDINAL;

PROCEDURE dmpTTX (paragraph:ARRAY OF CHAR;width:CARDINAL;
                  VAR S:ARRAY OF CHAR;firstcall:BOOLEAN) : BOOLEAN;
CONST
    blank=" ";
VAR
    len,count:CARDINAL;
BEGIN
    IF firstcall THEN
        basePos := 0;
        baseLen := Str.Length(paragraph); (* do not recalc it each time *)
    END;

    IF basePos = baseLen THEN RETURN FALSE; END; (* we're done now *)

    count:=width;
    IF (basePos+count) < baseLen THEN (* more than 'width' remaining chars *)
        LOOP
            IF paragraph[basePos+count]=blank THEN
                EXIT;
            END;
            DEC(count);
            IF count=0 THEN (* no space at all in string to get *)
                count:=width;
                EXIT;
            END;
        END;
        Str.Slice(S,paragraph,basePos,count); (* slice required portion *)
        INC(basePos,count); (* advance *)
        IF paragraph[basePos] = blank THEN INC(basePos); END; (* skip possible first space *)
    ELSE (* remaining chars are just 'width' or less *)
        Str.Slice(S,paragraph,basePos,baseLen-basePos);
        basePos:=baseLen; (* right on nullchar *)
    END;
    RETURN TRUE;
END dmpTTX;


PROCEDURE Waitkey () : str2;
VAR
    ch : str2;
BEGIN
    Str.Copy (ch,IO.RdKey());
    IF ch = str2(CHR(0)) THEN
        (* appending to CHR(0) is NOT a good idea if we need functions keys *)
        (* change special char to CHR(255) instead *)
        ch:=str2(CHR(255));
        Str.Append (ch,IO.RdKey());
        RETURN ch;
    ELSE
        RETURN ch;
    END;
END Waitkey;

PROCEDURE Flushkey ();
VAR
    ch : str2;
BEGIN
    WHILE ( IO.KeyPressed()=TRUE ) DO
        ch := Waitkey();
    END;
END Flushkey;



(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE chkJoker (S : ARRAY OF CHAR) : BOOLEAN;
CONST
    star     = "*";
    question = "?";
VAR
    p : CARDINAL;
BEGIN
    IF Str.CharPos(S,star) # MAX(CARDINAL) THEN RETURN TRUE; END;
    IF Str.CharPos(S,question) # MAX(CARDINAL) THEN RETURN TRUE; END;
    RETURN FALSE;
END chkJoker;

PROCEDURE isOption (S : ARRAY OF CHAR) : BOOLEAN;
BEGIN
CASE S[0] OF
| slash, dash :
    RETURN TRUE;
ELSE
    RETURN FALSE;
END;
END isOption;

(* ------------------------------------------------------------ *)

PROCEDURE fixEquate (VAR S:ARRAY OF CHAR);
CONST
    NOTFOUND = MAX(CARDINAL);
VAR
    pc,pe:CARDINAL;
BEGIN
    pc:=Str.CharPos(S,colon);
    pe:=Str.CharPos(S,equal);

    IF pc = NOTFOUND THEN
        IF pe # NOTFOUND THEN   (* "=" only *)
            Str.Subst(S,equal,colon); (* would do nothing IF equal does NOT exist *)
        END;
    ELSE
        IF pe # NOTFOUND THEN   (* both ":" and "=" *)
            IF pe < pc THEN     (* "=" before ":" *)
                Str.Subst(S,equal,colon);
            END;
        END;
    END;
END fixEquate;

PROCEDURE GetOptIndex (S : ARRAY OF CHAR; options : ARRAY OF CHAR) : CARDINAL;
VAR
    i   : CARDINAL;
    str : str80; (* option tested against s *)
BEGIN
    Str.Delete(S,0,1); (* remove leading - or / because we arrive if option *)
    Str.Append(options,delim);
    (*
    Str.Subst(S,equal,colon); (* command line option xxx= becomes xxx: *)
    *)
    fixEquate(S);
    ReplaceChar(options,equal,colon); (* all possibles option xxx= become xxx: *)

    i := 0;
    LOOP
        Str.ItemS(str,options,delim,i);
        IF Str.Length(str)=0 THEN RETURN 0; END;
        IF Str.CharPos(str,colon)=MAX(CARDINAL) THEN (* normal option *)
             IF Str.Compare(str,S)=0 THEN RETURN (i+1); END;
        ELSE (* option with a number *)
             Str.Append(str,"*"); (* match any number after : *)
             IF Str.Match(S, str )=TRUE THEN RETURN (i+1); END;
        END;
        INC(i);
    END;
END GetOptIndex;



(* //FIXED with fixEquate(), we no longer always replace first "=" with ":" *)

PROCEDURE GetLongCard (s : ARRAY OF CHAR; VAR value : LONGCARD) : BOOLEAN;
VAR
    p  : CARDINAL;
    v  : LONGCARD;
    ok : BOOLEAN;
BEGIN
    (*
    Str.Subst(s,equal,colon); (* command line option xxx= becomes xxx: *)
    *)
    fixEquate(s);

    p := Str.CharPos(s,colon); (* exclude MAX(CARDINAL) here! ;-) *)

    Str.Delete(s,0,p+1);
    v := Str.StrToCard(s,10,ok);
    value := v;
    RETURN ok;
END GetLongCard;

PROCEDURE GetLongInt (s : ARRAY OF CHAR; VAR value : LONGINT) : BOOLEAN;
VAR
    p  : CARDINAL;
    v  : LONGINT;
    ok : BOOLEAN;
BEGIN
    (*
    Str.Subst(s,equal,colon); (* command line option xxx= becomes xxx: *)
    *)
    fixEquate(s);

    p := Str.CharPos(s,colon); (* exclude MAX(CARDINAL) here! ;-) *)

    Str.Delete(s,0,p+1);
    v := Str.StrToInt(s,10,ok);
    value := v;
    RETURN ok;
END GetLongInt;

PROCEDURE GetString (s : ARRAY OF CHAR; VAR r : ARRAY OF CHAR);
VAR
    p : CARDINAL;
BEGIN
    (*
    Str.Subst(s,equal,colon); (* command line option xxx= becomes xxx: *)
    *)
    fixEquate(s);

    p := Str.CharPos(s,colon); (* exclude MAX(CARDINAL) here! ;-) *)

    Str.Delete(s,0,p+1);
    Str.Copy (r, s);
END GetString;

(* ------------------------------------------------------------ *)

PROCEDURE CharCount (S : ARRAY OF CHAR; what : CHAR) : CARDINAL;
VAR
    n    : CARDINAL;
    oldp : CARDINAL;
    nxtp : CARDINAL;
    len  : CARDINAL;
BEGIN
    n    := 0;
    len  := Str.Length(S);
    oldp := 0;
    LOOP
        nxtp := Str.NextPos(S,what,oldp);
        IF nxtp = MAX(CARDINAL) THEN EXIT; END;
        INC(n);
        oldp := nxtp+1;
        IF oldp >= len THEN EXIT; END;
    END;
    RETURN n;
END CharCount;

(* ------------------------------------------------------------ *)

PROCEDURE same(S1,S2 : ARRAY OF CHAR) : BOOLEAN;
BEGIN
    IF Str.Compare(S1,S2)=0 THEN RETURN TRUE; END;
    RETURN FALSE;
END same;

(* ------------------------------------------------------------ *)

PROCEDURE fixDirectory (VAR S : ARRAY OF CHAR);
VAR
    len : CARDINAL;
BEGIN
    len := Str.Length(S);
    IF len=0 THEN RETURN; END;
    IF S[len-1] = antislash THEN RETURN; END;
    Str.Append(S,antislash);
END fixDirectory;

(* take care of "\" or "u:\" *)

PROCEDURE unfixDirectory (VAR S : ARRAY OF CHAR);
VAR
    len : CARDINAL;
BEGIN
    len := Str.Length(S);
    IF len=0 THEN RETURN; END;
    IF S[len-1] # antislash THEN RETURN; END;
    IF same(S,"\") THEN RETURN;END;
    IF Str.Match(S,"?:\") THEN RETURN;END;
    S[len-1] := CHR(0);
END unfixDirectory;

PROCEDURE isDirectory (S : ARRAY OF CHAR) : BOOLEAN;
VAR
    entry : FIO.DirEntry;
    found : BOOLEAN;
BEGIN
    unfixDirectory(S);
    CASE Str.Length(S) OF
    | 2 : (* avoid "u:" alone ! but what about "*.", eh ? *)
        IF S[1]=colon THEN RETURN TRUE; END; (* always but we do not check for its real life *)
    | 3 : (* avoid "u:\" alone ! *)
        IF S[1]=colon THEN
            IF S[2]=antislash THEN RETURN TRUE; END; (* always but we do not check for its real life *)
        END;
    ELSE (* //XDS USELESS SYNTAX REQUIREMENT MAKING PROGRAM CRASH WHEN RUN ! *)
        ;
    END;
    found := FIO.ReadFirstEntry(S,everything,entry);
    IF found = FALSE THEN RETURN FALSE; END;
    RETURN (aDD IN entry.attr);
END isDirectory;

PROCEDURE Belongs (S : ARRAY OF CHAR; c:CHAR  ) : BOOLEAN;
BEGIN
    IF Str.CharPos(S,c)=MAX(CARDINAL) THEN
        RETURN FALSE;
    ELSE
        RETURN TRUE;
    END;
END Belongs;

(* ------------------------------------------------------------ *)

PROCEDURE LtrimBlanks (VAR S : ARRAY OF CHAR);
CONST
    espace=ORD(" ");
VAR
    i,len : CARDINAL;
BEGIN
    len := Str.Length (S);
    i   := 0;
    LOOP
        IF i = len THEN EXIT; END;
        IF ORD(S[i]) > espace THEN EXIT; END;
        INC (i);
    END;
    Str.Delete (S,0,i); (* procedure handles removal of i=0 *)
END LtrimBlanks;

(* ------------------------------------------------------------ *)

PROCEDURE RtrimBlanks (VAR S : ARRAY OF CHAR);
CONST
    espace=ORD(" ");
VAR
    i,len : CARDINAL;
BEGIN
    len := Str.Length (S);
    i   := len;
    LOOP
        IF i = 0 THEN EXIT; END;
        IF ORD(S[i-1]) > espace THEN EXIT; END;
        DEC (i);
    END;
    Str.Delete (S,i,len-i);
END RtrimBlanks;

(* ------------------------------------------------------------ *)

PROCEDURE getStrIndex (delimiter:CHAR;what,list: ARRAY OF CHAR) : CARDINAL;
VAR
    i   : CARDINAL;
    str : str80; (* option tested against s *)
    myList:str1024; (* long just in case *)
BEGIN
    Str.Concat(myList,list,delimiter);

    i := 0;
    LOOP
        Str.ItemS(str,myList,delimiter,i);
        IF Str.Length(str)=0 THEN RETURN 0; END;
        IF same(str,what) THEN RETURN (i+1); END;
        INC(i);
    END;
END getStrIndex;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    SetOfChars = SET OF CHAR;

PROCEDURE isoleItem(VAR R: ARRAY OF CHAR;
                    S: ARRAY OF CHAR; T: SetOfChars; N: CARDINAL);

VAR
    len,i,j,k:CARDINAL;
BEGIN
    len := Str.Length(S);
    i   := 0;
    WHILE ( (i<len) AND (N > 0) ) DO
        IF (S[i] IN T) THEN DEC(N); END;
        INC(i);
    END;
    (* i is at first char after separator *)
    j := i;
    WHILE ( (j<len) AND NOT (S[j] IN T) ) DO
        INC(j);
    END;
    k := 0;
    WHILE (i<j) DO
        R[k] := S[i];
        INC(k);
        INC(i);
    END;
    R[k]:=CHR(0);
END isoleItem;

(* now we can safely parse "a,b,c,,e" ! what about ",,c,d,e" ? *)
(* first element is 0 *)

PROCEDURE isoleItemS(VAR R: ARRAY OF CHAR;
                     S: ARRAY OF CHAR; T: ARRAY OF CHAR; N: CARDINAL);
VAR
    CS  : SetOfChars;
    len : CARDINAL;
BEGIN
    len := Str.Length(T);
    CS := SetOfChars{};
    WHILE len > 0 DO
        DEC(len);
        INCL(CS,T[len]);
    END;
    isoleItem(R,S,CS,N);
END isoleItemS;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE verifyString (S,allowed:ARRAY OF CHAR):BOOLEAN;
VAR
    i,mismatches : CARDINAL;
    ch : CHAR;
BEGIN
    mismatches := 0;
    FOR i := 1 TO Str.Length(S) DO
        ch := S[i-1];
        IF Belongs(allowed,ch)=FALSE THEN INC(mismatches);END;
    END;
    RETURN (mismatches=0);
END verifyString;

(* ------------------------------------------------------------ *)

(* try and fix TopSpeed bug when tabs are in command line *)

PROCEDURE cleantabs (VAR S : ARRAY OF CHAR);
CONST
    tab=CHR(9);
VAR
    p : CARDINAL;
BEGIN
    LOOP
        p:=Str.CharPos(S,tab);
        IF p = MAX(CARDINAL) THEN EXIT; END;
        Str.Delete(S,p,1); (* delete unwanted tabulation *)
    END;
END cleantabs;


PROCEDURE ChkEscape (  ) : BOOLEAN;
CONST
    escape  = CHR(27);
    special = CHR(0);
VAR
    ch : CHAR;
BEGIN
    IF IO.KeyPressed() = TRUE THEN
        ch := IO.RdKey();
        CASE ch OF
        | escape :
            RETURN TRUE;
        | special :
            ch := IO.RdKey();
            RETURN FALSE;
        ELSE
            RETURN FALSE;
        END;
    ELSE
        RETURN FALSE;
    END;
END ChkEscape;

PROCEDURE fileIsRO(S : ARRAY OF CHAR) : BOOLEAN;
VAR
    entry : FIO.DirEntry;
    found : BOOLEAN;
BEGIN
    found := FIO.ReadFirstEntry(S,allfiles,entry);
    IF found = FALSE THEN RETURN FALSE; END;
    RETURN (aR IN entry.attr);
END fileIsRO;

(* //FIXME //XDS to be done ! *)

PROCEDURE fileSetRW(S:ARRAY OF CHAR);
BEGIN
    ;
END fileSetRW;

PROCEDURE getFileSize(S:ARRAY OF CHAR):LONGCARD;
VAR
    hnd:FIO.File;
    fsize:LONGCARD;
BEGIN
    hnd:=FIO.OpenRead(S);
    fsize:=FIO.Size(hnd);
    FIO.Close(hnd);
    RETURN fsize;
END getFileSize;

PROCEDURE vidWrStr (s : ARRAY OF CHAR);
(*
CONST
    stderr        = 02H;  (* write to stderr as not to write control chars to possible redirected file *)
    stderr        = FIO.ErrorOutput;
*)
BEGIN
    FIO.WrStr(FIO.ErrorOutput,s);
END vidWrStr;

PROCEDURE video (S : ARRAY OF CHAR; flag:BOOLEAN);
CONST
    charBS    = 10C;
    charBlank = 40C;
    bsblankbs = charBS+charBlank+charBS;
VAR
    len : CARDINAL;
    i   : CARDINAL;
BEGIN
    len := Str.Length(S);
    CASE flag OF
    | TRUE :
        vidWrStr(S);
    | FALSE:
        FOR i := 1 TO len DO
            vidWrStr(bsblankbs);
        END;
    ELSE
        ; (* //XDS no kidding ??? ternary boolean logic ??? *)
    END;
END video;


(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* // QD_RAND *)

CONST
    firstalea = 1;
    lastalea  = 128;
VAR
    pool : ARRAY [firstalea..lastalea] OF REAL;

(* ------------------------------------------------------------ *)

PROCEDURE InitRnd ();
VAR
    i : CARDINAL;
BEGIN
    Lib.RANDOMIZE;
    FOR i := firstalea TO lastalea DO
        pool[i] := Lib.RAND();
    END;
END InitRnd;

(* ------------------------------------------------------------ *)

(*
PROCEDURE GetRnd () : REAL;
VAR
    index:CARDINAL;
    lowerbound:REAL;
    upperbound:REAL;
    rnd:REAL;
BEGIN
    lowerbound:=REAL(firstalea);
    upperbound:=REAL(lastalea);
    rnd := (upperbound-lowerbound+1.0) * Lib.RAND() + lowerbound ;
    index := CARDINAL(rnd);
    rnd:=pool[index];
    pool[index]:=Lib.RAND();
    RETURN rnd;
END GetRnd;
*)

PROCEDURE GetRnd () : REAL;
VAR
    index:CARDINAL;
    rnd  :REAL;
BEGIN
    index := Lib.RANDOM(lastalea-firstalea+1); (* [0..128[ [0..127] *)
    rnd:=pool[firstalea+index];
    pool[firstalea+index]:=Lib.RAND();
    RETURN rnd;
END GetRnd;

(* ------------------------------------------------------------ *)

PROCEDURE GetRndCardRange (lowerlimit,upperlimit:CARDINAL) : CARDINAL;
VAR
    lowerbound:REAL;
    upperbound:REAL;
    rnd:REAL;
BEGIN
    lowerbound:=REAL(lowerlimit);
    upperbound:=REAL(upperlimit);
    rnd :=  (upperbound-lowerbound+1.0) * GetRnd()+ lowerbound ;
    RETURN CARDINAL(rnd);
END GetRndCardRange;

PROCEDURE GetRndLngCardRange (lowerlimit,upperlimit:LONGCARD) : LONGCARD;
VAR
    lowerbound:REAL;
    upperbound:REAL;
    rnd:REAL;
BEGIN
    lowerbound:=REAL(lowerlimit);
    upperbound:=REAL(upperlimit);
    rnd :=  (upperbound-lowerbound+1.0) * GetRnd()+ lowerbound ;
    RETURN LONGCARD(rnd);
END GetRndLngCardRange;

(* ------------------------------------------------------------ *)

(*
    eh, the formula does not work for reals ! try removing the -1.0
    maybe add an adjustement to upper limit, such as 0.0001 ?
*)

PROCEDURE GetRndLngRealRange (lowerlimit,upperlimit:LONGREAL) : LONGREAL;
VAR
    lowerbound:REAL;
    upperbound:REAL;
    rnd:REAL;
BEGIN
    lowerbound:=REAL(lowerlimit);
    upperbound:=REAL(upperlimit);
    rnd :=  (upperbound-lowerbound) * GetRnd()+ lowerbound ;
    RETURN LONGREAL(rnd);
END GetRndLngRealRange;

PROCEDURE GetRndRealRange (lowerbound,upperbound:REAL) : REAL;
VAR
    rnd:REAL;
BEGIN
    rnd :=  (upperbound-lowerbound) * GetRnd()+ lowerbound ;
    RETURN rnd;
END GetRndRealRange;

(* ------------------------------------------------------------ *)

PROCEDURE GetRndIntRange (lowerlimit,upperlimit:INTEGER) : INTEGER;
VAR
    lowerbound:REAL;
    upperbound:REAL;
    rnd:REAL;
BEGIN
    lowerbound:=REAL(lowerlimit);
    upperbound:=REAL(upperlimit);
    rnd :=  (upperbound-lowerbound+1.0) * GetRnd()+ lowerbound ;
    RETURN INTEGER(rnd);
END GetRndIntRange;

PROCEDURE GetRndLngIntRange (lowerlimit,upperlimit:LONGINT) : LONGINT;
VAR
    lowerbound:REAL;
    upperbound:REAL;
    rnd:REAL;
BEGIN
    lowerbound:=REAL(lowerlimit);
    upperbound:=REAL(upperlimit);
    rnd :=  (upperbound-lowerbound+1.0) * GetRnd()+ lowerbound ;
    RETURN LONGINT(rnd);
END GetRndLngIntRange;


(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* we could use VAR cli as to go a little faster : nay, not even really faster in fact ! *)

PROCEDURE argc (cli : ARRAY OF CHAR; clean : BOOLEAN) : CARDINAL;
VAR
    state    : (empty,intoken,instring);
    cliLen   : CARDINAL;
    cliPos   : CARDINAL;
    argCount : CARDINAL;
    ch       : CHAR;
    code     : CARDINAL;
    arg      : str128;
    sdelim   : CARDINAL;
BEGIN
    IF clean THEN
        LtrimBlanks(cli);
        RtrimBlanks(cli);
    END;
    cliLen   := Str.Length(cli);
    cliPos   := 1;
    argCount := 0;
    state    := empty;
    LOOP
        IF cliPos > cliLen THEN EXIT; END;
        ch := cli[cliPos-1];
        code := ORD(ch);
        CASE state OF
        | empty :
            IF ( (code = ORD(doublequote)) OR (code=ORD(singlequote)) ) THEN
                sdelim:= code;
                state := instring;        (* begin new string *)
                Str.Copy(arg,ch);
                INC(argCount);
            ELSIF code > ORD(blank) THEN  (* quote already trapped *)
                state := intoken;         (* begin new token *)
                Str.Copy(arg,ch);
                INC(argCount);
            END;
        | intoken :
            IF ( (code = ORD(doublequote)) OR (code=ORD(singlequote)) ) THEN (* quote in token *)
                sdelim:= code; (* fix *)
                state := instring;        (* if string in parameter *)
                Str.Append(arg,ch);
            ELSIF code > ORD(blank) THEN
                Str.Append(arg,ch);
            ELSE
                state := empty;           (* end of token *)
            END;
        | instring :
            IF code = sdelim THEN
                Str.Append(arg,ch);
                state := empty;           (* end of string *)
            ELSIF code > ORD(blank) THEN
                Str.Append(arg,ch);
            ELSE
                Str.Append(arg,blank);    (* remove TAB and controls if any *)
            END;
        ELSE ; (* //ISO fixed silly requirement *) END;
        INC(cliPos);
    END;
    RETURN argCount;
END argc;

PROCEDURE argv (VAR argument : ARRAY OF CHAR;
                cli : ARRAY OF CHAR; n : CARDINAL; clean:BOOLEAN);
VAR
    state    : (empty,intoken,instring);
    cliLen   : CARDINAL;
    cliPos   : CARDINAL;
    argCount : CARDINAL;
    ch     : CHAR;
    code     : CARDINAL;
    arg      : str128;
    sdelim   : CARDINAL;
BEGIN
    IF ( (n < 1) OR (n > argc(cli,clean)) ) THEN
        Str.Copy(argument,"");
        RETURN;
    END;
    IF clean THEN
        LtrimBlanks(cli);
        RtrimBlanks(cli);
    END;
    cliLen   := Str.Length(cli);
    cliPos   := 1;
    argCount := 0;
    state    := empty;
    LOOP
        IF cliPos > cliLen THEN EXIT; END;
        ch := cli[cliPos-1];
        code := ORD(ch);
        CASE state OF
        | empty :
            IF n = argCount THEN EXIT; END; (* argV$ test *)
            IF ( (code = ORD(doublequote)) OR (code=ORD(singlequote)) ) THEN
                sdelim:= code;
                state := instring;        (* begin new string *)
                Str.Copy(arg,ch);
                INC(argCount);
            ELSIF code > ORD(blank) THEN  (* quote already trapped *)
                state := intoken;         (* begin new token *)
                Str.Copy(arg,ch);
                INC(argCount);
            END;
        | intoken :
            IF ( (code = ORD(doublequote)) OR (code=ORD(singlequote)) ) THEN (* quote in token *)
                sdelim:= code; (* fix *)
                state := instring;        (* if string in parameter *)
                Str.Append(arg,ch);
            ELSIF code > ORD(blank) THEN
                Str.Append(arg,ch);
            ELSE
                state := empty;           (* end of token *)
            END;
        | instring :
            IF code = sdelim THEN
                Str.Append(arg,ch);
                state := empty;           (* end of string *)
            ELSIF code > ORD(blank) THEN
                Str.Append(arg,ch);
            ELSE
                Str.Append(arg,blank);    (* remove TAB and controls if any *)
            END;
        ELSE ; (* //ISO fixed silly requirement *) END;
        INC(cliPos);
    END;
    Str.Copy(argument,arg);
END argv;


(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)



(*
    handle both "\" and "/" separators
    *:*\[n[.[e]]]

    grab last "\" if any then check for next "."

    we could fix "/" and "\" mix, as well as empty extension ?
*)

PROCEDURE qdSplitAllPath (udne:ARRAY OF CHAR;VAR u,d,n,e:ARRAY OF CHAR);
VAR
    F:str1024; (* bigger than pathtype *)
    p:CARDINAL;
BEGIN
    (*
    Lib.SplitAllPath(F,u,d,n,e);
    Lib.MakeAllPath(F,u,d,n,e);
    IF Str.Compare(udne,F) = 0 THEN RETURN; END;
    *)

    (* let's fix JPI buggy function so we still get "u:" "\*\" "f" ".e" *)

    p:=Str.CharPos(udne,colon);
    IF p = MAX(CARDINAL) THEN
        Str.Copy(u,"");
    ELSE
        Str.Slice(u,udne,0,p+1);
        Str.Delete(udne,0,p+1);
    END;

    Str.Copy(F,udne);
    LOOP
        p:=Str.CharPos(F,slash);
        IF p = MAX(CARDINAL) THEN EXIT;END;
        F[p]:=antislash;
    END;
    p:=Str.RCharPos(F,antislash);
    IF p = MAX(CARDINAL) THEN
        Str.Copy(d,"");
    ELSE
        Str.Slice(d,udne,0,p+1);
        Str.Delete(udne,0,p+1);
    END;

    p:=Str.RCharPos(udne,dot);
    CASE p OF
    | MAX(CARDINAL) :
        Str.Copy(n,udne);
        Str.Copy(e,"");
    | 0 :
        Str.Copy(n,"");
        Str.Copy(e,udne);
    ELSE
        Str.Slice(n,udne,0,p);
        Str.Delete(udne,0,p);
        Str.Copy(e,udne);
    END;

END qdSplitAllPath;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    pathtype = FIO.FileName;

(* ------------------------------------------------------------ *)

CONST
    DBG = FALSE;

(*%T DBG  *)
PROCEDURE dbgv (S:ARRAY OF CHAR;v:CARDINAL );
CONST
    F = "WOLFRULE.DBG";
VAR
    hout:FIO.File;
BEGIN
    IF FIO.Exists(F) THEN
        hout:=FIO.Append(F);
    ELSE
        hout:=FIO.Create(F);
    END;
    FIO.WrStr(hout,S);FIO.WrStr(hout," = ");
    FIO.WrCard(hout,v,0);FIO.WrLn(hout);
    FIO.Close(hout);
END dbgv;

CONST
    DBGFILE = "WOLFRULE.DBG";

PROCEDURE dbg (S,R:ARRAY OF CHAR   );
VAR
    hout:FIO.File;
BEGIN
    IF FIO.Exists(DBGFILE) THEN
        hout:=FIO.Append(DBGFILE);
    ELSE
        hout:=FIO.Create(DBGFILE);
    END;
    FIO.WrStr(hout,S); FIO.WrStr(hout," = ");
    FIO.WrStr(hout,R);FIO.WrLn(hout);
    FIO.Close(hout);
END dbg;
(*%E  *)

(* ------------------------------------------------------------ *)

CONST
    exe        = "WOLFRU32";
    allowed    = "01 .X";
    allowedrnd = allowed+"?"; (* "?" means random "0" OR "1" *)
    ver        = "v0.9 (from v1.0b for DOS)";
    extTXT     = ".TXT";
    extWR      = ".RUL";
CONST
    msghelp= "Q&D Wolfram Rule (Win32) "+ver+" by PhG"+nl+
nl+
"Syntax : "+exe+" [file["+extWR+"|"+extTXT+"]] [option]..."+nl+
nl+
"-r:#   8 bits bitmask"+nl+
"-w:#   initial cellular automaton width (default is 1)"+nl+
"-i:$   initial binary pattern"+nl+
"-n     do not force at least one bit on in initial pattern"+nl+
"-b     black and white mode (default is color mode)"+nl+
"-v     wait for screen vertical retrace"+nl+
"-t[:#] testmode (-t:# = -t -r:#)"+nl+
nl+
"a) If no bit is on in initial pattern, background will be red."+nl+
"b) Note wrapping is enabled for first and last cellular automaton cells."+nl+
"c) -r:# bitmask assumes 111 110 101 100 011 010 001 000 order ;"+nl+
"   it can be expressed as a $hexadecimal, decimal or %binary number."+nl+
"d) While program is running, [Space] = pause, [Return] = quit."+nl+
'   In test mode, "+" = increment rule, "-" = decrement rule.'+nl+
"e) -t option requires -r:# option."+nl+
"f) -r:# less uninteresting displays seem to be :"+nl+
"   18 22 26 30 45 60 75 82 86 89 90"+nl+
"   101 102 105 110 124 126 129 135 137 146 149"+nl+
"   150 153 154 161 165 167 169 181 182 193 195 210 218 225."+nl+
"g) Here is an example of rule file (Wolframm's Rule 30) :"+nl+
nl+
"   ; line 1 : each of the 8 possible patterns"+nl+
"   ; line 2 : new center cell state"+nl+
"   ; optional line 3 : initial pattern (-w:# ignored)"+nl+
"   ; without optional line 3, initial pattern is randomly created"+nl+
"   111 110 101 100 011 010 001 000"+nl+
"    0   0   0   1   1   1   1   0"+nl+
"   1"+nl+
'   ; alternate characters are "." for 0 and "x" for 1'+nl+
'   ; "?" will randomly return either "0" or "1"'+nl+
'   ; undefined patterns will force "0" for new center cell state'+nl+
"h) The original J.P.I. TopSpeed Modula-2 DOS source was reluctantly ported"+nl+
"   to XDS compiler using the so-called TopSpeed Compatibility Kit."+nl+
"   Although successful compiling required (too) many modifications in source,"+nl+
"   results should match the original DOS program. ERRORINF.$$$ may happen ! ;-)"+nl

(*%F DBG  *)
;
(*%E  *)
(*%T DBG  *)
+
nl+
"Debug data dumped to "+DBGFILE+" text file."+nl;
(*%E  *)

CONST
    errNone         = 0;
    errHelp         = 1;
    errBadWidth     = 2;
    errWidthRange   = 3;
    errOption       = 4;
    errParm         = 5;
    errJoker        = 6;
    errNotFound     = 7;
    errGraphMode    = 8;
    errCharSet      = 9;
    errThreeLines   = 10;
    errMaxPatterns  = 11;
    errMaxInitPattern=12;
    errCount        = 13;
    errBadMask      = 14;
    errRuleOrFile   = 15;
    errBadInitial   = 16;
    errInitialOrFile= 17;
    errTestMode     = 18;
    errRuleRequired = 19;

PROCEDURE abort (e:CARDINAL);
VAR
   S : str128;
BEGIN
   CASE e OF
   | errNone:   ;
   | errHelp:        WrStr(msghelp);
   | errBadWidth :   S := "Illegal -w:#";
   | errWidthRange:  S := "Out of [1..640] range -w:#";
   | errOption :     S := "Illegal option";
   | errParm   :     S := "At least one parameter too many";
   | errJoker  :     S := "Jokers are not allowed";
   | errNotFound:    S := "File does not exist";
   | errGraphMode:   S := "Cannot set graphics mode";
   | errCharSet:     S := "Only characters from ["+allowed+"][?] are allowed in rule file entries";
   | errThreeLines:  S := "Rule file cannot contain more than 3 lines";
   | errMaxPatterns :S := "More than 8 patterns or new states";
   | errMaxInitPattern:S:="More than 1 initial pattern";
   | errCount      : S := "Patterns and new states count mismatch";
   | errBadMask    : S := "Illegal -r:# value";
   | errRuleOrFile : S := "-r:# option and rule file specification are mutually exclusive";
   | errBadInitial : S := "Illegal -i:$ pattern";
   | errInitialOrFile:S:= "-i:$ option and rule file specification are mutually exclusive";
   | errTestMode:    S := "-t option and rule file specification are mutually exclusive";
   | errRuleRequired:S := "-t option requires -r:# option";
   ELSE
                     S := "This is impossible, Captain";
   END;
   CASE e OF
   | errNone,errHelp:
       e:=errNone;
   ELSE
       WrStr(exe+" : ");WrStr(S);WrStr(" !");WrLn;
   END;
   (* Lib.SetReturnCode( SHORTCARD(e) ); *)
   HALT (e);
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    gmode  = Graph._VRES16COLOR; (* 640x480x4 *)
    tmode  = Graph._DEFAULTMODE;
CONST
    black  = 0;
    white  = 15;
    yellow = 14;
    mauve  = 13;
    red    = 12;
    blue   = 11;
    green  = 10;
    minink = black+1;
    maxink = white;

(* ------------------------------------------------------------ *)

VAR
    xmin,xmax,ymin,ymax : CARDINAL;

PROCEDURE retrace (n:CARDINAL );
VAR
    i:CARDINAL;
BEGIN
    (*
    FOR i:=1 TO n DO
        WHILE (SYSTEM.In(03DAH) AND 08H) # 0 DO
        END;
        WHILE (SYSTEM.In(03DAH) AND 08H) = 0 DO
        END;
    END;
    *)
END retrace;

PROCEDURE HiresON () : BOOLEAN;
BEGIN
    RETURN Graph.SetVideoMode(gmode);
END HiresON;

PROCEDURE HiresOFF() : BOOLEAN;
BEGIN
    RETURN Graph.SetVideoMode(tmode);
END HiresOFF;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE XOR (b1,b2 : CARDINAL) : CARDINAL;
BEGIN
    RETURN CARDINAL ( (BITSET (b1) / BITSET (b2)) );
END XOR;

CONST
    maxpatterns     = 8;
    maxinitpattern  = 1;
    defaultPat      = "111 110 101 100 011 010 001 000"; (* all bitmasks from 7 to 0 *)
    defaultNewState = " 0   0   0   1   1   1   1   0";
    defaultInit     = ""; (* empty = random, could be "1" *)

PROCEDURE readRule (VAR sPat,sNewState,sInit:ARRAY OF CHAR;
                   useLFN:BOOLEAN;rulefile:pathtype  ):CARDINAL;
VAR
    hin:FIO.File;
    S:str1024;
    rc:CARDINAL;
    grabmode:(waiting,gotpat,gotnewstate,gotinit);
BEGIN
    rc:=errNone;
    Str.Copy(sPat,"");
    Str.Copy(sNewState,"");
    Str.Copy(sInit,"");
    grabmode:=waiting;
    FIO.EOF:=FALSE;
    hin:=FIO.OpenRead(rulefile);
    LOOP
        FIO.RdStr(hin,S);
        (*%T DBG  *) dbg("S",S); (*%E *)
        LtrimBlanks(S);
        RtrimBlanks(S);
        CASE S[0] OF
        | charnull,semicolon,pound : ;
        ELSE
            ReplaceChar(S,tabul,space);
            CASE grabmode OF
            | waiting:     IF verifyString(S,allowed)=FALSE THEN
                               rc:=errCharSet;
                               EXIT;
                           END;
                           Str.Copy(sPat,S);
            | gotpat:      IF verifyString(S,allowedrnd)=FALSE THEN
                               rc:=errCharSet;
                               EXIT;
                           END;
                           Str.Copy(sNewState,S);
            | gotnewstate: IF verifyString(S,allowed)=FALSE THEN
                               rc:=errCharSet;
                               EXIT;
                           END;
                           Str.Copy(sInit,S);
            | gotinit:     rc:=errThreeLines; EXIT;
            ELSE ; (* //ISO fixed silly requirement *) END;
            INC(grabmode);
        END;
        IF FIO.EOF THEN EXIT;END;
    END;
    FIO.Close(hin);
    IF rc=errNone THEN
        (* any error is an error is an error is an error... *)
        IF argc(sPat,FALSE) > maxpatterns THEN rc:=errMaxPatterns;END;
        IF argc(sNewState,FALSE) > maxpatterns THEN rc:=errMaxPatterns;END;
        IF argc(sInit,FALSE) > maxinitpattern THEN rc:=errMaxInitPattern;END;
        IF rc=errNone THEN
            IF argc(sPat,FALSE) # argc(sNewState,FALSE) THEN rc:=errCount;END;
        END;
        (* //FIXME yes, yes, we should check lmr's and newcenter's length *)
    END;
    RETURN rc;
END readRule;

(* ------------------------------------------------------------ *)

CONST
    MAXHASH = MAX(CARDINAL)-1;
    CAmin = 0;
    CAmax = 640-1;
    firstclock = 1;
    lastclock  = 2;
    minbitpattern = 1;
    maxbitpattern = 8;
TYPE
    str3 = ARRAY [0..2] OF CHAR;
    transformtype = RECORD
        hash : CARDINAL;
        lmr : str3;
        newcenter : CHAR;
    END;

TYPE
    CAtype = ARRAY[CAmin..CAmax] OF CHAR; (* bit state i.e. BOOLEAN *)
VAR
    CA : ARRAY [firstclock..lastclock] OF CAtype;
    bitpattern : ARRAY[minbitpattern..maxbitpattern] OF transformtype;

PROCEDURE chToVal ( ch:CHAR ):CARDINAL ;
VAR
    v:CARDINAL;
BEGIN
    CASE ch OF
    | "0", dot : v:=0;
    | "1", "X" : v:=1;
    ELSE
        v:=MAX(CARDINAL); (* should never happen *)
        v:=0; (* //HACK *)
    END;
    RETURN v;
END chToVal;

PROCEDURE valToCh ( v:CARDINAL ):CHAR;
VAR
    ch:CHAR;
BEGIN
    CASE v OF
    | 0 : ch:="0";
    | 1 : ch:="1";
    ELSE
        ch:=charnull;
        ch:="0";       (* //HACK *)
    END;
    RETURN ch;
END valToCh;

PROCEDURE initCA (VAR ndx,inkbk:CARDINAL;
                 initwi,last:CARDINAL; forceonelit:BOOLEAN;
                 sPat,sNewState,sInit:ARRAY OF CHAR);
VAR
     hash,i,p,v,cx,x,lit:CARDINAL;
     rnd:BOOLEAN;
     lmr:str3;
     newcenter:CHAR;
BEGIN
     (*%T DBG *) dbgv("initwi at entry",initwi); (*%E  *)
     (*%T DBG *) dbg("sPat",sPat); (*%E  *)
     (*%T DBG *) dbg("sNewState",sNewState); (*%E  *)
     (*%T DBG *) dbg("sInit",sInit); (*%E  *)
     rnd:=same(sInit,"");
     IF NOT(rnd) THEN
         initwi := Str.Length(sInit);
     END;
     (*%T DBG *) dbgv("initwi",initwi); (*%E  *)
     DEC(initwi);

     FOR i:=CAmin TO last DO
         CA[firstclock,i]:="0";
         CA[lastclock ,i]:="0";
     END;

     lit:=0;
     cx:=(last-CAmin) DIV 2;   (* 639-0 => 319 *)
     x:=cx - (initwi DIV 2);     (* initwi=0 => 319-0 , initwi=8 => 319-4 *)
     FOR i:=CAmin TO initwi DO
         p:=x+i;
         (*%T DBG *) dbgv("screen index",p); (*%E *)
         IF rnd THEN
             v:=Lib.RANDOM(2); (* 0..1 *)
         ELSE
             v:=chToVal( sInit[i] );
         END;
         IF v = 1 THEN INC(lit);END;
         CA[firstclock,p]:=valToCh(v);
     END;
     (*%T DBG   *)dbgv("lit cells",lit); (*%E *)
     IF forceonelit THEN
         IF lit = 0 THEN
             p:=cx - Lib.RANDOM(initwi+1) DIV 2;  (* +1 because of previous DEC() *)
             CA[firstclock,p]:="1";
             lit:=1;
             (*%T DBG *) dbgv("force lit cell",p); (*%E  *)
         END;
     END;

     IF lit=0 THEN
         inkbk:=red;
     ELSE
         inkbk:=black;
     END;

     (* now parse sPat and sNewState *)

     lmr:="";     newcenter:="";
     lmr:="000";  newcenter:="0";  (* //HACK *)

     hash:=Lib.HashString(lmr,MAXHASH);
     FOR i:=minbitpattern TO maxbitpattern DO
         hash:=hash;
         bitpattern[i].lmr:=lmr;
         bitpattern[i].newcenter:=newcenter;
     END;

     FOR i:=minbitpattern TO argc(sPat,FALSE) DO
         argv(lmr, sPat,i,FALSE);
         argv(newcenter, sNewState,i,FALSE);
         hash:=Lib.HashString(lmr,MAXHASH);
         bitpattern[i].hash := hash;
         bitpattern[i].lmr := lmr;
         bitpattern[i].newcenter := newcenter;
     END;

     ndx:=firstclock;
END initCA;

(* for now, limited to 3 bits patterns *)

PROCEDURE processline (ndx,last:CARDINAL ):CARDINAL ;
VAR
    j,hash,i,nextndx:CARDINAL;
    l,m,r:CHAR;
    newcenter:CHAR;
    lmr:str3;
BEGIN
    CASE ndx OF
    | firstclock : nextndx:=lastclock;
    | lastclock  : nextndx:=firstclock;
    ELSE ; (* //ISO fixed silly requirement *) END;

    FOR i:= CAmin TO last DO
        m := CA[ ndx,i ];
        IF i = CAmin THEN
            l := CA [ndx,last];
            r := CA [ndx,i+1 ];
        ELSIF i= last THEN
            l := CA [ndx,last-1];
            r := CA [ndx,CAmin];
        ELSE
            l := CA [ndx,i-1];
            r := CA [ndx,i+1];
        END;

        (* yes, yes, storing to string would DO too *)
        Str.Concat(lmr,l,m);
        Str.Append(lmr,r);
        hash:=Lib.HashString(lmr,MAXHASH);

        j:=minbitpattern-1;
        LOOP
            INC(j);
            IF j > maxbitpattern THEN
                newcenter:="";
                newcenter:="0"; (* //HACK *)
                EXIT;
            END;
            IF hash = bitpattern[j].hash THEN
                IF lmr = bitpattern[j].lmr THEN (* collision may happen ! ;-) *)
                    newcenter:=bitpattern[j].newcenter;
                    IF newcenter = "?" THEN
                        newcenter := valToCh( Lib.RANDOM(2) ); (* 0..1 *)
                    END;
                    EXIT;
                END;
            END;
        END;
        CA[nextndx,i]:=newcenter;
    END;
    RETURN nextndx;
END processline;

PROCEDURE showstate (ndx,last,ycurr,ink,inkbk,vsync:CARDINAL):CARDINAL;
VAR
    myink,x:CARDINAL;
BEGIN
    retrace(vsync);
    FOR x:=CAmin TO xmax DO
         IF CA[ndx,x]="1" THEN
             myink:=ink;
         ELSE
             myink:=inkbk;
         END;
         Graph.Plot(x,ycurr,myink);
    END;
    INC(ycurr);
    IF ycurr > ymax THEN ycurr:=ymin;END;
    RETURN ycurr;
END showstate;

PROCEDURE nextink ( ink,ycurr:CARDINAL;cycle:BOOLEAN  ):CARDINAL;
BEGIN
    IF cycle THEN
        IF ycurr = ymin THEN
            (* ugly
            INC(ink);
            IF ink > maxink THEN ink:=minink;END; (* black avoided *)
            *)
            CASE ink OF
            | white : ink:=green;
            | green : ink:=white;
            ELSE ; (* //ISO fixed silly requirement *) END;
        END;
    END;
    RETURN ink;
END nextink;

PROCEDURE gotkey (VAR ch:CHAR):BOOLEAN;
VAR
    got:BOOLEAN;
BEGIN
    ch:="";
    got:=IO.KeyPressed();
    IF got THEN
        ch:=IO.RdKey();
        IF ch = CHR(0) THEN ch:=IO.RdKey();ch:="";END;
    END;
    RETURN got;
END gotkey;

PROCEDURE showparms (testmode:BOOLEAN;rulemask:CARDINAL );
VAR
    i:CARDINAL;
BEGIN
    WrLn;
    WrStr("Original pattern : ");
    FOR i:=minbitpattern TO maxbitpattern DO
        WrStr( bitpattern[i].lmr);
        IF i < maxbitpattern THEN WrStr(" ");END;
    END;
    WrLn;
    WrStr("New center cell  : ");
    FOR i:=minbitpattern TO maxbitpattern DO
        WrStr(" ");
        WrStr( bitpattern[i].newcenter);
        IF i < maxbitpattern THEN WrStr("  ");END;
    END;
    WrLn;
    IF testmode THEN
        WrStr("Rule             : ");IO.WrCard(rulemask,0);WrLn;
    END;
END showparms;

PROCEDURE grabrulemask (VAR rule:CARDINAL;S:ARRAY OF CHAR    ):BOOLEAN;
VAR
    ok:BOOLEAN;
    R:str16;
    allowed:str80;
    base:CARDINAL;
    lc:LONGCARD;
BEGIN
    GetString(S, R);
    CASE R[0] OF
    | "$": base:=16; allowed:=digits+"ABCDEF"; Str.Delete(R,0,1);
    | "%": base:=2;  allowed:="01";            Str.Delete(R,0,1);
    ELSE
           base:=10; allowed:=digits;
    END;
    Str.Caps(R); (* already uppercase but who knows ? *)
    ok:=verifyString(R,allowed);
    IF ok THEN
        lc:=Str.StrToCard(R,base,ok);
        IF ok THEN
            ok:= ( lc <= MAX(SHORTCARD) );
            IF ok THEN rule:=CARDINAL(lc);END;
        END;
    END;
    RETURN ok;
END grabrulemask;

PROCEDURE grabinitial ( VAR R:ARRAY OF CHAR; S:ARRAY OF CHAR  ):BOOLEAN;
VAR
    ok:BOOLEAN;
BEGIN
    GetString(S, R);
    IF R[0] = "%" THEN Str.Delete(R,0,1);END;
    ok:=verifyString(R,"01");
    RETURN ok;
END grabinitial;

PROCEDURE testbit(Target, BitNum : CARDINAL) : BOOLEAN;
BEGIN
    RETURN ( (BitNum MOD 16) IN BITSET(Target) );
END testbit;

PROCEDURE byteToMask (VAR R:ARRAY OF CHAR;v:CARDINAL );
VAR
    i:CARDINAL;
BEGIN
    Str.Copy(R,"");
    FOR i:=maxbitpattern TO minbitpattern BY -1 DO (* 8..1 *)
        IF testbit(v,i-1) THEN (* 0-based *)
            Str.Append(R," 1");
        ELSE
            Str.Append(R," 0");
        END;
    END;
END byteToMask;

PROCEDURE maskToByte ( S:ARRAY OF CHAR  ):CARDINAL;
VAR
    lc:LONGCARD;
    base:CARDINAL;
    ok:BOOLEAN;
BEGIN
    ReplaceChar(S,blank,"");
    lc:=Str.StrToCard(S,2,ok);
    RETURN CARDINAL(lc);
END maskToByte;

PROCEDURE newvalue (k:INTEGER;v,mini,maxi:CARDINAL ):CARDINAL ;
VAR
    n:INTEGER;
BEGIN
    n:=INTEGER(v);
    INC(n,k);
    IF n < INTEGER(mini) THEN
        n:=maxi;
    ELSIF n > INTEGER(maxi) THEN
        n:=mini;
    END;
    RETURN CARDINAL(n);
END newvalue;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    parmcount,i,opt : CARDINAL;
    S,R,rulefile    : pathtype;
    lc              : LONGCARD;
    ok              : BOOLEAN;
    ch              : CHAR;
    rulemask,ink,inkbk,vsync,ycurr,ndx,initwi            : CARDINAL;
    V:Graph.VideoConfig;
    useLFN,cycle,forceonelit,testmode:BOOLEAN;
    state : (waiting,gotfile);
    sPat,sNewState:str128;
    sInit:str1024;
    sInitCLI:str128;
    k:INTEGER;
BEGIN
    FIO.IOcheck:=FALSE;
    WrLn;

    initwi := 1;
    cycle  := TRUE;
    forceonelit := TRUE;
    vsync  := 0;
    rulemask := MAX(CARDINAL);
    sInitCLI:="";
    testmode:=FALSE;
    useLFN := TRUE;
    state  := waiting;

    parmcount := Lib.ParamCount();

    (* IF parmcount = 0 THEN abort(errHelp);END; *)

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "W:"+delim+"WIDTH:"+delim+
                                  "B"+delim+"BW"+delim+
                                  "V"+delim+"VSYNC"+delim+
                                  "N"+delim+"DONOTFORCE"+delim+
                                  "R:"+delim+"RULE:"+delim+
                                  "I:"+delim+"INIT:"+delim+
                                  "T"+delim+"TEST"+delim+
                                  "T:"+delim+"TEST:"
                              );
            CASE opt OF
            | 1,2,3  : abort(errHelp);
            | 4,5    : ok:=GetLongCard(S,lc);
                       IF NOT(ok) THEN abort(errBadWidth);END;
                       initwi:=CARDINAL(lc);
                       IF ( ( initwi < (CAmin+1) ) OR (initwi > (CAmax+1)) ) THEN
                           abort(errWidthRange);
                       END;
            | 6,7    : cycle := FALSE;
            | 8,9    : vsync:=1;
            |10,11:    forceonelit:=FALSE;
            |12,13:    IF grabrulemask(rulemask,R)=FALSE THEN abort(errBadMask);END;
            |14,15:    IF grabinitial(sInitCLI,R)=FALSE THEN abort(errBadInitial);END;
            |16,17:    testmode:=TRUE;
            |18,19:    IF grabrulemask(rulemask,R)=FALSE THEN abort(errBadMask);END;
                       testmode:=TRUE;
            ELSE
                abort(errOption);
            END;
        ELSE
            CASE state OF
            | waiting:
                IF same(S,"?") THEN abort(errHelp);END;
                Str.Copy(rulefile,S);
            ELSE
                abort(errParm);
            END;
            INC(state);
        END;
    END;

    (* useLFN := (useLFN AND fileSupportLFN() ); *)

    (*%T DBG *) dbg("//","NEW RUN !"); (*%E  *)

    Lib.RANDOMIZE;

    CASE state OF
    | waiting:
        (*%T DBG *) dbg("rulefile","none"); (*%E  *)
        sPat      := defaultPat;
        sNewState := defaultNewState;
        sInit     := defaultInit;
        IF rulemask = MAX(CARDINAL) THEN
            IF testmode THEN abort(errRuleRequired);END;
        ELSE
            byteToMask(sNewState,rulemask);
        END;
        IF same(sInitCLI,"")=FALSE THEN Str.Copy(sInit,sInitCLI);END;
    | gotfile:
        IF testmode THEN abort(errTestMode);END;
        IF rulemask # MAX(CARDINAL) THEN abort(errRuleOrFile);END;
        IF same(sInitCLI,"")=FALSE THEN abort(errInitialOrFile);END;

        (*%T DBG *) dbg("rulefile",rulefile); (*%E  *)
        IF chkJoker(rulefile) THEN abort(errJoker);END;
        Str.Copy(S,rulefile);
        IF Str.CharPos(rulefile,dot)=MAX(CARDINAL) THEN Str.Append(rulefile,extWR);END;
        IF FIO.Exists(rulefile)=FALSE THEN
            Str.Concat(rulefile,S,extTXT);
            IF FIO.Exists(rulefile)=FALSE THEN abort(errNotFound);END;
        END;
        i:=readRule(sPat,sNewState,sInit, useLFN,rulefile);
        IF i # errNone THEN abort(i);END;
    ELSE ; (* //ISO fixed silly requirement *) END;

    IF HiresON() = FALSE THEN ok:=HiresOFF();abort(errGraphMode); END;


    Graph.GetVideoConfig(V);
    xmin:=0; xmax:=V.numxpixels-1;
    ymin:=0; ymax:=V.numypixels-1;

    Graph.ClearScreen(Graph._GCLEARSCREEN);
    initCA(ndx,inkbk,  initwi,xmax,forceonelit,sPat,sNewState,sInit);
    ink:=white;
    ycurr:=ymin;

    LOOP
        ycurr:=showstate(ndx,xmax,ycurr,ink,inkbk,vsync);
        ndx:=processline(ndx,xmax);
        ink:=nextink(ink,ycurr,cycle);

        IF gotkey(ch) THEN
            CASE ch OF
            | space:
                LOOP
                    IF gotkey(ch) THEN EXIT; END;
                END;
            | "+", "-":
                IF testmode THEN
                    IF ch="+" THEN
                        k:=1;
                    ELSE
                        k:=-1;
                    END;
                    rulemask:=maskToByte(sNewState);
                    rulemask:=newvalue(k,rulemask,MIN(SHORTCARD), MAX(SHORTCARD));
                    byteToMask(sNewState,rulemask);

                    Graph.ClearScreen(Graph._GCLEARSCREEN);
                    initCA(ndx,inkbk,  initwi,xmax,forceonelit,sPat,sNewState,sInit);
                    ink:=white;
                    ycurr:=ymin;
                END;
            ELSE
                EXIT;
            END;
        END;
    END;

    IF HiresOFF()= FALSE THEN HALT; END;

    showparms(testmode,rulemask);

    abort(errNone);
END wolfRule32.


