(* ---------------------------------------------------------------
Title         Q&D Solar System
Author        PhG
Overview
Notes
Bugs
Wish List     utiliser flagAbrev dans doRange/DumpCoor,
              tsl, domification, aspects ? remove "" and "," if "." parm?
              rewrite code : aborting from procedures is ugly

--------------------------------------------------------------- *)

MODULE SolarSys32;

IMPORT IO;
IMPORT Lib;
IMPORT Str;
IMPORT MATHLIB;
IMPORT SYSTEM;

FROM IO IMPORT WrStr,WrLn;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed,
reallyeverything,
AltAnimation, str16, getCurrentDirectory,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed;


FROM QD_Names IMPORT sPOURTU;

FROM QD_Sky IMPORT parseDeltaTU, minYear, maxYear,
parseDate,rangeWarning;

FROM MATHLIB IMPORT Cos,Sin,Tan,Sqrt,ATan,ATan2,ASin,Mod;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE fixCLI (VAR S : ARRAY OF CHAR);
CONST
    fr   = "颪뤫";
    frup = "EECAUAEIOUAEIOUY";
VAR
    p,i,len : CARDINAL;
    c,ch  : CHAR;
BEGIN
    len := Str.Length(S);
    FOR i := 1 TO len DO
        c  := S[i-1];
	p:=Str.CharPos(fr,c);
	IF p # MAX(CARDINAL) THEN
	    ch:=frup[p];
	    S[i-1]:=ch;
	END;
    END;
END fixCLI;

(* ------------------------------------------------------------ *)


TYPE
    flag16 = SET OF [0..15]; (* 16 bits *)
CONST
    BITsoleil  = 0; (* bit #, number MUST match enumeration order in Calc module *)
    BITlune    = 1;
    BITmercure = 2;
    BITvenus   = 3;
    BITmars    = 4;
    BITjupiter = 5;
    BITsaturne = 6;
    BITuranus  = 7;
    BITneptune = 8;
    BITpluton  = 9;
    firstplanet = BITsoleil;
    lastplanet  = BITpluton;
CONST
    BITheliocentric = 0; (* l b r *)
    BITecliptic     = 1; (* lambda beta *)
    BITequatorial   = 2; (* alpha delta *)
    BITastro        = 3; (* position in sign *)
    firstoutput = 0;
    lastoutput = 3;
CONST
    BITlongdegmin = 0; (* DDDMM *)
    BITlongdeg    = 1; (* DDDdd *)
    BITlonghms    = 2; (* h mn s *)
    BITlongpad    = 3; (* clear=0, set=space *)
    firstlongfmt= 0;
    lastlongfmt = 2;
CONST
    BITlatdegmin  = 0; (* +DDMM *)
    BITlatdeg     = 1; (* +DDdd *)
    BITlatpadzero = 2; (* clear=space, set=0 *)
    BITlatpadsign = 3; (* clear=space, set=+ *)
    firstlatfmt = 0;
    lastlatfmt =  1;
CONST
	equal		= "=";
	tabul 		= CHR(9);
	semicolon 	= ";";
	dash		= "-";
    now       = "*";
    empty     = "";
    zero      = "0";
    doubledot = ":";
    hourdelim = "H";
    (* DOS chars
    stralpha     = CHR(0E0H);
    strdelta     = CHR(0EBH);
    strbeta      = CHR(0E1H);
    *)
    stralpha     = "alpha";
    strdelta     = "delta";
    strbeta      = "beta";

    strlambda    = "lambda"; (* alas, no lambda char... *)
    plus      = "+";
    sizeLambdaStr  = 6; (* "lambda" *)
    nl             = 015C+012C; (* CHR(13)+CHR(10); *)
CONST
    minInterval = 1;
    maxInterval = 365-1;

(* ------------------------------------------------------------ *)

CONST
    debug         = FALSE;
    ProgEXEname   = "SOLARSYS32";
    ProgEXEsep    = "          "; (* 2 de plus pour 32 *)
    ProgTitle     = "Q&D Solar System (Win32)";
    ProgVersion   = "v0.9 (from v1.1e for DOS)";
    ProgCopyright = "par PhG";
    ProgUtility   = "Calcul de positions plantaires [1600-2200]";
    extEXE        = ".EXE";
CONST
    errNone    = 0;
    errHelp    = 1;
    errCannotBe= 2;
    errUnknown = 3;
    errConversion= 4;
    errNotCardinal=5;
    errDate    = 6;
    errHour    = 7;
    errInterval= 8;
    errDMY = 9;
    errHM = 10;
    errIntervalSpec = 11;
    errIntervalRange = 12;
    errDeltaYears = 13;
    errMissingStart = 14;
    errMissingEnd = 15;
    errNotBoth = 16;
    errNeedBoth = 17;
    errMissingInterval = 18;
    errMissingTime = 19;
    errLongFmt = 20;
    errLatFmt = 21;
    errOnlyOneOutput = 22;
    errLongFmtRepeated = 23;
    errLatFmtRepeated = 24;
    errConversionInt = 25;
    errCannotBeToo = 26;
    errBadEnvVar = 27;
    errCSVfmt = 28;
    errNotMarked     = 254;
    errTampered      = 255;

PROCEDURE abort (e : CARDINAL);
CONST
    helpmsg = nl+
ProgTitle+" "+ProgVersion+" "+ProgCopyright+nl+
nl+
ProgUtility+nl+
nl+
"Syntaxe 1 : "+ProgEXEname+" <d= fi= h= i=> [plante]... [paramtre]..."+nl+
"Syntaxe 2 : "+ProgEXEname+" <d= h=> [plante]... [paramtre]..."+nl+
nl+
"  <d>part=      date de dpart au format jj-mm-aaaa"+nl+
"  <fi>n=         date de fin au format jj-mm-aaaa"+nl+
"  <i>ntervalle=  intervalle en jours ou au format HH:MM"+nl+
"  <d>ate=        date au format jj-mm-aaaa ("+now+"=date actuelle)"+nl+
"  <h>eure=       heure T.U. au format HH:MM ("+now+"=heure actuelle)"+nl+
"  plante(s)     nom de(s) plante(s) (par dfaut : toutes)"+nl+
"                 <so>leil, <lu>ne, <me>rcure, <v>nus, <ma>rs,"+nl+
"                 <ju>piter, <sa>turne, <ur>anus, <ne>ptune, <pl>uton"+nl+
nl+
"  coordonnes    <astro>, <lambda>, <alpha> ou <lbr> (par dfaut : <astro>)"+nl+
"  affichage      FMTLONG=[0]<HMS|DEGMIN|DEG> (par dfaut : 0DEG = ###,##)"+nl+
"                 FMTLAT=[+][0]<DEGMIN|DEG> (par dfaut : 0EG = ##,##)"+nl+
'                 ("0" ou espace, HMS DDDMM DDDdd, "+" ou espace, DDMM DDdd)'+nl+
"  sparateur     CSV=<ESPACE|POINTVIRGULE|TABULATION> (par dfaut : ESPACE)"+nl+
nl+
"Exemples : "+ProgEXEname+" d="+now+" h="+now+nl+
"           "+ProgEXEname+" d=1/1/2400 h=0h /u /alpha"+nl+
"           "+ProgEXEname+" de=1/1/1900 fi=31/12/1900 h=0h i=15 me alpha"+nl+
"           "+ProgEXEsep +" fmtlong=hms fmtlat=+0degmin csv=tabulation"+nl+
nl+
"Dans les paramtres, d'ventuels accents ne sont pas convertis identiquement"+nl+
"en fonction du systme d'exploitation (Win98 ou WinXP) : pour cette raison,"+nl+
"ils sont dconseills."+nl+
nl+
"La variable d'environnement "+sPOURTU+", si elle est dfinie, contient"+nl+
"la correction requise pour passer de l'heure systme  l'heure T.U."+nl+
"Cette correction est applique si et seulement si les date et heure"+nl+
"sont toutes deux spcifies avec "+now+"."+nl+
nl+
"L'option /UNSAFE permet d'ignorer la restriction  l'intervalle [1600..2200],"+nl+
"en faveur de l'intervalle [0..*]. Le recours  cette option est dconseill"+nl+
"en raison des limites de la thorie Van Flandern et Pulkkinen (1979)."+nl+
nl+
"Le code source original, crit pour le compilateur J.P.I. TopSpeed Modula-2,"+nl+
"a t adapt, au prix de nombreuses modifications, au compilateur XDS"+nl+
"avec son trs imparfait TopSpeed Compatibility Kit."+nl+
"L'excutable Win32 devrait fonctionner comme sa version DOS..."+nl+
"mais ce n'est pas garanti !"+nl+
nl+
"Nota bene : quand l'affichage est redirig vers un fichier,"+nl+
"toute squence CRLFCRLF devient CRCRLF sans raison aucune :"+nl+
"le programme CRLF32 permet de remdier  cet irritant mystre."+nl;


VAR
    S : str128;
BEGIN
    CASE e OF
    | errHelp :
        WrStr(helpmsg);
    | errCannotBe :
        S := "How can such things be ?";
    | errUnknown :
        S := "Option imprvue !";
    | errConversion :
        S := "StrToCard error !"; (* converting an empty string for example ! *)
    | errNotCardinal :
        S := "Not a CARDINAL !";
    | errDate :
        S := "Erreur dans la formulation de la date !";
    | errHour :
        S := "Erreur dans la formulation de l'heure !";
    | errInterval :
        S := "Erreur dans la formulation de l'intervalle !";
    | errDMY :
        S := "Valeur illgale pour le jour, le mois ou l'anne !";
    | errHM :
        S := "Valeur illgale pour des heures ou des minutes !";
    | errIntervalSpec :
        S := "Intervalle en jours OU en heures et minutes !";
    | errIntervalRange :
        S := "Delta hors de l'intervalle autoris !";
    | errDeltaYears :
        S := "L'intervalle entre les dates de dpart et de fin est illgal !";
    | errMissingStart :
        S := "La date de dpart manque !";
    | errMissingEnd :
        S := "La date de fin manque !";
    | errNotBoth :
        S := "Intervalle en annes OU date et heure !";
    | errNeedBoth :
        S := "Les date ET heure sont obligatoires !";
    | errMissingInterval :
        S := "L'intervalle manque !";
    | errMissingTime :
        S := "L'heure manque !";
    | errLongFmt :
        S := "Erreur dans la formulation du format d'affichage de la longitude !";
    | errLatFmt :
        S := "Erreur dans la formulation du format d'affichage de la latitude !";
    | errOnlyOneOutput :
        S := "Un seul type de coordonnes  la fois !";
    | errLongFmtRepeated :
        S := "Un seul format d'affichage de la longitude  la fois !";
    | errLatFmtRepeated :
        S := "Un seul format d'affichage de la latitude  la fois !";
    | errConversionInt :
        S := "IntToStr error!";
    | errCannotBeToo :
        S := "How can such things be too ?";
    | errBadEnvVar:
        S:="La valeur de la variable d'environnement "+sPOURTU+" est illgale !";
    | errCSVfmt:
        S := "Erreur dans la formulation du sparateur !";
    | errNotMarked :
        S:="No !";
    | errTampered :
        S:=ProgEXEname+extEXE+" seems to have been tampered !";
    ELSE
        S := "How did you get THERE ?";
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        WrStr(ProgEXEname+" : "); WrStr(S);
    END;
    HALT(e);
END abort;

(* ------------------------------------------------------------ *)

PROCEDURE Print (S : ARRAY OF CHAR );
BEGIN
    WrStr(S);
    WrLn;
END Print;

(* because Str.Subst bugs!!! with several occurrences of the same substring *)
PROCEDURE Replace (VAR s:ARRAY OF CHAR; target:CHAR; new:CHAR );
VAR
    p : CARDINAL;
BEGIN
    LOOP
        p:=Str.CharPos(s,target);
        IF p = MAX(CARDINAL) THEN EXIT; END;
        s[p] := new;
    END;
END Replace;

(* ------------------------------------------------------------------------ *)
(* ------------------------------------------------------------------------ *)


TYPE
    displayfmt = (HHMMSS,DDDdd,DDDMM,DDdd,DDMM,DDd,
                  latDDdd,latDDMM,latDDd,
                  DDddddd);
VAR
    padlongchar    : CHAR; (* "0" or space *)
    padlatchar     : CHAR; (* "0" or space *)
    padlatsignchar : CHAR; (* "+" or space *)
CONST
    blank     = " ";
    minus     = "-";
    dot       = ".";
    coma      = ",";
    degree    = "";
    arcminute = "'";
    strdegree = degree+blank;
    StrHHMM = "h ";
    StrMMSS = "m "; (* this is just for ALPHA ! *)
    StrSS   = "s";

(* ------------------------------------------------------------ *)

PROCEDURE PutSign (VAR S : ARRAY OF CHAR;positive : BOOLEAN);
BEGIN
    IF positive = TRUE THEN
        Str.Prepend (S,padlatsignchar); (* only useful for latitude *)
    ELSE
        Str.Prepend (S,minus);
    END;
END PutSign;

PROCEDURE KeepSign (value : LONGREAL) : BOOLEAN;
BEGIN
    IF value < 0.0 THEN
        RETURN FALSE;
    ELSE
        RETURN TRUE;
    END;
END KeepSign;

PROCEDURE doPad (VAR S : ARRAY OF CHAR; v : LONGINT;
                 digits : CARDINAL; padchar:CHAR);
VAR
    i     : CARDINAL;
    power : LONGINT;
    ok    : BOOLEAN; (* flag for conversion *)
BEGIN
    Str.IntToStr (v,S,10,ok);
    power := 1;
    FOR i := 1 TO (digits - 1) DO
        power := power * 10;
        IF v < power THEN
            Str.Prepend (S,padchar);
        END;
    END;
END doPad;

PROCEDURE doFix (VAR S : ARRAY OF CHAR;padchar:CHAR); (* avoid ---,-0 *)
VAR
    plast,poscoma,i : CARDINAL;
BEGIN
    plast := Str.Length (S) - 1; (* pos of last char but we'll go till char just before *)
    poscoma := Str.CharPos (S,coma); (* we'll patch string directly *)
    DEC (plast);   (* test jusqu'au penultime chiffre *)
    DEC (poscoma); (* test depuis le premier chiffre avant la virgule *)
    FOR i := poscoma TO plast DO
        IF S[i] = padchar THEN
            S[i] := zero; (* replace with a 0 after coma *)
        END;
    END;
END doFix;

(* ------------------------------------------------------------ *)

(* degrees / 15 = hours, then hours * 60 * 60 = seconds *)

PROCEDURE doHHMMSS (val : LONGREAL; VAR s : ARRAY OF CHAR);
VAR
    v,hh,mm,ss : LONGINT;
    strhh,strmm,strss : ARRAY [0..15] OF CHAR;
BEGIN
    v := VAL (LONGINT,(val * 240.0 + 0.5)); (* in seconds now *)
    IF v >= (24 * 60 * 60) THEN
        v := 0;
    END;
    hh := v DIV (60 * 60);
    mm := (v - hh * 60 * 60) DIV 60;
    ss := (v - hh * 60 * 60) MOD 60;
    doPad (strhh,hh,2,padlongchar); (* ## *)
    Str.Append (strhh,StrHHMM);
    doPad (strmm,mm,2,padlongchar); (* ## *)
    Str.Append (strmm,StrMMSS);
    doPad (strss,ss,2,padlongchar); (* ## *)
    Str.Append (strss,StrSS);
    Str.Append (strhh,strmm);
    Str.Concat (s,strhh,strss);
END doHHMMSS;

PROCEDURE doDDDdd (val : LONGREAL; VAR s : ARRAY OF CHAR); (* 0..360 *)
VAR
    v : LONGINT;
BEGIN
    v := VAL (LONGINT,(val * 100.0 + 0.5));
    IF v >= 36000 THEN (* 36000 maximum even if 359.99999999 *)
        v := 0;
    END;
    doPad (s,v,5,padlongchar); (* ##### *)
    Str.Insert (s,coma,3);     (* ###,## *)
    doFix (s,padlongchar);
    Str.Append (s,degree);
END doDDDdd;

PROCEDURE doDDdd (val : LONGREAL; VAR s : ARRAY OF CHAR);
VAR
    v : LONGINT;
BEGIN
    v := VAL (LONGINT,(ABS(val) * 100.0 + 0.5));
    IF v > 9000 THEN (* should never happen *)
        v := 9000;
    END;
    doPad (s,v,4,padlongchar); (* #### *)
    Str.Insert (s,coma,2);     (* ##,## *)
    doFix (s,padlongchar);
    Str.Append (s,degree);
END doDDdd;

PROCEDURE doDDd (val : LONGREAL; VAR s : ARRAY OF CHAR);
VAR
    v : LONGINT;
BEGIN
    v := VAL (LONGINT,(ABS(val) * 10.0 + 0.5));
    IF v > 900 THEN (* should never happen *)
        v := 900;
    END;
    doPad (s,v,3,padlongchar); (* ### *)
    Str.Insert (s,coma,2);     (* ##,# *)
    doFix (s,padlongchar);
    Str.Append (s,degree);
END doDDd;

PROCEDURE doDDDMM (val : LONGREAL; VAR s : ARRAY OF CHAR);
VAR
    v,dd,mm : LONGINT;
    strdd,strmm : ARRAY [0..15] OF CHAR;
BEGIN
    v := VAL (LONGINT,(val * 60.0 + 0.5)); (* in minutes now *)
    IF v >= (360 * 60) THEN
        v := 0;
    END;
    dd := v DIV 60;
    mm := v MOD 60;
    doPad (strdd,dd,3,padlongchar); (* ### *)
    Str.Append (strdd,strdegree);
    doPad (strmm,mm,2,padlongchar); (* ## *)
    Str.Append (strmm,arcminute);
    Str.Concat (s,strdd,strmm);
END doDDDMM;

PROCEDURE doDDMM (val : LONGREAL; VAR s : ARRAY OF CHAR);
VAR
    v,dd,mm : LONGINT;
    strdd,strmm : ARRAY [0..15] OF CHAR;
BEGIN
    v := VAL (LONGINT,(ABS(val) * 60.0 + 0.5)); (* in minutes now *)
    IF v > (90 * 60) THEN (* should never happen *)
        v := 90 * 60;
    END;
    dd := v DIV 60;
    mm := v MOD 60;
    doPad (strdd,dd,2,padlongchar); (* ## *)
    Str.Append (strdd,strdegree);
    doPad (strmm,mm,2,padlongchar); (* ## *)
    Str.Append (strmm,arcminute);
    Str.Concat (s,strdd,strmm);
END doDDMM;

PROCEDURE dolatDDdd (val : LONGREAL; VAR s : ARRAY OF CHAR); (* 0..90 *)
VAR
    positive : BOOLEAN;
    v : LONGINT;
BEGIN
    positive := KeepSign (val);
    v := VAL (LONGINT,(ABS(val) * 100.0 + 0.5));
    IF v > 9000 THEN (* should never happen *)
        v := 9000;
    END;
    doPad (s,v,4,padlatchar); (* #### *)
    Str.Insert (s,coma,2);    (* ##,## *)
    doFix (s,padlatchar);
    Str.Append (s,degree);
    PutSign (s,positive);
END dolatDDdd;

PROCEDURE dolatDDd (val : LONGREAL; VAR s : ARRAY OF CHAR);
VAR
    positive : BOOLEAN;
    v : LONGINT;
BEGIN
    positive := KeepSign (val);
    v := VAL (LONGINT,(ABS(val) * 10.0 + 0.5));
    IF v > 900 THEN (* should never happen *)
        v := 900;
    END;
    doPad (s,v,3,padlatchar); (* ### *)
    Str.Insert (s,coma,2);    (* ##,# *)
    doFix (s,padlatchar);
    Str.Append (s,degree);
    PutSign (s,positive);
END dolatDDd;

PROCEDURE dolatDDMM (val : LONGREAL; VAR s : ARRAY OF CHAR);
VAR
    positive : BOOLEAN;
    v,dd,mm : LONGINT;
    strdd,strmm : ARRAY [0..15] OF CHAR;
BEGIN
    positive := KeepSign (val);
    v := VAL (LONGINT,(ABS(val) * 60.0 + 0.5)); (* in minutes now *)
    IF v > (90 * 60) THEN (* should never happen *)
        v := 90 * 60;
    END;
    dd := v DIV 60;
    mm := v MOD 60;
    doPad (strdd,dd,2,padlatchar); (* ## *)
    Str.Append (strdd,strdegree);
    doPad (strmm,mm,2,padlatchar); (* ## *)
    Str.Append (strmm,arcminute);
    Str.Concat (s,strdd,strmm);
    PutSign (s,positive);
END dolatDDMM;

(* for radius vector *)

PROCEDURE doDDddddd (val : LONGREAL; VAR s : ARRAY OF CHAR);
VAR
    v : LONGINT;
BEGIN
    v := VAL (LONGINT,(val * 100000.0 + 0.5));
    doPad (s,v,7,padlongchar); (* ####### *)
    Str.Insert (s,coma,2);     (* ##,##### *)
    doFix (s,padlongchar);
END doDDddddd;

(* ------------------------------------------------------------ *)

(* angle should be normalized and in degrees ! *)

PROCEDURE AngleToStr (val : LONGREAL; f : displayfmt;
                      VAR s : ARRAY OF CHAR);
BEGIN
    CASE f OF
    | HHMMSS  : doHHMMSS  (val,s); (* ##h ##m ##s *)
    | DDDdd   : doDDDdd   (val,s); (* ###,##     *)
    | DDdd    : doDDdd    (val,s); (* ##,##      *)
    | DDd     : doDDd     (val,s); (* ##,#       *)
    | DDDMM   : doDDDMM   (val,s); (* ### ##'    *)
    | DDMM    : doDDMM    (val,s); (* ## ##'     *)
    | latDDdd : dolatDDdd (val,s); (* ##,##     *)
    | latDDd  : dolatDDd  (val,s); (* ##,#      *)
    | latDDMM : dolatDDMM (val,s); (* ## ##'    *)
    | DDddddd : doDDddddd(val,s); (* ###,#####   *)
    ELSE ; (* //ISO fixed silly requirement *) END;
END AngleToStr;

PROCEDURE PadsSetup(padlong,padlat,padlatsign:CHAR);
BEGIN
    padlongchar   := padlong;
    padlatchar    := padlat;
    padlatsignchar:= padlatsign;
END PadsSetup;


(* ------------------------------------------------------------------------ *)
(* ------------------------------------------------------------------------ *)




TYPE
    objet = (soleil,lune,mercure,venus,mars,
             jupiter,saturne,uranus,neptune,pluton); (* must MATCH bit positions *)
    skypos = RECORD
        l,b,r       : LONGREAL;
        lambda,beta : LONGREAL;
        alpha,delta : LONGREAL;
    END;
    calcmode     = (partial,full);

(* ------------------------------------------------------------ *)

PROCEDURE Int (V : LONGREAL) : LONGREAL;
VAR
    tmpV : LONGINT; (* a LONGINT should be enough here ! *)
BEGIN
    tmpV := VAL (LONGINT,V);
    RETURN VAL (LONGREAL,tmpV);
END Int;

PROCEDURE Frac (V : LONGREAL) : LONGREAL;
BEGIN
    RETURN (V - Int (V));
END Frac;

VAR
    PI         : LONGREAL;
    MaxDegrees : LONGREAL;
    MaxRadians : LONGREAL;
    DegRad     : LONGREAL;

PROCEDURE DegreesToRadians (VAR Angle : LONGREAL);
BEGIN
    Angle := Mod (Angle,MaxDegrees);
    Angle := Angle * DegRad;
END DegreesToRadians;

PROCEDURE SecondsToRadians (VAR Angle : LONGREAL);
BEGIN
    Angle := Mod (Angle / 3600.0,MaxDegrees);
    Angle := Angle * DegRad;
END SecondsToRadians;

PROCEDURE RadiansToDegrees (VAR Angle : LONGREAL);
BEGIN
    Angle := Mod (Angle,MaxRadians);
    Angle := Angle / DegRad;
END RadiansToDegrees;

PROCEDURE Mod360 (VAR Angle : LONGREAL); (* for angles whose range is 0-360 *)
BEGIN
    Angle := Mod (Angle,MaxDegrees); (* should be ALREADY done but... *)
    IF Angle < 0.0 THEN
        Angle := Angle + MaxDegrees;
    Angle := Mod (Angle,MaxDegrees); (* probably useless... *)
    END;
END Mod360;

PROCEDURE ModRadians (VAR Angle : LONGREAL); (* Mod360 in radians *)
BEGIN
    Angle := Mod (Angle,MaxRadians); (* should be ALREADY done but... *)
    IF Angle < 0.0 THEN
        Angle := Angle + MaxRadians;
    Angle := Mod (Angle,MaxRadians); (* probably useless... *)
    END;
END ModRadians;

(* === GLOBAL VARIABLES ======================================= *)

(* On sait, on sait... mais c'est plus rapide comme a ! *)

VAR
  X,Y    : LONGREAL; (* coordonnes rectangulaires du Soleil *)
  l      : LONGREAL; (* longitude hliocentrique vraie *)
  b      : LONGREAL; (* latitude hliocentrique vraie *)
  r      : LONGREAL; (* rayon vecteur *)
  lambda : LONGREAL; (* longitude cliptique -- gocentrique *)
  beta   : LONGREAL; (* latitude cliptique *)
  alpha  : LONGREAL; (* ascension droite -- quatoriale *)
  delta  : LONGREAL; (* declinaison *)

(* === DATES ================================================== *)

(*
    --- don't bother with Dynamical Time : about 10-60 seconds !!!
    --- day already with decimals
    --- correction in already in days with sign
*)

(* assume year will be POSITIVE !!! *)

PROCEDURE JDtoDate (jd:LONGREAL; VAR dd,mm,yyyy,hh,mn : CARDINAL);
VAR
    z,f,a,alf,b,c,d,e,day,fracday : LONGREAL;
    minutes : CARDINAL;
BEGIN
    jd := jd + 0.5;
    z := Int(jd);
    f := Frac(jd);
    IF z < 2299161.0 THEN
        a := z;
    ELSE
        alf := Int( (z - 1867216.25) / 36524.25);
        a := z + 1.0 + alf - Int( alf / 4.0 );
    END;
    b := a + 1524.0;
    c := Int ( (b - 122.1) / 365.25);
    d := Int ( 365.25 * c );
    e := Int ( (b - d) / 30.6001 );
    fracday := b - d - Int( 30.6001 * e) + f;
    day := Int ( fracday );
    dd := CARDINAL (day);
    IF CARDINAL(e) < 14 THEN
        mm := CARDINAL(e)-1;
    ELSE
        mm := CARDINAL(e)-13;
    END;
    IF mm > 2 THEN
        yyyy := CARDINAL(c) - 4716;
    ELSE
        yyyy := CARDINAL(c) - 4715;
    END;
    day := Frac(fracday);
    (* val(cardinal, expression) seems equal to cardinal(expression) *)
    minutes := VAL(CARDINAL, (day * 24.0 * 60.0 +0.5) ); (* fraction of day * hours * minutes *)
    (* compute hh and mn *)
    hh := minutes DIV 60;
    mn := minutes MOD 60;
    IF hh > 23 THEN (* should never happen *)
        hh := 23;
        mn := 59;
    END;
END JDtoDate;

VAR
    T,T2,T3       : LONGREAL;

PROCEDURE DateToJD (Day,Month,Year,DeltaTU : LONGREAL) : LONGREAL;
VAR
    YYYYMMDD  : LONGREAL;
    gregorian : LONGREAL; (* gregorian correction *)
    JD        : LONGREAL;
BEGIN
    YYYYMMDD := Year * 10000.0 + Month * 100.0 + Day;
    IF Month < 2.5 THEN
        Year  := Year - 1.0;
        Month := Month + 12.0;
    END;
    IF YYYYMMDD < 15821015.0 THEN
        gregorian := 0.0;
    ELSE
        gregorian := Int (Year / 100.0);
        gregorian := 2.0 - gregorian + Int (gregorian / 4.0);
    END;
    JD := Int (365.25 * (Year + 4716.0)) + Int (30.6001 * (Month + 1.0));
    JD := JD + Day + gregorian - 1524.5;
    JD := JD + DeltaTU; (* in days with sign : -1/24, etc. *)
    RETURN JD;
END DateToJD;

PROCEDURE CalcT (JD : LONGREAL);
BEGIN
    T  := (JD - 2415020.0) / 36525.0;
    T2 := T * T;
    T3 := T * T2;
END CalcT;

(* === GLOBAL SETUP =========================================== *)

VAR
  L0,M0    : LONGREAL;
  L1,M1,U1 : LONGREAL;
  L2,M2,U2 : LONGREAL;
  L3,M3,U3 : LONGREAL;
  L4,M4,U4 : LONGREAL;
  L5,M5,U5 : LONGREAL;
  L6,M6,U6 : LONGREAL;
  L7,M7,U7 : LONGREAL;
  L8,M8,U8 : LONGREAL;
  L9,M9    : LONGREAL;
  OmegaLune : LONGREAL; (* longitude du noeud ascendant *)
  Epsilon   : LONGREAL; (* obliquit de l'cliptique *)
  NutationLambda  : LONGREAL;
  NutationEpsilon : LONGREAL;
  NutationSoleil  : LONGREAL;

(*
  -- lments moyens des orbites plantaires
     (rapports  l'quinoxe moyen de la date)
  -- L = longitude moyenne
     M = anomalie moyenne
     U = argument de latitude
  -- procdure  excuter AVANT tout autre calcul !!!
*)

PROCEDURE CalcSetup ();
VAR
    OmegaNutation : LONGREAL; (* temporaire et # d'OmegaLune *)
BEGIN
    L0 := 279.6964027 +  36000.7695173 * T;     (* Soleil *)
    M0 := 358.4758635 +  35999.0494965 * T;
    L1 := 178.178814  + 149474.071386  * T;     (* Mercure *)
    M1 := 102.279426  + 149472.515334  * T;
    U1 := 131.032888  + 149472.885872  * T;
    L2 := 342.766738  +  58519.212542  * T;     (* Vnus *)
    M2 := 212.6018923 +  58517.8063877 * T;
    U2 := 266.987445  +  58518.311835  * T;
    L3 := 293.747201  +  19141.699879  * T;     (* Mars *)
    M3 := 319.5292728 +  19139.8588872 * T;
    U3 := 244.960887  +  19140.928953  * T;
    L4 := 237.352259  +   3034.906621  * T;     (* Jupiter *)
    M4 := 225.4445943 +   3034.9066206 * T;     (* 444539 ? *)
    U4 := 138.419219  +   3034.906621  * T;
    L5 := 265.869357  +   1222.116843  * T;     (* Saturne *)
    M5 := 175.758477  +   1222.116843  * T;
    U5 := 153.521637  +   1222.116843  * T;
    L6 := 243.362437  +    429.898403  * T;     (* Uranus *)
    M6 :=  74.313637  +    429.898403  * T;
    U6 := 169.872293  +    429.388747  * T;
    L7 :=  85.024943  +    219.863377  * T;     (* Neptune *)
    M7 :=  41.269103  +    219.863377  * T;
    U7 := 314.346275  +    218.761885  * T;
    L8 :=  92.312712  +    146.674728  * T;     (* Pluton *)
    M8 := 229.488633  +    145.278567  * T;
    U8 := 343.369233  +    145.278567  * T;

    L9 := 270.435378  + 481267.880862  * T;     (* Lune *)
    M9 := 296.095334  + 477198.867586  * T;
    OmegaLune := 259.181303 - 1934.137823 * T;  (* pour phmrides Lune *)

    DegreesToRadians (L0);
    DegreesToRadians (M0);
    DegreesToRadians (L1);
    DegreesToRadians (M1);
    DegreesToRadians (U1);
    DegreesToRadians (L2);
    DegreesToRadians (M2);
    DegreesToRadians (U2);
    DegreesToRadians (L3);
    DegreesToRadians (M3);
    DegreesToRadians (U3);
    DegreesToRadians (L4);
    DegreesToRadians (M4);
    DegreesToRadians (U4);
    DegreesToRadians (L5);
    DegreesToRadians (M5);
    DegreesToRadians (U5);
    DegreesToRadians (L6);
    DegreesToRadians (M6);
    DegreesToRadians (U6);
    DegreesToRadians (L7);
    DegreesToRadians (M7);
    DegreesToRadians (U7);
    DegreesToRadians (L8);
    DegreesToRadians (M8);
    DegreesToRadians (U8);
    DegreesToRadians (L9);
    DegreesToRadians (M9);
    DegreesToRadians (OmegaLune);

    Epsilon   := 23.4522944 - 0.0130125 * T - 0.000001638 * T2;
    DegreesToRadians (Epsilon);

(*
    --- MUST be computed right after CalcSetup and BEFORE CalcSoleil
    --- deltaPSI is too small to consider here for apparent TSL
    --- WAS a separate procedure
*)

    OmegaNutation := 259.18327 - 1934.142 * T;
    DegreesToRadians (OmegaNutation);

    NutationLambda  := - 0.004722 * Sin(OmegaNutation);
    NutationEpsilon := 0.00256 * Cos(OmegaNutation);
    NutationSoleil  := - 0.00569 - 0.00479 * Sin(OmegaNutation);

    DegreesToRadians (NutationLambda);
    DegreesToRadians (NutationEpsilon);
    DegreesToRadians (NutationSoleil);

    Epsilon := Epsilon + NutationEpsilon; (* !!! *)

END CalcSetup;


(* === COORDINATES ============================================ *)

PROCEDURE CalcAlphaDelta (); (* we use global variables here *)
VAR
   numerator,divisor : LONGREAL;
BEGIN
   numerator := Sin(lambda) * Cos(Epsilon) - Tan(beta) * Sin(Epsilon);
   divisor   := Cos(lambda);
   alpha     := ATan2 (divisor,numerator); (* REVERSE LOGIC here !!! *)
   delta := Sin(beta) * Cos(Epsilon) + Cos(beta) * Sin(Epsilon) * Sin(lambda);
   delta := ASin(delta);
END CalcAlphaDelta;

(* === COMMON TO PLANETS ====================================== *)

VAR
    Mode : calcmode; (* astrology or astronomy *)

PROCEDURE Store (VAR P : skypos);
BEGIN
     RadiansToDegrees(lambda);
     Mod360 (lambda);
     P.lambda := lambda;
     IF Mode = full THEN
         RadiansToDegrees(beta);
         RadiansToDegrees(alpha);
         RadiansToDegrees(delta);
         Mod360 (alpha);
         P.beta  := beta;
         P.alpha := alpha;
         P.delta := delta;

         RadiansToDegrees(l);
         RadiansToDegrees(b);
         Mod360(l);
         P.l := l;
         P.b := b;
         P.r := r;
    END;
END Store;

(* === COMMON FOR PLANETS ===================================== *)

PROCEDURE CalcPlanete (VAR P : skypos);
VAR
    x,y,z    : LONGREAL;
    bigdelta : LONGREAL;
BEGIN
    x := X + r * Cos(b) * Cos(l);
    y := Y + r * Cos(b) * Sin(l);
    lambda := ATan2 (x,y); (* arc tangente de Y sur X !!! *)
    lambda := lambda + NutationLambda; (* !!! *)
    IF Mode = full THEN
        z := r * Sin(b);
        bigdelta := Sqrt (x * x + y * y + z * z);
        beta := ASin (z / bigdelta);
        CalcAlphaDelta ();
    END;
    Store (P);
END CalcPlanete;

(* === PLANETS ================================================ *)

(*
  pour chaque plante, calcul des lments l, b, r
  (rapports  l'quinoxe moyen de la date)
*)

(* L1 M1 M2 U1 *)
PROCEDURE CalcMercure (VAR P : skypos);
BEGIN
    l := 8.0 * Sin(M1) * T;
    l := l + 84378.0 * Sin(M1);
    l := l + 10733.0 * Sin(2.0 * M1);
    l := l +  1892.0 * Sin(3.0 * M1);
    l := l +   381.0 * Sin(4.0 * M1);
    l := l +    83.0 * Sin(5.0 * M1);
    l := l +    19.0 * Sin(6.0 * M1);
    l := l -   646.0 * Sin(2.0 * U1);
    l := l -   306.0 * Sin(M1 - 2.0 * U1);
    l := l -   274.0 * Sin(M1 + 2.0 * U1);
    l := l -    92.0 * Sin(2.0 * M1 + 2.0 * U1);
    l := l -    28.0 * Sin(3.0 * M1 + 2.0 * U1);
    l := l +    25.0 * Sin(2.0 * M1 - 2.0 * U1);
    l := l -     9.0 * Sin(4.0 * M1 + 2.0 * U1);
    l := l +     7.0 * Cos(2.0 * M1 - 5.0 * M2);
    SecondsToRadians (l);
    l := L1 + l;
    b := 24134.0 * Sin(U1);
    b := b -   10.0 * Sin(3.0 * U1);
    b := b + 5180.0 * Sin(M1 - U1);
    b := b + 4910.0 * Sin(M1 + U1);
    b := b + 1124.0 * Sin(2.0 * M1 + U1);
    b := b +  271.0 * Sin(3.0 * M1 + U1);
    b := b +  132.0 * Sin(2.0 * M1 - U1);
    b := b +   67.0 * Sin(4.0 * M1 + U1);
    b := b +   18.0 * Sin(3.0 * M1 - U1);
    b := b +   17.0 * Sin(5.0 * M1 + U1);
    b := b -    9.0 * Sin(M1 - 3.0 * U1);
    SecondsToRadians (b);
    r := 0.39528 - 0.07834 * Cos(M1);
    r := r - 0.00795 * Cos(2.0 * M1);
    r := r - 0.00121 * Cos(3.0 * M1);
    r := r - 0.00022 * Cos(4.0 * M1);
    CalcPlanete (P);
END CalcMercure;

(* L2 M2 U2 M0 *)
PROCEDURE CalcVenus (VAR P : skypos);
BEGIN
    l := -20.0 * Sin(M2) * T;
    l := l + 2814.0 * Sin(M2);
    l := l +   12.0 * Sin(2.0 * M2);
    l := l -  181.0 * Sin(2.0 * U2);
    l := l -   10.0 * Cos(2.0 * M0 - 2.0 * M2);
    l := l +    7.0 * Cos(3.0 * M0 - 3.0 * M2);
    SecondsToRadians (l);
    l := L2 + l;
    b := 12215.0 * Sin(U2);
    b := b + 83.0 * Sin(M2 + U2);
    b := b + 83.0 * Sin(M2 - U2);
    SecondsToRadians (b);
    r := 0.72335 - 0.00493 * Cos(M2);
    CalcPlanete (P);
END CalcVenus;

(* L3 M0 M2 M3 M4 U3 *)
PROCEDURE CalcMars (VAR P : skypos);
BEGIN
    l := 37.0 * Sin(M3) * T;
    l := l +     4.0 * Sin(2.0 * M3) * T;
    l := l + 38451.0 * Sin(M3);
    l := l +  2238.0 * Sin(2.0 * M3);
    l := l +   181.0 * Sin(3.0 * M3);
    l := l +    17.0 * Sin(4.0 * M3);
    l := l -    52.0 * Sin(2.0 * U3);
    l := l -    22.0 * Cos(M3 - 2.0 * M4);
    l := l -    19.0 * Sin(M3 - M4);
    l := l +    17.0 * Cos(M3 - M4);
    l := l -    16.0 * Cos(2.0 * M3 - 2.0 * M4);
    l := l +    13.0 * Cos(M0 - 2.0 * M3);
    l := l -    10.0 * Sin(M3 - 2.0 * U3);
    l := l -    10.0 * Sin(M3 + 2.0 * U3);
    l := l +     7.0 * Cos(M0 - M3);
    l := l -     7.0 * Cos(2.0 * M0 - 3.0 * M3);
    l := l -     5.0 * Sin(M2 - 3.0 * M3);
    l := l -     5.0 * Sin(M0 - M3);
    l := l -     5.0 * Sin(M0 - 2.0 * M3);
    l := l -     4.0 * Cos(2.0 * M0 - 4.0 * M3);
    l := l +     4.0 * Cos(M4);
    l := l +     3.0 * Cos(M2 - 3.0 * M3);
    l := l +     3.0 * Sin(2.0 * M3 - 2.0 * M4);
    SecondsToRadians (l);
    l := L3 + l;
    b := 6603.0 * Sin(U3);
    b := b + 622.0 * Sin(M3 - U3);
    b := b + 615.0 * Sin(M3 + U3);
    b := b +  64.0 * Sin(2.0 * M3 + U3);
    SecondsToRadians (b);
    r := 1.53031 - 0.1417 * Cos(M3);
    r := r - 0.0066 * Cos(2.0 * M3);
    r := r - 0.00047 * Cos(3.0 * M3);
    CalcPlanete (P);
END CalcMars;

(* L4 M4 M5 M6 *)
PROCEDURE CalcJupiter (VAR P : skypos);
BEGIN
    l := 2511.0 + 5023.0 * T;
    l := l + 19934.0 * Sin(M4);
    l := l +   601.0 * Sin(2.0 * M4);
    l := l +  1093.0 * Cos(2.0 * M4 - 5.0 * M5);
    l := l -   479.0 * Sin(2.0 * M4 - 5.0 * M5);
    l := l -   185.0 * Sin(2.0 * M4 - 2.0 * M5);
    l := l +   137.0 * Sin(3.0 * M4 - 5.0 * M5);
    l := l -   131.0 * Sin(M4 - 2.0 * M5);
    l := l +    79.0 * Cos(M4 - M5);
    l := l -    76.0 * Cos(2.0 * M4 - 2.0 * M5);
    l := l -    74.0 * Cos(M4) * T;
    l := l +    68.0 * Sin(M4) * T;
    l := l +    66.0 * Cos(2.0 * M4 - 3.0 * M5);
    l := l +    63.0 * Cos(3.0 * M4 - 5.0 * M5);
    l := l +    53.0 * Cos(M4 - 5.0 * M5);
    l := l +    49.0 * Sin(2.0 * M4 - 3.0 * M5);
    l := l -    43.0 * Sin(2.0 * M4 - 5.0 * M5) * T;
    l := l -    37.0 * Cos(M4);
    l := l +    25.0 * Sin(2.0 * L4);
    l := l +    25.0 * Sin(3.0 * M4);
    l := l -    23.0 * Sin(M4 - 5.0 * M5);
    l := l -    19.0 * Cos(2.0 * M4 - 5.0 * M5) * T;
    l := l +    17.0 * Cos(2.0 * M4 - 4.0 * M5);
    l := l +    17.0 * Cos(3.0 * M4 - 3.0 * M5);
    l := l -    14.0 * Sin(M4 - M5);
    l := l -    13.0 * Sin(3.0 * M4 - 4.0 * M5);
    l := l -     9.0 * Cos(2.0 * L4);
    l := l +     9.0 * Cos(M5);
    l := l -     9.0 * Sin(M5);
    l := l -     9.0 * Sin(3.0 * M4 - 2.0 * M5);
    l := l +     9.0 * Sin(4.0 * M4 - 5.0 * M5);
    l := l +     9.0 * Sin(2.0 * M4 - 6.0 * M5 + 3.0 * M6);
    l := l -     8.0 * Cos(4.0 * M4 - 10.0 * M5);
    l := l +     7.0 * Cos(3.0 * M4 - 4.0 * M5);
    l := l -     7.0 * Cos(M4 - 3.0 * M5);
    l := l -     7.0 * Sin(4.0 * M4 - 10.0 * M5);
    l := l -     7.0 * Sin(M4 - 3.0 * M5);
    l := l +     6.0 * Cos(4.0 * M4 - 5.0 * M5);
    l := l -     6.0 * Sin(3.0 * M4 - 3.0 * M5);
    l := l +     5.0 * Cos(2.0 * M5);
    l := l -     4.0 * Sin(4.0 * M4 - 4.0 * M5);
    l := l -     4.0 * Cos(3.0 * M5);
    l := l +     4.0 * Cos(2.0 * M4 - M5);
    l := l -     4.0 * Cos(3.0 * M4 - 2.0 * M5);
    l := l -     4.0 * Cos(2.0 * M4) * T;
    l := l +     3.0 * Sin(2.0 * M4) * T;
    l := l +     3.0 * Cos(5.0 * M5);
    l := l +     3.0 * Cos(5.0 * M4 - 10.0 * M5);
    l := l +     3.0 * Sin(2.0 * M5);
    l := l -     2.0 * Sin(2.0 * L4 - M4);
    l := l +     2.0 * Sin(2.0 * L4 + M4);
    l := l -     2.0 * Sin(3.0 * M4 - 5.0 * M5) * T;
    l := l -     2.0 * Sin(M4 - 5.0 * M5) * T;
    SecondsToRadians (l);
    l := L4 + l;
    b := -4692.0 * Cos(M4);
    b := b + 259.0 * Sin(M4);
    b := b + 227.0;
    b := b - 227.0 * Cos(2.0 * M4);
    b := b +  30.0 * Sin(M4) * T;
    b := b +  21.0 * Cos(M4) * T;
    b := b +  16.0 * Sin(3.0 * M4 - 5.0 * M5);
    b := b -  13.0 * Sin(M4 - 5.0 * M5);
    b := b -  12.0 * Cos(3.0 * M4);
    b := b +  12.0 * Sin(2.0 * M4);
    b := b +   7.0 * Cos(3.0 * M4 - 5.0 * M5);
    b := b -   5.0 * Cos(M4 - 5.0 * M5);
    SecondsToRadians (b);
    r := 5.20883 - 0.25122 * Cos(M4);
    r := r - 0.00604 * Cos(2.0 * M4);
    r := r + 0.0026  * Cos(2.0 * M4 - 2.0 * M5);
    r := r - 0.0017  * Cos(3.0 * M4 - 5.0 * M5);
    r := r - 0.0016  * Sin(2.0 * M4 - 2.0 * M5);
    r := r - 0.00091 * Sin(M4) * T;
    r := r - 0.00084 * Cos(M4) * T;
    r := r + 0.00069 * Sin(2.0 * M4 - 3.0 * M5);
    r := r - 0.00067 * Sin(M4 - 5.0 * M5);
    r := r + 0.00066 * Sin(3.0 * M4 - 5.0 * M5);
    r := r + 0.00063 * Sin(M4 - M5);
    r := r - 0.00051 * Cos(2.0 * M4 - 3.0 * M5);
    r := r - 0.00046 * Sin(M4);
    r := r - 0.00029 * Cos(M4 - 5.0 * M5);
    r := r + 0.00027 * Cos(M4 - 2.0 * M5);
    r := r - 0.00022 * Cos(3.0 * M4);
    r := r - 0.00021 * Sin(2.0 * M4 - 5.0 * M5);
    CalcPlanete (P);
END CalcJupiter;

(* M4 M5 M6 L5 *)
PROCEDURE CalcSaturne (VAR P : skypos);
BEGIN
    l := 2507.0 + 5014.0 * T;
    l := l + 23043.0 * Sin(M5);
    l := l -  2689.0 * Cos(2.0 * M4 - 5.0 * M5);
    l := l +  1177.0 * Sin(2.0 * M4 - 5.0 * M5);
    l := l -   826.0 * Cos(2.0 * M4 - 4.0 * M5);
    l := l +   802.0 * Sin(2.0 * M5);
    l := l +   425.0 * Sin(M4 - 2.0 * M5);
    l := l -   229.0 * Cos(M5) * T;
    l := l -   142.0 * Sin(M5) * T;
    l := l -   153.0 * Cos(2.0 * M4 - 6.0 * M5);
    l := l -   114.0 * Cos(M5);
    l := l +   101.0 * Sin(2.0 * M4 - 5.0 * M5) * T;
    l := l -    70.0 * Cos(2.0 * L5);
    l := l +    67.0 * Sin(2.0 * L5);
    l := l +    66.0 * Sin(2.0 * M4 - 6.0 * M5);
    l := l +    60.0 * Cos(2.0 * M4 - 5.0 * M5) * T;
    l := l +    41.0 * Sin(M4 - 3.0 * M5);
    l := l +    39.0 * Sin(3.0 * M5);
    l := l +    31.0 * Sin(M4 - M5);
    l := l +    31.0 * Sin(2.0 * M4 - 2.0 * M5);
    l := l -    29.0 * Cos(2.0 * M4 - 3.0 * M5);
    l := l -    28.0 * Sin(2.0 * M4 - 6.0 * M5 + 3.0 * M6);
    l := l +    28.0 * Cos(M4 - 3.0 * M5);
    l := l +    22.0 * Sin(2.0 * M4 - 4.0 * M5) * T;
    l := l -    22.0 * Sin(M5 - 3.0 * M6);
    l := l +    20.0 * Sin(2.0 * M4 - 3.0 * M5);
    l := l +    20.0 * Cos(4.0 * M4 - 10.0 * M5);
    l := l +    19.0 * Cos(2.0 * M5 - 3.0 * M6);
    l := l +    19.0 * Sin(4.0 * M4 - 10.0 * M5);
    l := l -    17.0 * Cos(2.0 * M5) * T;
    l := l -    16.0 * Cos(M5 - 3.0 * M6);
    l := l -    12.0 * Sin(2.0 * M4 - 4.0 * M5);
    l := l +    12.0 * Cos(M4);
    l := l -    12.0 * Sin(2.0 * M5 - 2.0 * M6);
    l := l -    11.0 * Sin(2.0 * M5) * T;
    l := l -    11.0 * Cos(2.0 * M4 - 7.0 * M5);
    l := l +    10.0 * Sin(2.0 * M5 - 3.0 * M6);
    l := l +    10.0 * Cos(2.0 * M4 - 2.0 * M5);
    l := l +     9.0 * Sin(4.0 * M4 - 9.0 * M5);
    l := l -     8.0 * Sin(M5 - 2.0 * M6);
    l := l -     8.0 * Cos(2.0 * L5 + M5);
    l := l +     8.0 * Cos(2.0 * L5 - M5);
    l := l +     8.0 * Cos(M5 - M6);
    l := l -     8.0 * Sin(2.0 * L5 - M5);
    l := l +     7.0 * Sin(2.0 * L5 + M5);
    l := l -     7.0 * Cos(M4 - 2.0 * M5);
    l := l -     7.0 * Cos(2.0 * M5);
    l := l -     6.0 * Sin(4.0 * M4 - 10.0 * M5) * T;
    l := l +     6.0 * Cos(4.0 * M4 - 10.0 * M5) * T;
    l := l +     6.0 * Sin(2.0 * M4 - 6.0 * M5) * T;
    l := l -     5.0 * Sin(3.0 * M4 - 7.0 * M5);
    l := l -     5.0 * Cos(3.0 * M4 - 3.0 * M5);
    l := l -     5.0 * Cos(2.0 * M5 - 2.0 * M6);
    l := l +     5.0 * Sin(3.0 * M4 - 4.0 * M5);
    l := l +     5.0 * Sin(2.0 * M4 - 7.0 * M5);
    l := l +     4.0 * Sin(3.0 * M4 - 3.0 * M5);
    l := l +     4.0 * Sin(3.0 * M4 - 5.0 * M5);
    l := l +     4.0 * Cos(M4 - 3.0 * M5) * T;
    l := l +     3.0 * Cos(2.0 * M4 - 4.0 * M5) * T;
    l := l +     3.0 * Cos(2.0 * M4 - 6.0 * M5 + 3.0 * M6);
    l := l -     3.0 * Sin(2.0 * L5) * T;
    l := l +     3.0 * Cos(2.0 * M4 - 6.0 * M5) * T;
    l := l -     3.0 * Cos(2.0 * L5) * T;
    l := l +     3.0 * Cos(3.0 * M4 - 7.0 * M5);
    l := l +     3.0 * Cos(4.0 * M4 - 9.0 * M5);
    l := l +     3.0 * Sin(3.0 * M4 - 6.0 * M5);
    l := l +     3.0 * Sin(2.0 * M4 - M5);
    l := l +     3.0 * Sin(M4 - 4.0 * M5);
    l := l +     2.0 * Cos(3.0 * M5 - 3.0 * M6);
    l := l +     2.0 * Sin(M4 - 2.0 * M5) * T;
    l := l +     2.0 * Sin(4.0 * M5);
    l := l -     2.0 * Cos(3.0 * M4 - 4.0 * M5);
    l := l -     2.0 * Cos(2.0 * M4 - M5);
    l := l -     2.0 * Sin(2.0 * M4 - 7.0 * M5 + 3.0 * M6);
    l := l +     2.0 * Cos(M4 - 4.0 * M5);
    l := l +     2.0 * Cos(4.0 * M4 - 11.0 * M5);
    l := l -     2.0 * Sin(M5 - M6);
    SecondsToRadians (l);
    l := L5 + l;
    b := 185.0 + 8297.0 * Sin(M5);
    b := b - 3346.0 * Cos(M5);
    b := b +  462.0 * Sin(2.0 * M5);
    b := b -  189.0 * Cos(2.0 * M5);
    b := b +   79.0 * Cos(M5) * T;
    b := b -   71.0 * Cos(2.0 * M4 - 4.0 * M5);
    b := b +   46.0 * Sin(2.0 * M4 - 6.0 * M5);
    b := b -   45.0 * Cos(2.0 * M4 - 6.0 * M5);
    b := b +   29.0 * Sin(3.0 * M5);
    b := b -   20.0 * Cos(2.0 * M4 - 3.0 * M5);
    b := b +   18.0 * Sin(M5) * T;
    b := b -   14.0 * Cos(2.0 * M4 - 5.0 * M5);
    b := b -   11.0 * Cos(3.0 * M5);
    b := b -   10.0 * T;
    b := b +    9.0 * Sin(M4 - 3.0 * M5);
    b := b +    8.0 * Sin(M4 - M5);
    b := b -    6.0 * Sin(2.0 * M4 - 3.0 * M5);
    b := b +    5.0 * Sin(2.0 * M4 - 7.0 * M5);
    b := b -    5.0 * Cos(2.0 * M4 - 7.0 * M5);
    b := b +    4.0 * Sin(2.0 * M4 - 5.0 * M5);
    b := b -    4.0 * Sin(2.0 * M5) * T;
    b := b -    3.0 * Cos(M4 - M5);
    b := b +    3.0 * Cos(M4 - 3.0 * M5);
    b := b +    3.0 * Sin(2.0 * M4 - 4.0 * M5) * T;
    b := b +    3.0 * Sin(M4 - 2.0 * M5);
    b := b +    2.0 * Sin(4.0 * M5);
    b := b -    2.0 * Cos(2.0 * M4 - 2.0 * M5);
    SecondsToRadians (b);
    r := 9.55774 - 0.00028 * T;
    r := r - 0.53252 * Cos(M5);
    r := r - 0.01878 * Sin(2.0 * M4 - 4.0 * M5);
    r := r - 0.01482 * Cos(2.0 * M5);
    r := r + 0.00817 * Sin(M4 - M5);
    r := r - 0.00539 * Cos(M4 - 2.0 * M5);
    r := r - 0.00524 * Sin(M5) * T;
    r := r + 0.00349 * Sin(2.0 * M4 - 5.0 * M5);
    r := r + 0.00347 * Sin(2.0 * M4 - 6.0 * M5);
    r := r + 0.00328 * Cos(M5) * T;
    r := r - 0.00225 * Sin(M5);
    r := r + 0.00149 * Cos(2.0 * M4 - 6.0 * M5);
    r := r - 0.00126 * Cos(2.0 * M4 - 2.0 * M5);
    r := r + 0.00104 * Cos(M4 - M5);
    r := r + 0.00101 * Cos(2.0 * M4 - 5.0 * M5);
    r := r + 0.00098 * Cos(M4 - 3.0 * M5);
    r := r - 0.00073 * Cos(2.0 * M4 - 3.0 * M5);
    r := r - 0.00062 * Cos(3.0 * M5);
    r := r + 0.00042 * Sin(2.0 * M5 - 3.0 * M6);
    r := r + 0.00041 * Sin(2.0 * M4 - 2.0 * M5);
    r := r - 0.0004  * Sin(M4 - 3.0 * M5);
    r := r + 0.0004  * Cos(2.0 * M4 - 4.0 * M5);
    r := r - 0.00023 * Sin(M4);
    r := r + 0.0002  * Sin(2.0 * M4 - 7.0 * M5);
    CalcPlanete (P);
END CalcSaturne;

(* L6 M4 M5 M6 M7 U6 T2 *)
PROCEDURE CalcUranus (VAR P : skypos);
BEGIN
    l := 32.0 * T2;
    l := l + 19397.0 * Sin(M6);
    l := l +   570.0 * Sin(2.0 * M6);
    l := l -   536.0 * Cos(M6) * T;
    l := l +   143.0 * Sin(M5 - 2.0 * M6);
    l := l +   110.0 * Sin(M6) * T;
    l := l +   102.0 * Sin(M5 - 3.0 * M6);
    l := l +    76.0 * Cos(M5 - 3.0 * M6);
    l := l -    49.0 * Sin(M4 - M6);
    l := l -    30.0 * Cos(2.0 * M6) * T;
    l := l +    29.0 * Sin(2.0 * M4 - 6.0 * M5 + 3.0 * M6);
    l := l +    29.0 * Cos(2.0 * M6 - 2.0 * M7);
    l := l -    28.0 * Cos(M6 - M7);
    l := l +    23.0 * Sin(3.0 * M6);
    l := l -    21.0 * Cos(M4 - M6);
    l := l +    20.0 * Sin(M6 - M7);
    l := l +    20.0 * Cos(M5 - 2.0 * M6);
    l := l -    19.0 * Cos(M5 - M6);
    l := l +    17.0 * Sin(2.0 * M6 - 3.0 * M3);
    l := l +    14.0 * Sin(3.0 * M6 - 3.0 * M7);
    l := l +    13.0 * Sin(M5 - M6);
    l := l -    12.0 * Cos(M6) * T2;
    l := l -    12.0 * Cos(M6);
    l := l +    10.0 * Sin(2.0 * M6 - 2.0 * M7);
    l := l -     9.0 * Sin(2.0 * U6);
    l := l -     9.0 * Sin(M6) * T2;
    l := l +     9.0 * Cos(2.0 * M6 - 3.0 * M7);
    l := l +     8.0 * Cos(M5 - 2.0 * M6) * T;
    l := l +     7.0 * Cos(M5 - 3.0 * M6) * T;
    l := l -     7.0 * Sin(M5 - 3.0 * M6) * T;
    l := l +     7.0 * Sin(2.0 * M6) * T;
    l := l +     6.0 * Sin(2.0 * M4 - 6.0 * M5 + 2.0 * M6);
    l := l +     6.0 * Cos(2.0 * M4 - 6.0 * M5 + 2.0 * M6);
    l := l +     5.0 * Sin(M5 - 4.0 * M6);
    l := l -     4.0 * Sin(3.0 * M6 - 4.0 * M7);
    l := l +     4.0 * Cos(3.0 * M6 - 3.0 * M7);
    l := l -     3.0 * Cos(M7);
    l := l -     2.0 * Sin(M7);
    SecondsToRadians (l);
    l := L6 + l;
    b := 2775.0 * Sin(U6);
    b := b + 131.0 * Sin(M6 - U6);
    b := b + 130.0 * Sin(M6 + U6);
    SecondsToRadians (b);
    r := 19.21216 - 0.90154 * Cos(M6);
    r := r - 0.02488 * Sin(M6) * T;
    r := r - 0.02121 * Cos(2.0 * M6);
    r := r - 0.00585 * Cos(M5 - 2.0 * M6);
    r := r - 0.00508 * Cos(M6) * T;
    r := r - 0.00451 * Cos(M4 - M6);
    r := r + 0.00336 * Sin(M5 - M6);
    r := r + 0.00198 * Sin(M4 - M6);
    r := r + 0.00118 * Cos(M5 - 3.0 * M6);
    r := r + 0.00107 * Sin(M5 - 2.0 * M6);
    r := r - 0.00103 * Sin(2.0 * M6) * T;
    r := r - 0.00081 * Cos(3.0 * M6 - 3.0 * M7);
    CalcPlanete (P);
END CalcUranus;

(* L6 L7 M4 M5 M6 M7 U7 T2 *)
PROCEDURE CalcNeptune (VAR P : skypos);
BEGIN
    l := 3523.0 * Sin(M7);
    l := l - 50.0 * Sin(2.0 * U7);
    l := l - 43.0 * Cos(M7) * T;
    l := l + 29.0 * Sin(M4 - M7);
    l := l + 19.0 * Sin(2.0 * M7);
    l := l - 18.0 * Cos(M4 - M7);
    l := l + 13.0 * Cos(M5 - M7);
    l := l + 13.0 * Sin(M5 - M7);
    l := l -  9.0 * Sin(2.0 * M6 - 3.0 * M7);
    l := l +  9.0 * Cos(2.0 * M6 - 2.0 * M7);
    l := l -  5.0 * Cos(2.0 * M6 - 3.0 * M7);
    l := l -  4.0 * Sin(M7) * T;
    l := l +  4.0 * Cos(M6 - 2.0 * M7);
    l := l +  4.0 * Sin(M7) * T2;
    SecondsToRadians (l);
    l := L7 + l;
    b := 6404.0 * Sin(U7);
    b := b + 55.0 * Sin(M7 + U7);
    b := b + 55.0 * Sin(M7 - U7);
    b := b - 33.0 * Sin(U7) * T;
    SecondsToRadians (b);
    r := 30.07175 - 0.25701 * Cos(M7);
    r := r - 0.00787 * Cos(2.0 * L6 - M6 - 2.0 * L7);
    r := r + 0.00409 * Cos(M4 - M7);
    r := r - 0.00314 * Sin(M7) * T;
    r := r + 0.0025  * Sin(M4 - M7);
    r := r - 0.00194 * Sin(M5 - M7);
    r := r + 0.00185 * Cos(M5 - M7);
    CalcPlanete (P);
END CalcNeptune;

(* L8 M8 U8 T2 *)
PROCEDURE CalcPluton (VAR P : skypos);
BEGIN
    l := 101577.0 * Sin(M8);
    l := l + 15517.0 * Sin(2.0 * M8);
    l := l -  3593.0 * Sin(2.0 * U8);
    l := l +  3414.0 * Sin(3.0 * M8);
    l := l -  2201.0 * Sin(M8 - 2.0 * U8);
    l := l -  1871.0 * Sin(M8 + 2.0 * U8);
    l := l +   839.0 * Sin(4.0 * M8);
    l := l -   757.0 * Sin(2.0 * M8 + 2.0 * U8);
    l := l -   285.0 * Sin(3.0 * M8 + 2.0 * U8);
    l := l +   227.0 * Sin(M8) * T2;
    l := l +   218.0 * Sin(2.0 * M8 - 2.0 * U8);
    l := l +   200.0 * Sin(M8) * T;
    SecondsToRadians (l);
    l := L8 + l;
    b := 57726.0 * Sin(U8);
    b := b + 15257.0 * Sin(M8 - U8);
    b := b + 14102.0 * Sin(M8 + U8);
    b := b +  3870.0 * Sin(2.0 * M8 + U8);
    b := b +  1138.0 * Sin(3.0 * M8 + U8);
    b := b +   472.0 * Sin(2.0 * M8 - U8);
    b := b +   353.0 * Sin(4.0 * M8 + U8);
    b := b -   144.0 * Sin(M8 - 3.0 * U8);
    b := b -   119.0 * Sin(3.0 * U8);
    b := b -   111.0 * Sin(M8 + 3.0 * U8);
    SecondsToRadians (b);
    r := 40.74638 - 9.58235 * Cos(M8);
    r := r - 1.16703 * Cos(2.0 * M8);
    r := r - 0.22649 * Cos(3.0 * M8);
    r := r - 0.04996 * Cos(4.0 * M8);
    CalcPlanete (P);
END CalcPluton;

(* === MOON =================================================== *)

PROCEDURE CalcLune (VAR P : skypos);
VAR
  F : LONGREAL; (* distance angulaire de la Lune moyenne au noeud ascendant *)
  D : LONGREAL; (* longation moyenne *)
  M : LONGREAL; (* alias for M9 in original *)
BEGIN
    F :=  11.254075 + 483202.018685 * T;
    D := 350.738975 + 445267.111345 * T;
    DegreesToRadians (F);
    DegreesToRadians (D);
    M := M9; (* suis flemmard, l... *)
    lambda := 22640.0 * Sin(M);
    lambda := lambda - 4586.0 * Sin(M - 2.0 * D);
    lambda := lambda + 2370.0 * Sin(2.0 * D);
    lambda := lambda +  769.0 * Sin(2.0 * M);
    lambda := lambda -  668.0 * Sin(M0);
    lambda := lambda -  412.0 * Sin(2.0 * F);
    lambda := lambda -  212.0 * Sin(2.0 * M - 2.0 * D);
    lambda := lambda -  206.0 * Sin(M - 2.0 * D + M0);
    lambda := lambda +  192.0 * Sin(M + 2.0 * D);
    lambda := lambda +  165.0 * Sin(2.0 * D - M0);
    lambda := lambda +  148.0 * Sin(M - M0);
    lambda := lambda -  125.0 * Sin(D);
    lambda := lambda -  110.0 * Sin(M + M0);
    lambda := lambda -   55.0 * Sin(2.0 * F - 2.0 * D);
    lambda := lambda -   45.0 * Sin(M + 2.0 * F);
    lambda := lambda +   40.0 * Sin(M - 2.0 * F);
    lambda := lambda -   38.0 * Sin(M - 4.0 * D);
    lambda := lambda +   36.0 * Sin(3.0 * M);
    lambda := lambda -   31.0 * Sin(2.0 * M - 4.0 * D);
    lambda := lambda +   28.0 * Sin(M - 2.0 * D - M0);
    lambda := lambda -   24.0 * Sin(2.0 * D + M0);
    lambda := lambda +   19.0 * Sin(M - D);
    lambda := lambda +   18.0 * Sin(D + M0);
    lambda := lambda +   15.0 * Sin(M + 2.0 * D - M0);
    lambda := lambda +   14.0 * Sin(2.0 * M + 2.0 * D);
    lambda := lambda +   14.0 * Sin(4.0 * D);
    lambda := lambda -   13.0 * Sin(3.0 * M - 2.0 * D);
    lambda := lambda -   11.0 * Sin(M + 16.0 * L0 - 18.0 * L2);
    lambda := lambda +   10.0 * Sin(2.0 * M - M0);
    lambda := lambda +    9.0 * Sin(M - 2.0 * F - 2.0 * D);
    lambda := lambda +    9.0 * Cos(M + 16.0 * L0 - 18.0 * L2);
    lambda := lambda -    9.0 * Sin(2.0 * M - 2.0 * D + M0);
    lambda := lambda -    8.0 * Sin(M + D);
    lambda := lambda +    8.0 * Sin(2.0 * D - 2.0 * M0);
    lambda := lambda -    8.0 * Sin(2.0 * M + M0);
    lambda := lambda -    7.0 * Sin(2.0 * M0);
    lambda := lambda -    7.0 * Sin(M - 2.0 * D + 2.0 * M0);
    lambda := lambda +    7.0 * Sin(OmegaLune);
    lambda := lambda -    6.0 * Sin(M - 2.0 * F + 2.0 * D);
    lambda := lambda -    6.0 * Sin(2.0 * F + 2.0 * D);
    lambda := lambda -    4.0 * Sin(M - 4.0 * D + M0);
    lambda := lambda +    4.0 * Cos(M + 16.0 * L0 - 18.0 * L2) * T;
    lambda := lambda -    4.0 * Sin(2.0 * M + 2.0 * F);
    lambda := lambda +    4.0 * Sin(M + 16.0 * L0 - 18.0 * L2) * T;
    lambda := lambda +    3.0 * Sin(M - 3.0 * D);
    lambda := lambda -    3.0 * Sin(M + 2.0 * D + M0);
    lambda := lambda -    3.0 * Sin(2.0 * M - 4.0 * D + M0);
    lambda := lambda +    3.0 * Sin(M - 2.0 * M0);
    lambda := lambda +    3.0 * Sin(M - 2.0 * D - 2.0 * M0);
    lambda := lambda -    2.0 * Sin(2.0 * M - 2.0 * D - M0);
    lambda := lambda -    2.0 * Sin(2.0 * F - 2.0 * D + M0);
    lambda := lambda +    2.0 * Sin(M + 4.0 * D);
    lambda := lambda +    2.0 * Sin(4.0 * M);
    lambda := lambda +    2.0 * Sin(4.0 * D - M0);
    lambda := lambda +    2.0 * Sin(2.0 * M - D);
    SecondsToRadians (lambda);
    lambda := L9 + lambda;
    lambda := lambda + NutationLambda; (* !!! *)
    IF Mode = full THEN
        beta := 18461.0 * Sin(F);
        beta := beta + 1010.0 * Sin(M + F);
        beta := beta + 1000.0 * Sin(M - F);
        beta := beta -  624.0 * Sin(F - 2.0 * D);
        beta := beta -  199.0 * Sin(M - F - 2.0 * D);
        beta := beta -  167.0 * Sin(M + F - 2.0 * D);
        beta := beta +  117.0 * Sin(F + 2.0 * D);
        beta := beta +   62.0 * Sin(2.0 * M + F);
        beta := beta +   33.0 * Sin(M - F + 2.0 * D);
        beta := beta +   32.0 * Sin(2.0 * M - F);
        beta := beta -   30.0 * Sin(F - 2.0 * D + M0);
        beta := beta -   16.0 * Sin(2.0 * M + F - 2.0 * D);
        beta := beta +   15.0 * Sin(M + F + 2.0 * D);
        beta := beta +   12.0 * Sin(F - 2.0 * D - M0);
        beta := beta -    9.0 * Sin(M - F - 2.0 * D + M0);
        beta := beta -    8.0 * Sin(F + OmegaLune);
        beta := beta +    8.0 * Sin(F + 2.0 * D - M0);
        beta := beta -    7.0 * Sin(M + F - 2.0 * D + M0);
        beta := beta +    7.0 * Sin(M + F - M0);
        beta := beta -    7.0 * Sin(M + F - 4.0 * D);
        beta := beta -    6.0 * Sin(F + M0);
        beta := beta -    6.0 * Sin(3.0 * F);
        beta := beta +    6.0 * Sin(M - F - M0);
        beta := beta -    5.0 * Sin(F + D);
        beta := beta -    5.0 * Sin(M + F + M0);
        beta := beta -    5.0 * Sin(M - F + M0);
        beta := beta +    5.0 * Sin(F - M0);
        beta := beta +    5.0 * Sin(F - D);
        beta := beta +    4.0 * Sin(3.0 * M + F);
        beta := beta -    4.0 * Sin(F - 4.0 * D);
        beta := beta -    3.0 * Sin(M - F - 4.0 * D);
        beta := beta +    3.0 * Sin(M - 3.0 * F);
        beta := beta -    2.0 * Sin(2.0 * M - F - 4.0 * D);
        beta := beta -    2.0 * Sin(3.0 * F - 2.0 * D);
        beta := beta +    2.0 * Sin(2.0 * M - F + 2.0 * D);
        beta := beta +    2.0 * Sin(M - F + 2.0 * D - M0);
        beta := beta +    2.0 * Sin(2.0 * M - F - 2.0 * D);
        beta := beta +    2.0 * Sin(3.0 * M - F);
        SecondsToRadians (beta);
        CalcAlphaDelta ();
    END;
    Store (P);
END CalcLune;

(* === SUN ==================================================== *)

(* v1.1 : attention ! faut toujours au moins faire le soleil *)
(* TOUJOURS A FAIRE MEME POUR UNE SEULE PLANETE !!! *)


(* L0 M0 M2 M3 M4 L9 T2 *)
PROCEDURE CalcSoleil (VAR P : skypos);
VAR
    e,e2,e3 : LONGREAL; (* eccentricit *)
BEGIN
    lambda := (1.919444 - 0.004722 * T) * Sin(M0);
    lambda := lambda + 0.02     * Sin(2.0 * M0);
    lambda := lambda - 0.001944 * Cos(M0 - M4);
    lambda := lambda + 0.001666 * Sin(L9 - L0);
    lambda := lambda + 0.001388 * Sin(4.0 * M0 - 8.0 * M3 + M4);
    lambda := lambda - 0.001388 * Cos(2.0 * M0 - 2.0 * M2);
    lambda := lambda - 0.001111 * Sin(M0 - M2);
    lambda := lambda + 0.001111 * Cos(4.0 * M0 - 8.0 * M3 + M4);
    lambda := lambda + 0.000833 * Sin(2.0 * M0 - 2.0 * M2);
    lambda := lambda - 0.000833 * Sin(M4);
    lambda := lambda - 0.000833 * Sin(2.0 * M0 - 2.0 * M4);
    DegreesToRadians (lambda);
    lambda := L0 + lambda;
    e  := 0.01675104 - 0.0000418 * T - 0.000000126 * T2;
    e2 := e * e;
    e3 := e * e2;
    r := 1.0 + e2 / 2.0;
    r := r - (e - e3 * 3.0 / 8.0) * Cos(M0);
    r := r - e2 / 2.0 * Cos(2.0 * M0);
    r := r - e3 * 3.0 / 8.0 * Cos(3.0 * M0);
    r := 1.0000002 * r;

(*  lambda := lambda + NutationSoleil; (* !!! *) moins prcis on dirait ! *)

    X := r * Cos(lambda); (* beta = 0 donc * Cos(beta) = * 1 ici *)
    Y := r * Sin(lambda);

    lambda := lambda + NutationSoleil; (* !!! *)

    IF Mode = full THEN
        beta   := 0.0; (* latitude du Soleil toujours infrieure  1"2 *)
        CalcAlphaDelta ();
    END;
    Store (P);
END CalcSoleil;

(* === CALC =================================================== *)

PROCEDURE Calc(planete : objet; VAR P : skypos; Cmode : calcmode);
BEGIN
    Mode := Cmode;
    P.l := 0.0;
    P.b := 0.0;
    P.r := 0.0;
    P.lambda := 0.0;
    P.beta   := 0.0;
    P.alpha  := 0.0;
    P.delta  := 0.0;
    CASE planete OF
    | soleil  : CalcSoleil (P);
    | lune    : CalcLune (P);
    | mercure : CalcMercure (P);
    | venus   : CalcVenus (P);
    | mars    : CalcMars (P);
    | jupiter : CalcJupiter (P);
    | saturne : CalcSaturne (P);
    | uranus  : CalcUranus (P);
    | neptune : CalcNeptune (P);
    | pluton  : CalcPluton (P);
    ELSE ; (* //ISO fixed silly requirement *) END;
END Calc;

(* === MAIN =================================================== *)

PROCEDURE initSkyCalc();
BEGIN
    PI         := 4.0 * ATan (1.0);
    MaxRadians := 8.0 * ATan (1.0); (* PI*2 *)
    MaxDegrees := 360.0;
    DegRad     := MaxRadians / MaxDegrees;
END initSkyCalc;

(* ------------------------------------------------------------------------ *)
(* ------------------------------------------------------------------------ *)

PROCEDURE ClearBit (VAR ensemble:flag16;bit:CARDINAL);
BEGIN
    EXCL (ensemble,bit);
END ClearBit;

PROCEDURE SetBit (VAR ensemble:flag16;bit:CARDINAL);
BEGIN
    INCL (ensemble,bit);
END SetBit;

PROCEDURE TestBit (ensemble:flag16;bit:CARDINAL) : BOOLEAN; (* 0..15 *)
BEGIN
    RETURN (bit IN ensemble );
END TestBit;

(* ------------------------------------------------------------ *)

PROCEDURE NumToStr (VAR s:ARRAY OF CHAR; val:CARDINAL);
VAR
    ok:BOOLEAN;
BEGIN
    Str.CardToStr(LONGCARD(val),s,10,ok);
END NumToStr;

PROCEDURE DumpWidth (s:ARRAY OF CHAR;width:CARDINAL;sep:ARRAY OF CHAR);
VAR
    i,len : CARDINAL;
BEGIN
    WrStr(s);
    len:=Str.Length(s);
    FOR i:= len TO (width-1) DO
        WrStr(blank);
    END;
    WrStr(sep);
END DumpWidth;

(* 1.. *)

PROCEDURE GetText (VAR text:ARRAY OF CHAR;
                     chaine:ARRAY OF CHAR;element:CARDINAL;delim:ARRAY OF CHAR);
BEGIN
    Str.Append(chaine,delim); (* safety *)
    Str.ItemS(text,chaine,delim,element-1);
END GetText;

CONST (* use coma as delim, and set data in BIT order *)
    Smois         ="janvier,fvrier,mars,avril,mai,juin,"+
                   "juillet,aot,septembre,octobre,novembre,dcembre";
    Splanete      ="Soleil,Lune,Mercure,Vnus,Mars,"+
                   "Jupiter,Saturne,Uranus,Neptune,Pluton";
    Ssigne        ="Blier,Taureau,Gmeaux,Cancer,Lion,Vierge,"+
                   "Balance,Scorpion,Sagittaire,Capricorne,Verseau,Poissons";
    Scoortype     ="hliocentriques,cliptiques,quatoriales,astrologiques";

    SplaneteAbrev ="Sol,Lun,Mer,Vn,Mar,Jup,Sat,Ura,Nep,Plu";
    SsigneAbrev   ="Bl,Tau,Gm,Can,Lio,Vie,Bal,Sco,Sag,Cap,Ver,Poi";

    sizePlaneteStr = 7;
    sizeSigneStr   = 10;
    positionStr    = "Position";
    signeStr       = "Signe";
    sizePositionStr= 8; (* "Position" *)
    sizeTxtSigne   = 5;
    txtPlanete     = "Plante";
    sizeTxtPlanete = 7;
    sizeJM         = 5; (* jj-mm *)
    sizeJMH        = 12;  (* jj-mm, hh:mm *)

PROCEDURE doFmtDMY (VAR s:ARRAY OF CHAR;d,m,y:CARDINAL);
VAR
    str : str16;
BEGIN
    IF d = 1 THEN
        Str.Copy(s,"1er");
    ELSE
        NumToStr (s,d);
    END;
    Str.Append(s,blank);
    GetText(str,Smois,m,coma);
    Str.Append(s,str);
    Str.Append(s,blank);
    NumToStr(str,y);
    Str.Append(s,str);
END doFmtDMY;

PROCEDURE doFmtHM (VAR s:ARRAY OF CHAR;h,m:CARDINAL);
VAR
    str : str16;
BEGIN
    CASE h OF
    | 0 : Str.Copy(s,"minuit");
    | 12: Str.Copy(s,"midi");
    ELSE
        NumToStr(s,h);
        Str.Append(s," heure");
        IF h # 1 THEN Str.Append(s,"s");END;
    END;
    IF m > 0 THEN
        Str.Append(s,blank);
        NumToStr(str,m);
        Str.Append(s,str);
        Str.Append(s," minute");
        IF m # 1 THEN Str.Append(s,"s");END;
    END;
END doFmtHM;

PROCEDURE doFmtCoor(VAR s:ARRAY OF CHAR;flagoutput:flag16;VAR coor:CARDINAL);
VAR
    i : CARDINAL;
BEGIN
    FOR i := firstoutput TO lastoutput DO
        IF TestBit(flagoutput,i)=TRUE THEN coor:=i; END; (* always happen ! *)
    END;
    GetText(s,Scoortype,coor+1,coma);
END doFmtCoor;

PROCEDURE doFmtLongLat (VAR s: ARRAY OF CHAR;
                        flagoutput,flaglongitude,flaglatitude:flag16;
                        VAR uselong,uselat:displayfmt);
CONST
    tstlong=0.1;
    tstlat =0.1;
VAR
    i,fmtlong,fmtlat : CARDINAL;
    str:str16;
    posinsign:BOOLEAN;
    astrologicalfmt : displayfmt; (* if astro mode enabled *)
BEGIN
    posinsign:= TestBit(flagoutput,BITastro);
    FOR i := firstlongfmt TO lastlongfmt DO
        IF TestBit(flaglongitude,i)=TRUE THEN fmtlong:=i; END; (* always happen !*)
    END;
    FOR i := firstlatfmt TO lastlatfmt DO
        IF TestBit(flaglatitude,i)=TRUE THEN fmtlat:=i; END;
    END;
    CASE fmtlong OF
    | BITlongdegmin : uselong:=DDDMM;
    | BITlongdeg    : uselong:=DDDdd;
    | BITlonghms    : uselong:=HHMMSS;
    ELSE ; (* //ISO fixed silly requirement *) END;
    CASE fmtlat OF
    | BITlatdegmin  : uselat :=latDDMM;
    | BITlatdeg     : uselat :=latDDdd;
    ELSE ; (* //ISO fixed silly requirement *) END;
    IF posinsign=TRUE THEN (* astrological mode, no latitude *)
       (* handle positions *)
       CASE uselong OF
       | DDDMM : astrologicalfmt := DDMM;
       | DDDdd : astrologicalfmt := DDd;
       | HHMMSS: astrologicalfmt := DDd;  (* silent conversion to rationality for HMS would be a nonsense here *)
       ELSE
           abort(errCannotBe);
       END;
       uselong := astrologicalfmt;

       AngleToStr(tstlong,uselong,s);
       Str.Append(s,"] dans le signe zodiacal");
       Str.Prepend(s,"[");
    ELSE
       AngleToStr(tstlong,uselong,s);
       Str.Append(s,"] et [");
       AngleToStr(tstlat,uselat,str);
       Str.Append(s,str);
       Str.Append(s,"]");
       Str.Prepend(s,"[");
    END;
END doFmtLongLat;

PROCEDURE doShowUsed (flagplanets:flag16);
CONST
    separator  = ", ";
VAR
    s   : str128;
    str : str128;
    i,len: CARDINAL;
BEGIN
    s:=empty;
    FOR i := firstplanet TO lastplanet DO
        IF TestBit(flagplanets,i)=TRUE THEN
           GetText(str,Splanete,i+1,coma);
           Print(str);
        END;
    END;
    WrLn;
END doShowUsed;

PROCEDURE doFmtEcart (VAR s: ARRAY OF CHAR;d,h,m:CARDINAL );
VAR
    str:str16;
BEGIN
    IF d = 0 THEN
        CASE h OF
        | 0 : Str.Copy(s,"minuit");
        | 12: Str.Copy(s,"midi");
        ELSE
            NumToStr(s,h);
            Str.Append(s," heure");
            IF h # 1 THEN Str.Append(s,"s");END;
        END;
        IF m > 0 THEN
            Str.Append(s,blank);
            NumToStr(str,m);
            Str.Append(s,str);
            Str.Append(s," minute");
            IF m # 1 THEN Str.Append(s,"s");END;
        END;
    ELSE
        NumToStr(s,d);
        Str.Append(s," jour");
        IF d # 1 THEN Str.Append(s,"s");END;
    END;
END doFmtEcart;

(* ------------------------------------------------------------ *)

CONST
    maxLenOne     = 11+4; (* max count of chars incliding ":::" *)
    maxLenRange   = 14+4;
    sSep          = " : ";

    sDate         = "::: Date";
    sHeure        = "::: Heure T.U.";

    sDateDepart   = "::: Date de dpart"; (* 14 *)
    sDateFin      = "::: Date de fin";
    sIntervalle   = "::: Intervalle";

    sOutput       = "::: Coordonnes"; (* 10 *)
    sFormat       = "::: Affichage";

PROCEDURE doHeaderOne (dd,mm,yyyy,hh,mn : CARDINAL;
                       flagplanets,flagoutput,flaglongitude,flaglatitude:flag16);
VAR
    s          : str128;
    dmy        : str128; (* long string for safety *)
    hm         : str128;
    coortype   : str128;
    fmtlonglat : str128;
    coorsystem    : CARDINAL;
    uselong,uselat:displayfmt;
BEGIN
    doFmtDMY (dmy,dd,mm,yyyy);
    doFmtHM  (hm, hh,mn);
    doFmtCoor(coortype,flagoutput,coorsystem);
    doFmtLongLat(fmtlonglat,flagoutput,flaglongitude,flaglatitude,
                 uselong,uselat);
    s := "::: "+ProgTitle+" "+ProgVersion+" "+ProgCopyright;
    WrLn;
    Print(s);
    WrLn;
    DumpWidth(sDate  ,maxLenOne,sSep);  Print(dmy);
    DumpWidth(sHeure ,maxLenOne,sSep);  Print(hm);
    DumpWidth(sOutput,maxLenOne,sSep);  Print(coortype);
    (* DumpWidth(sFormat,maxLenOne,sSep);  Print(fmtlonglat); *)
    WrLn;
    (* doShowUsed(flagplanets); (* seems useless! *) *)
END doHeaderOne;

PROCEDURE doHeaderRange (departDD,departMM,departYYYY,finDD,finMM,finYYYY:CARDINAL;
                         hh,mn,deltadd,deltahh,deltamm:CARDINAL;
                         flagplanets,flagoutput,flaglongitude,flaglatitude:flag16);
VAR
    s          : str128;
    debut      : str128; (* long string for safety *)
    fin        : str128;
    hm         : str128;
    intervalle : str128;
    coortype   : str128;
    fmtlonglat : str128;
    coorsystem    : CARDINAL;
    uselong,uselat:displayfmt;
BEGIN
    doFmtDMY (debut,departDD,departMM,departYYYY);
    doFmtDMY (fin,finDD,finMM,finYYYY);
    doFmtHM  (hm, hh,mn);
    doFmtEcart(intervalle,deltadd,deltahh,deltamm);
    doFmtCoor(coortype,flagoutput,coorsystem);
    doFmtLongLat(fmtlonglat,flagoutput,flaglongitude,flaglatitude,
                 uselong,uselat);
    s := "::: "+ProgTitle+" "+ProgVersion+" "+ProgCopyright;
    WrLn;
    Print(s);
    WrLn;
    DumpWidth(sDateDepart,maxLenRange,sSep);  Print(debut);
    DumpWidth(sDateFin   ,maxLenRange,sSep);  Print(fin);
    DumpWidth(sHeure     ,maxLenRange,sSep);  Print(hm);
    DumpWidth(sIntervalle,maxLenRange,sSep);  Print(intervalle);
    DumpWidth(sOutput,maxLenRange,sSep);  Print(coortype);
    (* DumpWidth(sFormat,maxLenRange,sSep);  Print(fmtlonglat); *)
    (* doShowUsed(flagplanets); *)
END doHeaderRange;

(* ------------------------------------------------------------ *)

PROCEDURE SetCalcMode (flagoutput:flag16; VAR mode:calcmode);
BEGIN
    IF TestBit(flagoutput,BITastro)=TRUE THEN
        mode:= partial;
    ELSE
        mode:= full;
    END;
END SetCalcMode;

PROCEDURE SetPads (flaglongitude,flaglatitude:flag16;
                   VAR padlong,padlat,padlatsign : CHAR );
BEGIN
    padlong    := blank;
    padlat     := blank;
    padlatsign := blank;
    IF TestBit(flaglongitude,BITlongpad)=TRUE THEN padlong:=zero; END;
    IF TestBit(flaglatitude,BITlatpadzero)=TRUE THEN padlat := zero; END;
    IF TestBit(flaglatitude,BITlatpadsign)=TRUE THEN padlatsign := plus; END;
END SetPads;

PROCEDURE SetJD (dd,mm,yyyy,hh,mn:CARDINAL) : LONGREAL;
VAR
    j,m,a : LONGREAL;
BEGIN
    j := LONGREAL(dd) + ( LONGREAL(hh) + LONGREAL(mn) / 60.0 ) / 24.0;
    m := LONGREAL(mm);
    a := LONGREAL(yyyy);
    RETURN DateToJD(j,m,a,0.0); (* 0.0 is deltaTU in days *)
END SetJD;

(* ------------------------------------------------------------ *)

PROCEDURE AngleToPos (val:LONGREAL; uselong:displayfmt;VAR p,s: ARRAY OF CHAR);
VAR
    pos : LONGREAL;
    i : CARDINAL;
    v : LONGINT;
BEGIN
    pos := MATHLIB.Mod (val,30.0);
    i := CARDINAL (val / 30.0); (* 0..11 *)

    (* handle rounding here for 29,99 or 29,9 or 30 0' etc. *)
    (* see code for DDMM and DDd functions in SkyPrint module *)
    (* could be directly implemented in these subroutines but well... *)

    CASE uselong OF
    | DDMM :
        v := VAL (LONGINT,(ABS(pos) * 60.0 + 0.5)); (* in minutes now *)
        IF v >= (30 * 60) THEN (* should never happen *)
            pos := 0.0;
            INC(i);
        END;
    | DDd  :
        v := VAL (LONGINT,(ABS(pos) * 10.0 + 0.5));
        IF v >= 300 THEN
            pos := 0.0;
            INC(i);
        END;
    ELSE
        abort(errCannotBeToo); (* should NEVER happen! *)
    END;
    AngleToStr( pos , uselong ,p );
    IF i > 11 THEN i := 0; END;
    GetText(s,Ssigne,i+1,coma); (* 1..12 *)
END AngleToPos;

(* ------------------------------------------------------------ *)

VAR
    chCSV      : CHAR;
    flagAbrev  : BOOLEAN;
    buf        : str256;
    w1,w2,w3   : CARDINAL; (* max width of each coor field *)
VAR
    coordonnees : ARRAY [ORD(soleil)..ORD(pluton)] OF skypos;

PROCEDURE InitBuf ();
BEGIN
    Str.Copy(buf,empty);
END InitBuf;

PROCEDURE AddBuf (s:ARRAY OF CHAR;w : CARDINAL; bigsep:BOOLEAN );
VAR
    len : CARDINAL;
    i   : CARDINAL;
BEGIN
    CASE chCSV OF
    | tabul,semicolon:
        Str.Append(buf,chCSV);
        Str.Append(buf,s);
    ELSE
        Str.Append(buf,s);
        len:=Str.Length(s)+1;
        IF bigsep = TRUE THEN
            INC(w,4);
        ELSE
            INC(w,2);
        END;
        FOR i := len TO w DO
            Str.Append(buf,blank);
        END;
    END;
END AddBuf;

PROCEDURE SendBuf ();
BEGIN
    CASE chCSV OF
    | tabul,semicolon:
        Str.Delete(buf,0,1);  (* leading csv is ONE char *)
    ELSE ; (* //ISO fixed silly requirement *) END;
    Rtrim(buf,blank);
    Print(buf);
END SendBuf;

(* ------------------------------------------------------------ *)

CONST
    sizeHHMMSS     = 11;
    sizeDDDdd      = 7;
    sizeDDdd       = 6;
    sizeDDd        = 5;
    sizeDDDMM      = 8;
    sizeDDMM       = 7;
    sizelatDDdd    = 7;
    sizelatDDd     = 6;
    sizelatDDMM    = 8;
    sizeDDddddd    = 8;

PROCEDURE getsize (fmt:displayfmt  ) : CARDINAL;
BEGIN
    CASE fmt OF
    | HHMMSS  : RETURN sizeHHMMSS ; (* ##h ##m ##s *)
    | DDDdd   : RETURN sizeDDDdd;   (* ###,##     *)
    | DDdd    : RETURN sizeDDdd ;   (* ##,##      *)
    | DDd     : RETURN sizeDDd  ;   (* ##,#       *)
    | DDDMM   : RETURN sizeDDDMM;   (* ### ##'    *)
    | DDMM    : RETURN sizeDDMM  ;  (* ## ##'     *)
    | latDDdd : RETURN sizelatDDdd; (* ##,##     *)
    | latDDd  : RETURN sizelatDDd ; (* ##,#      *)
    | latDDMM : RETURN sizelatDDMM; (* ## ##'    *)
    | DDddddd : RETURN sizeDDddddd; (* ##,#####    *)
    ELSE ; (* //ISO fixed silly requirement *) END;
END getsize;

PROCEDURE GetMaxWidthInfo (coorsystem:CARDINAL;uselong,uselat:displayfmt);
BEGIN
    CASE coorsystem OF
    | BITheliocentric :
         w1 := getsize(uselong);
         w2 := getsize(uselat) ;
         w3 := sizeDDddddd;
    | BITecliptic     :
         w1 := getsize(uselong);
         IF w1 < sizeLambdaStr THEN w1 := sizeLambdaStr; END; (* "lambda" takes 6 chars *)
         w2 := getsize(uselat);
    | BITequatorial   :
         w1 := getsize(uselong);
         w2 := getsize(uselat);
    | BITastro        :
         w1 := getsize(uselong);
         IF w1 < sizePositionStr THEN w1 := sizePositionStr; END; (* "Position" takes 8 chars *)
         IF sizeTxtSigne < sizeSigneStr THEN
             w2 := sizeSigneStr;
         ELSE
             w2 := sizeTxtSigne;
         END;
    ELSE ; (* //ISO fixed silly requirement *) END;
END GetMaxWidthInfo;


PROCEDURE DumpCoor (i:CARDINAL;coorsystem:CARDINAL;uselong,uselat:displayfmt);
VAR
    s : str16;
    p : str16;
BEGIN
    CASE coorsystem OF
    | BITheliocentric :
         CASE i OF
         | BITsoleil : (* only r is computed *)
             AddBuf(empty,w1,FALSE);
             AddBuf(empty,w2,FALSE);
             AngleToStr( coordonnees[i].r ,DDddddd,s );
             AddBuf(s,w3,TRUE);
         | BITlune   : (* nonsense here *)
             AddBuf(empty,w1,FALSE);
             AddBuf(empty,w2,FALSE);
             AddBuf(empty,w3,TRUE);
         ELSE
             AngleToStr( coordonnees[i].l, uselong ,s );
             AddBuf(s,w1,FALSE);
             AngleToStr( coordonnees[i].b, uselat  ,s );
             AddBuf(s,w2,FALSE);
             AngleToStr( coordonnees[i].r ,DDddddd,s );
             AddBuf(s,w3,TRUE);
         END;
    | BITecliptic     :
         AngleToStr( coordonnees[i].lambda, uselong ,s );
         AddBuf(s,w1,FALSE);
         AngleToStr( coordonnees[i].beta  , uselat ,s );
         AddBuf(s,w2,TRUE);
    | BITequatorial   :
         AngleToStr( coordonnees[i].alpha , uselong ,s );
         AddBuf(s,w1,FALSE);
         AngleToStr( coordonnees[i].delta , uselat ,s );
         AddBuf(s,w2,TRUE);
    | BITastro        :
         AngleToPos( coordonnees[i].lambda, uselong , p, s );
         AddBuf(p,w1,FALSE);
         AddBuf(s,w2,TRUE);
    ELSE ; (* //ISO fixed silly requirement *) END;
END DumpCoor;

PROCEDURE doOne (dd,mm,yyyy,hh,mn:CARDINAL;
                 flagplanets,flagoutput,flaglongitude,flaglatitude:flag16;
                 tudefined,tuapplycli:BOOLEAN;tudeltadays:LONGREAL);
VAR
    jd : LONGREAL;
    i : CARDINAL;
    mode : calcmode;
    s : str16;
    ok : BOOLEAN;
    padlong,padlat,padlatsign : CHAR;
    signe : CARDINAL;

    coorsystem    : CARDINAL;
    uselong,uselat:displayfmt;

    w0 : CARDINAL;
BEGIN
    SetCalcMode(flagoutput,mode);
    SetPads(flaglongitude,flaglatitude,padlong,padlat,padlatsign);
    PadsSetup(padlong,padlat,padlatsign);

    jd:=SetJD(dd,mm,yyyy,hh,mn);
    IF (tudefined AND tuapplycli) THEN
        jd:= jd + tudeltadays;
        JDtoDate(jd,dd,mm,yyyy,hh,mn); (* recalc date and time ! *)
    END;

    doHeaderOne(dd,mm,yyyy,hh,mn,
                flagplanets,flagoutput,flaglongitude,flaglatitude);

    doFmtCoor(buf,flagoutput,coorsystem); (* dummy string, just to reset vars *)
    doFmtLongLat(buf,flagoutput,flaglongitude,flaglatitude,
                 uselong,uselat); (* dummy string, just to reset vars *)

    CalcT (jd);
    CalcSetup;

    (* v1.1 : SOLEIL must ALWAYS be computed even if not selected !!! *)

    Calc(soleil, coordonnees[ORD(soleil)], mode); (* ALWAYS !!! *)

    FOR i := ORD(lune) TO ORD(pluton) DO (* useless to redo sun even if specified *)
        IF TestBit(flagplanets,i)=TRUE THEN
            Calc(VAL(objet,i), coordonnees[i], mode);
        END;
    END;
    (* and now... dump'em, boy! *)

    GetMaxWidthInfo(coorsystem,uselong,uselat);
    InitBuf();

    IF sizePlaneteStr < sizeTxtPlanete THEN
        w0 := sizePlaneteStr;
    ELSE
        w0 := sizeTxtPlanete;
    END;
    AddBuf(txtPlanete,w0,TRUE);
    CASE coorsystem OF
    | BITheliocentric :
        AddBuf("l",w1,FALSE);
        AddBuf("b",w2,FALSE);
        AddBuf("r",w3,TRUE);
    | BITecliptic     :
        AddBuf(strlambda,w1,FALSE);
        AddBuf(strbeta,w2,TRUE);
    | BITequatorial   :
        AddBuf(stralpha,w1,FALSE);
        AddBuf(strdelta,w2,TRUE);
    | BITastro        :
        AddBuf(positionStr,w1,FALSE);
        AddBuf(signeStr,w2,TRUE);
    ELSE ; (* //ISO fixed silly requirement *) END;
    SendBuf();
    WrLn;
    FOR i := ORD(soleil) TO ORD(pluton) DO
        IF TestBit(flagplanets,i)=TRUE THEN
            GetText(s,Splanete,i+1,coma);
            InitBuf();
            AddBuf(s,sizePlaneteStr,TRUE);
            DumpCoor(i,coorsystem,uselong,uselat);
            SendBuf();
        END;
    END;
END doOne;

(* ------------------------------------------------------------ *)


PROCEDURE TwoToStr (VAR s:ARRAY OF CHAR; val:CARDINAL);
VAR
    ok:BOOLEAN;
BEGIN
    Str.CardToStr(LONGCARD(val),s,10,ok);
    IF val < 10 THEN Str.Prepend(s,"0"); END;
END TwoToStr;

PROCEDURE fmtDateTime (jour,mois,annee,heure,minute:CARDINAL;
                       VAR strJJMM,strHHMM,strYYYY : ARRAY OF CHAR );
VAR
    s : str16;
BEGIN
    TwoToStr(strJJMM,jour);
    TwoToStr(s,mois);
    Str.Append(strJJMM,dash);
    Str.Append(strJJMM,s);

    TwoToStr(strHHMM,heure);
    TwoToStr(s,minute);
    Str.Append(strHHMM,doubledot);
    Str.Append(strHHMM,s);

    NumToStr(strYYYY,annee);
END fmtDateTime;

PROCEDURE doRange (departDD,departMM,departYYYY,finDD,finMM,finYYYY:CARDINAL;
                   deltadd,deltahh,deltamm,hh,mm:CARDINAL;
                   flagplanets,flagoutput,flaglongitude,flaglatitude:flag16);
VAR
    firstjd,lastjd,jd : LONGREAL;
    deltadays : LONGREAL;
    i : CARDINAL;
    mode : calcmode;
    s : str16;
    ok : BOOLEAN;
    padlong,padlat,padlatsign : CHAR;

    coorsystem    : CARDINAL;
    uselong,uselat:displayfmt;

    jour,mois,annee,heure,minute,oldannee:CARDINAL;
    strJJMM,strHHMM,strYYYY : str16;
    w0 : CARDINAL;
BEGIN
    SetCalcMode(flagoutput,mode);
    SetPads(flaglongitude,flaglatitude,padlong,padlat,padlatsign);
    PadsSetup(padlong,padlat,padlatsign);
    firstjd:=SetJD(departDD,departMM,departYYYY,hh,mm);
    lastjd :=SetJD(finDD,finMM,finYYYY,hh,mm);
    (* deltadays *)
    IF deltadd = 0 THEN
        deltadays := ( LONGREAL(deltahh) + LONGREAL(deltamm) / 60.0 ) / 24.0;
    ELSE
        deltadays := LONGREAL(deltadd);
    END;
    doHeaderRange(departDD,departMM,departYYYY,finDD,finMM,finYYYY,
                  hh,mm,deltadd,deltahh,deltamm,
                  flagplanets,flagoutput,flaglongitude,flaglatitude);

    doFmtCoor(buf,flagoutput,coorsystem);
    doFmtLongLat(buf,flagoutput,flaglongitude,flaglatitude,
                 uselong,uselat);

    GetMaxWidthInfo(coorsystem,uselong,uselat);
    InitBuf();

    IF deltadd = 0 THEN
        w0 := sizeJMH;
    ELSE
        w0 := sizeJM;
    END;

    oldannee := 0;

    jd := firstjd;
    LOOP
        CalcT (jd);
        CalcSetup;

        (* v1.1 : SOLEIL must ALWAYS be computed even if not selected !!! *)

        Calc(soleil, coordonnees[ORD(soleil)], mode); (* ALWAYS !!! *)

        FOR i := ORD(lune) TO ORD(pluton) DO (* useless to calc sun again even if specified *)
            IF TestBit(flagplanets,i)=TRUE THEN
               Calc(VAL(objet,i), coordonnees[i], mode);
            END;
        END;
        JDtoDate(jd,jour,mois,annee,heure,minute);
        fmtDateTime(jour,mois,annee,heure,minute,strJJMM,strHHMM,strYYYY);
        IF annee # oldannee THEN
            InitBuf();
            SendBuf();
            AddBuf(strYYYY,w0,TRUE);
            FOR i := ORD(soleil) TO ORD(pluton) DO
                IF TestBit(flagplanets,i)=TRUE THEN
                    GetText(s,Splanete,i+1,coma);
                    AddBuf(s,w1,FALSE);
                    CASE coorsystem OF
                    | BITheliocentric :
                        AddBuf(empty,w2,FALSE);
                        AddBuf(empty,w3,TRUE);
                    | BITecliptic     :
                        AddBuf(empty,w2,TRUE);
                    | BITequatorial   :
                        AddBuf(empty,w2,TRUE);
                    | BITastro        :
                        AddBuf(empty,w2,TRUE);
                    ELSE ; (* //ISO fixed silly requirement *) END;
                END;
            END;
            SendBuf();
            InitBuf();
            AddBuf(empty,w0,TRUE);
            FOR i := ORD(soleil) TO ORD(pluton) DO
                IF TestBit(flagplanets,i)=TRUE THEN
                    CASE coorsystem OF
                    | BITheliocentric :
                        AddBuf("l",w1,FALSE);
                        AddBuf("b",w2,FALSE);
                        AddBuf("r",w3,TRUE);
                    | BITecliptic     :
                        AddBuf(strlambda,w1,FALSE);
                        AddBuf(strbeta,w2,TRUE);
                    | BITequatorial   :
                        AddBuf(stralpha,w1,FALSE);
                        AddBuf(strdelta,w2,TRUE);
                    | BITastro        :
                        AddBuf(positionStr,w1,FALSE);
                        AddBuf(signeStr,w2,TRUE);
                    ELSE ; (* //ISO fixed silly requirement *) END;
                END;
            END;
            SendBuf();
            InitBuf();
            SendBuf();
            oldannee:=annee;
        END;
        InitBuf();
        Str.Copy(s,strJJMM);
        IF deltadd = 0 THEN
            Str.Append(s,", ");
            Str.Append(s,strHHMM);
        END;
        AddBuf(s,w0,TRUE);
        FOR i := ORD(soleil) TO ORD(pluton) DO
            IF TestBit(flagplanets,i)=TRUE THEN
                DumpCoor(i,coorsystem,uselong,uselat); (* later handle short *)
            END;
        END;
        SendBuf();
        jd := jd + deltadays;
        IF jd > lastjd THEN EXIT; END; (* >= *)
    END;
END doRange;


(* ------------------------------------------------------------ *)

PROCEDURE GetCardinal (s : ARRAY OF CHAR) : CARDINAL;
VAR
    p : CARDINAL;
    v : LONGCARD;
    ok : BOOLEAN;
BEGIN
    Replace(s,equal,doubledot); (* command line option xxx= becomes xxx: *)
    p := Str.CharPos(s,doubledot); (* MAX(CARDINAL) already excluded! *)
    Str.Delete(s,0,p+1);
    v:=Str.StrToCard(s,10,ok);
    IF ok = FALSE THEN abort(errConversion); END;
    IF v > MAX(CARDINAL) THEN abort(errNotCardinal); END;
    RETURN CARDINAL(v);
END GetCardinal;

(* ------------------------------------------------------------ *)

(* get value then remove from string *)

PROCEDURE GetNum (VAR S : ARRAY OF CHAR; p : CARDINAL) : CARDINAL;
VAR
    str : str16;
    v : LONGCARD;
    ok : BOOLEAN;
BEGIN
    Str.Slice (str,S,0,p); (* retrieve leading num *)
    Str.Delete (S,0,p+1); (* remove num and separator *)
    v:=Str.StrToCard(str,10,ok);
    IF ok = FALSE THEN abort(errConversion); END;
    IF v > MAX(CARDINAL) THEN abort(errNotCardinal); END;
    RETURN CARDINAL(v);
END GetNum;

(* get last value *)

PROCEDURE GetLastNum (VAR S : ARRAY OF CHAR) : CARDINAL;
VAR
    v : LONGCARD;
    ok : BOOLEAN;
BEGIN
    v:=Str.StrToCard(S,10,ok);
    IF ok = FALSE THEN abort(errConversion); END;
    IF v > MAX(CARDINAL) THEN abort(errNotCardinal); END;
    RETURN CARDINAL(v);
END GetLastNum;

PROCEDURE ChkDateNow (VAR s : ARRAY OF CHAR) : BOOLEAN;
VAR
    jj,mm,aa : CARDINAL;
    dow : Lib.DayType;
    ok : BOOLEAN;
    sjj,smm,saa : str16;
BEGIN
    ok:=FALSE;
    IF Str.Compare(s,now)=0 THEN
        Lib.GetDate(aa,mm,jj,dow);
        Str.CardToStr(LONGCARD(jj), sjj ,10,ok);
        Str.CardToStr(LONGCARD(mm), smm ,10,ok);
        Str.CardToStr(LONGCARD(aa), saa ,10,ok);
        Str.Copy (s, sjj);
        Str.Append(s,dash);
        Str.Append(s,smm);
        Str.Append(s,dash);
        Str.Append(s,saa);
        ok:=TRUE;
    END;
    RETURN ok;
END ChkDateNow;

PROCEDURE ChkTimeNow (VAR s : ARRAY OF CHAR) : BOOLEAN;
VAR
    hh,mm,ss,centiemes : CARDINAL;
    ok : BOOLEAN;
    shh,smm : str16;
BEGIN
    ok:=FALSE;
    IF Str.Compare(s,now)=0 THEN
        Lib.GetTime (hh,mm,ss,centiemes);
        Str.CardToStr(LONGCARD(hh), shh ,10,ok);
        Str.CardToStr(LONGCARD(mm), smm ,10,ok); (* tant pis pour les secondes ! *)
        Str.Copy (s, shh);
        Str.Append(s,hourdelim);
        Str.Append(s,smm);
        ok:=TRUE;
    END;
    RETURN ok;
END ChkTimeNow;

(*

(* dd-mm-yyyy, separator is / : -, and if yyyy < 100, add 1900 *)
PROCEDURE ParseDate (s:ARRAY OF CHAR; VAR dd,mm,yyyy:CARDINAL ) : BOOLEAN;
VAR
    p : CARDINAL;
BEGIN
    Replace (s,dash,doubledot);  (* if separator - *)
    Replace (s,slash,doubledot); (* if separator / *)
    p := Str.CharPos(s,doubledot);
    IF p = MAX(CARDINAL) THEN RETURN FALSE; END;
    dd := GetNum(s,p);
    p := Str.CharPos(s,doubledot);
    IF p = MAX(CARDINAL) THEN RETURN FALSE; END;
    mm := GetNum(s,p);
    IF Str.Length(s)=0 THEN RETURN FALSE; END;
    yyyy := GetLastNum(s);
    IF yyyy < 100 THEN INC(yyyy,1900); END;
    RETURN TRUE;
END ParseDate;

*)

(* hhHmm or hhH, separator is H or : *)
PROCEDURE ParseHour (s:ARRAY OF CHAR;VAR hh,mm:CARDINAL  ) : BOOLEAN;
VAR
    p : CARDINAL;
BEGIN
    UpperCase(s); (* already done but... *)
    Replace (s,hourdelim,doubledot);  (* if separator H *)
    p := Str.CharPos(s,doubledot);
    IF p = MAX(CARDINAL) THEN RETURN FALSE; END;
    hh := GetNum(s,p);
    IF Str.Length(s)=0 THEN
        mm := 0;
    ELSE
        mm := GetLastNum(s);
    END;
    RETURN TRUE;
END ParseHour;

(* nn, hhHmm or hhH, separator is H or : *)
PROCEDURE ParseInterval (s:ARRAY OF CHAR; VAR dd,hh,mm:CARDINAL ) : BOOLEAN;
VAR
    p : CARDINAL;
BEGIN
    UpperCase(s); (* already done but... *)
    Replace (s,hourdelim,doubledot);  (* if separator H *)
    p := Str.CharPos(s,doubledot);
    IF p = MAX(CARDINAL) THEN
        dd := GetLastNum(s);
        hh := 0;
        mm := 0;
    ELSE
        hh := GetNum(s,p);
        IF Str.Length(s)=0 THEN
            mm := 0;
        ELSE
            mm := GetLastNum(s);
        END;
        dd := 0;
    END;
    RETURN TRUE;
END ParseInterval;

PROCEDURE ChkDate (flagUNSAFE:BOOLEAN; dd,mm,yyyy:CARDINAL  );
BEGIN
    IF (dd < 1) OR (dd > 31) THEN abort(errDMY); END;
    IF (mm < 1) OR (mm > 12) THEN abort(errDMY); END;
    IF flagUNSAFE=FALSE THEN
        IF (yyyy < minYear) OR (yyyy > maxYear) THEN abort(errDMY); END;
    END;
END ChkDate;

PROCEDURE ChkHour ( hh,mm:CARDINAL   );
BEGIN
    IF (hh < 0) OR (hh > 23) THEN abort(errHM); END;
    IF (mm < 0) OR (mm > 59) THEN abort(errHM); END;
END ChkHour;

PROCEDURE ChkInterval ( dd,hh,mm:CARDINAL  );
BEGIN
    CASE dd OF (* probably useless but... *)
    | 0 :
        ChkHour(hh,mm);
    ELSE
        IF (hh # 0) OR (mm # 0) THEN abort(errIntervalSpec); END;
        IF (dd < minInterval) OR (dd > maxInterval) THEN abort(errIntervalRange); END;
    END;
END ChkInterval;

PROCEDURE ChkLongFmt (s : ARRAY OF CHAR; VAR flagformat : flag16  );
VAR
    n : CARDINAL;
BEGIN
    UpperCase(s);
    IF isOption(s)=FALSE THEN Str.Prepend(s,dash);END; (* force option *)
    n := GetOptIndex (s, "0HMS"+delim+
                         "0'"+delim+"0DEGMIN"+delim+
                         "0"+delim+"0DEG"+delim+
                         "HMS"+delim+
                         "'"+delim+"DEGMIN"+delim+
                         ""+delim+"DEG");
    CASE n OF
    | 1 :
         SetBit(flagformat,BITlongpad);
         SetBit(flagformat,BITlonghms);
    | 2,3 :
         SetBit(flagformat,BITlongpad);
         SetBit(flagformat,BITlongdegmin);
    | 4,5 :
         SetBit(flagformat,BITlongpad);
         SetBit(flagformat,BITlongdeg);
    | 6 :
         ClearBit(flagformat,BITlongpad);
         SetBit(flagformat,BITlonghms);
    | 7,8 :
         ClearBit(flagformat,BITlongpad);
         SetBit(flagformat,BITlongdegmin);
    | 9,10 :
         ClearBit(flagformat,BITlongpad);
         SetBit(flagformat,BITlongdeg);
    ELSE
         abort(errLongFmt);
    END;
END ChkLongFmt;

PROCEDURE ChkLatFmt (s : ARRAY OF CHAR; VAR flagformat : flag16 );
VAR
    n : CARDINAL;
BEGIN
    UpperCase(s);
    IF s[0] = plus THEN
        SetBit(flagformat,BITlatpadsign);
        Str.Delete(s,0,1);
    END;
    IF s[0] = zero THEN
        SetBit(flagformat,BITlatpadzero);
        Str.Delete(s,0,1);
    END;

    IF isOption(s)=FALSE THEN Str.Prepend(s,dash);END; (* force option *)
    n := GetOptIndex (s, "'"+delim+"DEGMIN"+delim+
                         ""+delim+"DEG");
    CASE n OF
    | 1,2 :
         SetBit(flagformat,BITlatdegmin);
    | 3,4 :
         SetBit(flagformat,BITlatdeg);
    ELSE
         abort(errLatFmt);
    END;
END ChkLatFmt;

PROCEDURE ChkFmtCSV (s : ARRAY OF CHAR; VAR ch:CHAR );
VAR
    n:CARDINAL;
BEGIN
    UpperCase(s);
    IF isOption(s)=FALSE THEN Str.Prepend(s,dash);END;
    n := GetOptIndex (s, "ESPACE"+delim+"SPACE"+delim+"BLANK"+delim+
                         "TAB"+delim+"TABUL"+delim+"TABULATION"+delim+
                         "POINTVIRGULE"+delim+"SEMICOLON");
    CASE n OF
    | 1,2,3: ch:=blank;
    | 4,5,6: ch:=tabul;
    | 7,8,9: ch:=semicolon;
    ELSE
        abort(errCSVfmt);
    END;
END ChkFmtCSV;

(* ------------------------------------------------------------ *)

VAR
    POURTUdeltaDays : LONGREAL;
    POURTUdefined   : BOOLEAN;
    POURTUstring    : str128;
    POURTUapplyCLI  : BOOLEAN;

PROCEDURE chkPOURTU () : BOOLEAN;
VAR
    rc, useLocal : BOOLEAN;
BEGIN
    (* default just in case ! *)
    POURTUdeltaDays := 0.0;
    POURTUdefined   := FALSE;

    Lib.EnvironmentFind(sPOURTU,POURTUstring); (* uppercase *)
    IF same(POURTUstring,"") THEN
        RETURN TRUE;            (* not defined, so success *)
    ELSE
        rc:=parseDeltaTU(POURTUstring,POURTUdeltaDays,useLocal); (* P is ok, but L is not ! *)
        IF rc = FALSE THEN
            POURTUdeltaDays := 0.0; (* hey, you never know... ;-) *)
            RETURN FALSE; (* defined and illegal *)
        END;
        IF useLocal THEN
            POURTUdeltaDays := 0.0; (* hey, you never know... ;-) *)
            RETURN FALSE; (* defined but illegal *)
        ELSE
            POURTUdefined := TRUE;
            RETURN TRUE; (* defined and legal *)
        END;
    END;
END chkPOURTU;

(* ------------------------------------------------------------ *)

VAR
    parms : CARDINAL;
    i     : CARDINAL;
    S     : str128;
    opt   : CARDINAL;
    tmp   : str16;

    flagdepart    : BOOLEAN;
    flagfin       : BOOLEAN;
    flaginterval  : BOOLEAN;
    flagdate      : BOOLEAN;
    flaghour      : BOOLEAN;

    intervalDD: CARDINAL;
    intervalHH: CARDINAL;
    intervalMM: CARDINAL;
    dateDD    : CARDINAL;
    dateMM    : CARDINAL;
    dateYYYY  : CARDINAL;
    hourHH    : CARDINAL;
    hourMM    : CARDINAL;

    departDD    : CARDINAL;
    departMM    : CARDINAL;
    departYYYY  : CARDINAL;
    finDD    : CARDINAL;
    finMM    : CARDINAL;
    finYYYY  : CARDINAL;

    flagplanets : flag16;
    flagoutput  : flag16;
    flaglongitude: flag16;
    flaglatitude : flag16;
    flagUNSAFE   : BOOLEAN;
    DEBUG : BOOLEAN;
    departjd,finjd:LONGREAL;

    dateWasNow, timeWasNow : BOOLEAN;
BEGIN

    parms := Lib.ParamCount();
    IF parms = 0 THEN abort (errHelp); END;

    (* init all variables and flags *)
    flagdepart    := FALSE;
    flagfin       := FALSE;
    flaginterval  := FALSE;
    flagdate      := FALSE;
    flaghour      := FALSE;
    flagAbrev     := FALSE;     (* global: Q&D!!! *)
    flagplanets   := flag16{};
    flagoutput    := flag16{};
    flaglongitude := flag16{};
    flaglatitude  := flag16{};
    chCSV         := blank;
    DEBUG	  := FALSE;
    flagUNSAFE    := FALSE;

    initSkyCalc();


    FOR i:= 1 TO parms DO
        Lib.ParamStr(S,i); cleantabs(S);
	fixCLI(S);
        UpperCase(S);
        IF isOption(S)=FALSE THEN Str.Prepend(S,dash); END; (* force option *)
        opt := GetOptIndex (S,
                              "UNSAFE"+delim+"U"+delim+
			      "DEBUG"
                           );
        CASE opt OF
        | 1,2: flagUNSAFE := TRUE;
	| 3 : DEBUG:=TRUE;
        ELSE ; (* //ISO fixed silly requirement *) END;
    END;

    FOR i := 1 TO parms DO
        Lib.ParamStr(S,i); cleantabs(S);
    (* fixCLI(S); (* win to PC *) *)
        UpperCase(S); (* accents lost *)
        IF isOption(S)=FALSE THEN Str.Prepend(S,dash); END; (* force option *)
IF DEBUG THEN WrStr("parm : ");WrStr(S);WrLn;END;
        opt := GetOptIndex (S,"?"+delim+"AIDE"+delim+"HELP"+delim+
                              "DEPART="+delim+"DE="+delim+
                              "FIN="+delim+"FI="+delim+
                              "INTERVALLE="+delim+"I="+delim+
                              "DATE="+delim+"D="+delim+
                              "HEURE="+delim+"H="+delim+
                              "SOLEIL"+delim+"SO"+delim+
                              "LUNE"+delim+"LU"+delim+
                              "MERCURE"+delim+"ME"+delim+
                              "VENUS"+delim+"VE"+delim+
                              "MARS"+delim+"MA"+delim+
                              "JUPITER"+delim+"JU"+delim+
                              "SATURNE"+delim+"SA"+delim+
                              "URANUS"+delim+"UR"+delim+
                              "NEPTUNE"+delim+"NE"+delim+
                              "PLUTON"+delim+"PL"+delim+
                              "HELIOCENTRIC"+delim+"LBR"+delim+
                              "ECLIPTIC"+delim+"LAMBDA"+delim+
                              "EQUATORIAL"+delim+"ALPHA"+delim+
                              "ASTROLOGY"+delim+"ASTRO"+delim+
                              "FMTLONG="+delim+"LO="+delim+
                              "FMTLAT="+delim+"LA="+delim+
                              "ABREGE"+delim+"SHORT"+delim+
                              "UNSAFE"+delim+"U"+delim+
                              "CSV="+delim+"SEP="+delim+
			      "DEBUG"
                              );
        CASE opt OF
        | 1,2,3 :
            abort(errHelp);

        | 4,5 :
            GetString(S,tmp); (* date de dpart *)
            (*         IF ParseDate(tmp,departDD,departMM,departYYYY) = FALSE THEN abort(errDate); END; *)
            IF parseDate(flagUNSAFE,tmp,departDD,departMM,departYYYY)=FALSE THEN abort(errDate); END;
            ChkDate(flagUNSAFE,departDD,departMM,departYYYY);
            flagdepart := TRUE;
        | 6,7 :
            GetString(S,tmp); (* date de fin *)
            (*         IF ParseDate(tmp,finDD,finMM,finYYYY) = FALSE THEN abort(errDate); END; *)
            IF parseDate(flagUNSAFE,tmp,finDD,finMM,finYYYY)=FALSE THEN abort(errDate); END;
            ChkDate(flagUNSAFE,finDD,finMM,finYYYY);
            flagfin    := TRUE;
        | 8,9 :
            GetString(S,tmp); (* intervalle *)
            IF ParseInterval(tmp,intervalDD,intervalHH,intervalMM) = FALSE THEN abort(errInterval); END;
            ChkInterval(intervalDD,intervalHH,intervalMM);
            flaginterval := TRUE;

        | 10,11 :
            GetString(S,tmp); (* date *)
            dateWasNow := ChkDateNow(tmp);
            (*         IF ParseDate(tmp,dateDD,dateMM,dateYYYY) = FALSE THEN abort(errDate); END; *)
            IF parseDate(flagUNSAFE,tmp,dateDD,dateMM,dateYYYY)=FALSE THEN abort(errDate); END;
            ChkDate(flagUNSAFE,dateDD,dateMM,dateYYYY);
            flagdate := TRUE;
        | 12,13 :
            GetString(S,tmp); (* heure *)
            timeWasNow := ChkTimeNow(tmp);
            IF ParseHour(tmp,hourHH,hourMM)=FALSE THEN abort(errHour); END;
            ChkHour(hourHH,hourMM);
            flaghour := TRUE;

        | 14,15 :
            SetBit(flagplanets,BITsoleil);
        | 16,17 :
            SetBit(flagplanets,BITlune);
        | 18,19 :
            SetBit(flagplanets,BITmercure);
        | 20,21 :
            SetBit(flagplanets,BITvenus);
        | 22,23 :
            SetBit(flagplanets,BITmars);
        | 24,25 :
            SetBit(flagplanets,BITjupiter);
        | 26,27 :
            SetBit(flagplanets,BITsaturne);
        | 28,29 :
            SetBit(flagplanets,BITuranus);
        | 30,31 :
            SetBit(flagplanets,BITneptune);
        | 32,33 :
            SetBit(flagplanets,BITpluton);

        | 34,35 :
            SetBit(flagoutput,BITheliocentric);
        | 36,37 :
            SetBit(flagoutput,BITecliptic);
        | 38,39 :
            SetBit(flagoutput,BITequatorial);
        | 40,41 :
            SetBit(flagoutput,BITastro);

        | 42,43 : (* flagformat for longitude *)
            GetString(S,tmp); (* [0]<HMS|'|> *)
            ChkLongFmt(tmp,flaglongitude);
        | 44,45 : (* flagformat for latitude *)
            GetString(S,tmp); (* [+][0]<'|> *)
            ChkLatFmt(tmp,flaglatitude);
        | 46,47 :
            flagAbrev := TRUE;
        | 48,49:
            (* flagUNSAFE := TRUE; *)
            ; (* already handled *)
        | 50,51:
            GetString(S,tmp);
            ChkFmtCSV(tmp,chCSV);
	| 52 :
	    DEBUG := TRUE;
        ELSE
            abort(errUnknown);
        END;
    END;
    (* consistency tests *)
    IF flagdepart = TRUE THEN
        IF flagfin = FALSE THEN abort(errMissingEnd); END;
        IF flaginterval = FALSE THEN abort(errMissingInterval); END;
        IF flaghour = FALSE THEN abort(errMissingTime); END;

        departjd:=SetJD(departDD,departMM,departYYYY,0,0);
        finjd:=SetJD(finDD,finMM,finYYYY,0,0);

        IF finjd < departjd THEN abort(errDeltaYears);END; (* allow start=end *)

        (* good years *)
        IF flagdate = TRUE THEN abort(errNotBoth); END;
    ELSE
        IF flagfin = TRUE THEN abort(errMissingStart); END;
        (* date and hour ? *)
        IF (flagdate = FALSE) OR (flaghour = FALSE) THEN abort(errNeedBoth);END;
    END;
    IF flagplanets = flag16 {} THEN
        (* FOR i := firstplanet TO lastplanet DO
            SetBit(flagplanets,i);
        END; *)
        flagplanets := flag16{BITsoleil,BITlune,
                              BITmercure,BITvenus,BITmars,
                              BITjupiter,BITsaturne,
                              BITuranus,BITneptune,BITpluton}; (* default *)
    END;
    (* check only one output type *)
    opt:=0;
    FOR i := firstoutput TO lastoutput DO
        IF TestBit(flagoutput,i)=TRUE THEN INC(opt); END;
    END;
    IF opt = 0 THEN
        flagoutput := flag16 {BITastro}; (* default is ASTRO *)
    ELSIF opt > 1 THEN
        abort(errOnlyOneOutput);
    END;
    (* check longitude format *)
    opt := 0;
    FOR i := firstlongfmt TO lastlongfmt DO
        IF TestBit(flaglongitude,i)=TRUE THEN INC(opt); END;
    END;
    IF opt = 0 THEN
        flaglongitude := flag16{BITlongdeg}; (* default is DDD,DD *)
    ELSIF opt > 1 THEN
        abort(errLongFmtRepeated);
    END;
    (* check latitude format *)
    opt := 0;
    FOR i := firstlatfmt TO lastlatfmt DO
        IF TestBit(flaglatitude,i)=TRUE THEN INC(opt); END;
    END;
    IF opt = 0 THEN
        flaglatitude := flag16{BITlatdeg}; (* default is DD,DD *)
    ELSIF opt > 1 THEN
        abort(errLatFmtRepeated);
    END;

    IF chkPOURTU()=FALSE THEN abort(errBadEnvVar); END;
    POURTUapplyCLI:=(dateWasNow AND timeWasNow);

    rangeWarning(flagUNSAFE,1,0);

    IF flagdate = TRUE THEN (* only one date to dump *)
        (* take now for date and time into account here only ! *)
        doOne   (dateDD,dateMM,dateYYYY,
                 hourHH,hourMM,
                 flagplanets,flagoutput,flaglongitude,flaglatitude,
                 POURTUdefined,POURTUapplyCLI,POURTUdeltaDays);
    ELSE

        doRange (departDD,departMM,departYYYY,
                 finDD,finMM,finYYYY,
                 intervalDD,intervalHH,intervalMM,
                 hourHH,hourMM,
                 flagplanets,flagoutput,flaglongitude,flaglatitude);

    END;
    abort(errNone);
END SolarSys32.


